/**************************************************************************/ 
/*                                                                        */ 
/*            Copyright (c) 1996-2007 by Express Logic Inc.               */ 
/*                                                                        */ 
/*  This Original Work may be modified, distributed, or otherwise used in */ 
/*  any manner with no obligations other than the following:              */ 
/*                                                                        */ 
/*    1. This legend must be retained in its entirety in any source code  */ 
/*       copies of this Work.                                             */ 
/*                                                                        */ 
/*    2. This software may not be used in the development of an operating */
/*	 system product.                                                  */ 
/*                                                                        */  
/*  This Original Work is hereby provided on an "AS IS" BASIS and WITHOUT */ 
/*  WARRANTY, either express or implied, including, without limitation,   */ 
/*  the warranties of NON-INFRINGEMENT, MERCHANTABILITY or FITNESS FOR A  */ 
/*  PARTICULAR PURPOSE. THE ENTIRE RISK AS TO THE QUALITY OF this         */ 
/*  ORIGINAL WORK IS WITH the user.                                       */ 
/*                                                                        */ 
/*  Express Logic, Inc. reserves the right to modify this software        */ 
/*  without notice.                                                       */ 
/*                                                                        */ 
/*  Express Logic, Inc.                     info@expresslogic.com         */
/*  11423 West Bernardo Court               http://www.expresslogic.com   */
/*  San Diego, CA  92127                                                  */
/*                                                                        */
/**************************************************************************/

/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  **
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  **
 **  <http://www.quasarsoft.com/license.html>                                       **
 **                                                                                 **
 *************************************************************************************/



/**************************************************************************/
/**************************************************************************/
/**                                                                       */ 
/** Thread-Metric Component                                              */
/**                                                                       */
/**   Porting Layer (Must be completed with RTOS specifics)               */
/**                                                                       */
/**************************************************************************/
/**************************************************************************/

// PIC24HJ256GP610 Configuration Bit Settings

#include <xc.h>

// FBS
#pragma config BWRP = WRPROTECT_OFF     // Boot Segment Write Protect (Boot Segment may be written)
#pragma config BSS = NO_FLASH           // Boot Segment Program Flash Code Protection (No Boot program Flash segment)
#pragma config RBS = NO_RAM             // Boot Segment RAM Protection (No Boot RAM)

// FSS
#pragma config SWRP = WRPROTECT_OFF     // Secure Segment Program Write Protect (Secure segment may be written)
#pragma config SSS = NO_FLASH           // Secure Segment Program Flash Code Protection (No Secure Segment)
#pragma config RSS = NO_RAM             // Secure Segment Data RAM Protection (No Secure RAM)

// FGS
#pragma config GWRP = OFF               // General Code Segment Write Protect (User program memory is not write-protected)
#pragma config GSS = OFF                // General Segment Code Protection (User program memory is not code-protected)

// FOSCSEL
#pragma config FNOSC = PRIPLL           // Oscillator Mode (Primary Oscillator (XT, HS, EC) w/ PLL)
#pragma config IESO = ON                // Two-speed Oscillator Start-Up Enable (Start up with FRC, then switch)

// FOSC
#pragma config POSCMD = EC              // Primary Oscillator Source (EC Oscillator Mode)
#pragma config OSCIOFNC = ON            // OSC2 Pin Function (OSC2 pin has digital I/O function)
#pragma config FCKSM = CSDCMD           // Clock Switching and Monitor (Both Clock Switching and Fail-Safe Clock Monitor are disabled)

// FWDT
#pragma config WDTPOST = PS32768        // Watchdog Timer Postscaler (1:32,768)
#pragma config WDTPRE = PR128           // WDT Prescaler (1:128)
#pragma config WINDIS = OFF             // Watchdog Timer Window (Watchdog Timer in Non-Window mode)
#pragma config FWDTEN = OFF             // Watchdog Timer Enable (Watchdog timer enabled/disabled by user software)

// FPOR
#pragma config FPWRT = PWR128           // POR Timer Value (128ms)

// FICD
#pragma config ICS = PGD1               // Comm Channel Select (Communicate on PGC1/EMUC1 and PGD1/EMUD1)
#pragma config JTAGEN = OFF             // JTAG Port Enable (JTAG is Disabled)


#define VARIABLE_MEMORY 1
#include "../qKernel.h"

/* Include necessary files.  */
#include <stdio.h>
#include    "tm_api.h"
pTCB all_tcb[32];
pPIP all_que[32];
pSEM all_sem[32];
#if VARIABLE_MEMORY==1
	pMPL all_mem[32];
#else
	pFIX all_mem[32];
#endif
/**************************************************************************************

	This test uses for all its functionality threads. It works but Q-Kernel has more 
	advanced mechanisms to do the same task that can be implemented with a thread.
	
	Q-Kernel is a Dual Mode RTOS that provides fibers that are optimized for cooperative 
	scheduling and very fast interrupt response to support the tight time window for 
	high data-load tasks, like DSP. Fibers are single stack, stateless and operate 
	"closer" to the interrupt. Fibers don't require a full context switch to become 
	active.
	
	The Thread Metrics tests are specifically designed for threads and don't show the
	versatility and power of Q-Kernel and don't use advanced mechanisms like fibers.
	
	All tests are using QKernelFast.a as object library. This is the fastest library that
	have the stack limit features and the variable psv-page disabled and are compiled 
	with the -O3 switch.

******************************************************************************************/

extern void tm_interrupt_preemption_handler(void);
extern void tm_interrupt_handler(void);




/*
	The interrupt handlers are called from the interrupt and interrupt pre-emption test. 
	The shadow keyword is used because it uses the shadow register set which is faster.
	The psv register is static so it will not be saved.
*/
void __attribute__ ((__interrupt__, shadow, __no_auto_psv__)) _INT0Interrupt(void) {
    _INT0IF = 0;							// Clear the interrupt 
	tm_interrupt_preemption_handler();
}
void __attribute__ ((__interrupt__, shadow, __no_auto_psv__)) _INT1Interrupt(void) {
    _INT1IF = 0;							// Clear the interrupt 
	tm_interrupt_handler();
}




/*  
	This is the central Q-Kernel Error handling. If an error occurs Q-Kernel will execute
	the function qErrNtf(). Q-Kernel functions sometimes call this function with a value
	of zero which is not a real error and in that case qNtfError() must return the 
	error itself.
*/

unsigned qErrNotify(unsigned err) {
	if (err==0) return err;
	qKrnError(err);
	return err;
}	




/* 
	This function called from main performs basic RTOS initialization, 
	calls the test initialization function, and then starts the RTOS function.  
*/
void  tm_initialize(void (*test_initialization_function)(void)) {
	/*
		The initialize enables INT0 and INT1 and set the priority at 5. Both interrupts 
		the same priority (5) so we can use the shadow register set. They have to be higher
		than 2 because Q-Kernel uses interrupt priority 1 for fibers, including the 
		scheduler, and interrupt priority 2 for the Q-Kernel timer.
		It then calls the Q-Kernel initialization and then starts Q-Kernel
	*/
	_INT0IP = 5;					// Priority 5
	_INT0IE = 1;					// Enable the interrupt
	_INT1IP = 5;					// Priority 5
	_INT1IE = 1;					// Enable the interrupt
	pVER ver;
	ver = qKrnVersion();
	printf("Q-Kernel version %i.%i-%i\n\n",(int)ver->major, (int)ver->minor, (int)ver->build);
	pERR err = qKrnInit(40000000, 256, 256, 4);
                                    // Frequency is 40Mhz (40MIPS)
                                    // IdleThreadStackSize=256
                                    // InterruptStackSize=256
                                    // FiberQueueSize=4
	if (err!=0) {
#if defined(__DEBUG)
	qBREAKPOINT;                     // activate the debugger and stop
#else
loop:
	goto loop;						// Loops, let the WDT take care of this.
#endif
	}
	test_initialization_function();
	qKrnStart();
}


/* 
	This function takes a thread ID and priority and attempts to create the
	file in the underlying RTOS.  Valid priorities range from 1 through 31, 
	where 1 is the highest priority and 31 is the lowest. If successful, 
	the function should return TM_SUCCESS. Otherwise, TM_ERROR should be returned.   
*/
int  tm_thread_create(int thread_id, int priority, void (*entry_function)(void)) {
	/*
		QKernel: Central error handling so this means always success in the main flow.
		If there is an issue the function qNtfError is executed

		A thread will be created without a name, with a 256 byte stack and the priority 
		will be reversed. Q-Kernel uses 254 for the highest priority and 1 for the lowest 
		priority.
	*/
	all_tcb[thread_id] = qThrCreateSuspended(0,(void*)(void*)entry_function,(void*)0,256,32-priority);
	if (all_tcb[thread_id] == 0) return TM_ERROR;
	return TM_SUCCESS;
}

/* 
	This function resumes the specified thread.  If successful, the function should
	return TM_SUCCESS. Otherwise, TM_ERROR should be returned.  
*/
int  tm_thread_resume(int thread_id){
	/*
		QKernel: Central error handling so this means always success in the main flow.
		If there is an issue the function qNtfError is executed.
	*/
	qThrResumeV4(all_tcb[thread_id]);
	return TM_SUCCESS;
}

/*	
	This function suspends the specified thread.  If successful, the function should
	return TM_SUCCESS. Otherwise, TM_ERROR should be returned.  
*/
int  tm_thread_suspend(int thread_id){
	/*
		QKernel: Central error handling so this means always success in the main flow.
		If there is an issue the function qNtfError is executed.
	*/
	qThrSuspendV4();
	return TM_SUCCESS;
}

/* 
	This function relinquishes to other ready threads at the same priority.  
*/
void tm_thread_relinquish(void){
	qThrYield();
}

/*
	This function suspends the specified thread for the specified number
	of seconds.  If successful, the function should return TM_SUCCESS. 
	Otherwise, TM_ERROR should be returned.  
*/
void tm_thread_sleep(int seconds) {
	qThrSleep(seconds*1000000);
}







/*
	This implementation of message processing is implemented with pipes. Our native 
	message processing mechanism is based on sending messages by reference and keeping
	track of the usage count. Special functions are available to manage the message.
	Because the TM tests are based on sending messages by value we use pipes in this 
	example.  
	
	Please don't use this as an example to send and receive messages between threads 
	and/or fibers and/ot lightweight threads but use qQueSend() and qQueReceive() 
	instaed.
*/

/*
	This function creates the specified queue.  If successful, the function should
	return TM_SUCCESS. Otherwise, TM_ERROR should be returned.  
*/
int  tm_queue_create(int queue_id){
	/*
		QKernel: Central error handling so this means always success in the main flow.
		If there is an issue the function qNtfError is executed.

		Create a pipe with no name, 16 bytes messages, and max 40 structures in the
		pipe. No readers or writes are notified. This function returns zero if there
		is no memory available for the pipe.
	*/
	all_que[queue_id] = qPipCreate(0,16,40,0,0);
	if (all_que[queue_id] == 0) return TM_ERROR;
  	return TM_SUCCESS; 
}

/*
	This function sends a 16 byte (8 ints) message to the specified queue.  If successful, 
	the function should return TM_SUCCESS. Otherwise, TM_ERROR should be returned.  
*/
int  tm_queue_send(int queue_id, unsigned long *message_ptr) {
	/*
		QKernel: Central error handling so this means always success in the main flow.
		If there is an issue the function qNtfError is executed
	*/
	qPipWriteFast(all_que[queue_id],message_ptr);
    return TM_SUCCESS;
}

/*
	This function receives a 16-byte message from the specified queue.  If successful, 
	the function should return TM_SUCCESS. Otherwise, TM_ERROR should be returned.  
*/
int  tm_queue_receive(int queue_id, unsigned long *message_ptr) {
	/*
		QKernel: Central error handling so this means always success in the main flow.
		If there is an issue the function qNtfError is executed
	*/
	qPipReadFast(all_que[queue_id],message_ptr);
	return TM_SUCCESS;
}








/*	
	This function creates the specified semaphore.  If successful, the function should
	return TM_SUCCESS. Otherwise, TM_ERROR should be returned.  
*/
int  tm_semaphore_create(int semaphore_id) {
	/*
		QKernel: Central error handling so this means always success in the main flow.
		If there is an issue the function qNtfError is executed. This function returns 
		zero if there is no memory available for the semaphore.
	*/
	all_sem[semaphore_id] = qSemCreate(0,100);
	if (all_sem[semaphore_id] == 0) return TM_ERROR;
	return TM_SUCCESS;
}

/*
	This function gets the specified semaphore.  If successful, the function should
	return TM_SUCCESS. Otherwise, TM_ERROR should be returned.  
*/
int  tm_semaphore_get(int semaphore_id) {
	/*
		QKernel: Central error handling so this means always success in the main flow.
		If there is an issue the function qNtfError is executed
	*/
	qSemAcquireFast(all_sem[semaphore_id]);
	return TM_SUCCESS;
}


/*	
	This function puts the specified semaphore.  If successful, the function should
	return TM_SUCCESS. Otherwise, TM_ERROR should be returned.  
*/
int  tm_semaphore_put(int semaphore_id) {
	/*
		QKernel: Central error handling so this means always success in the main flow.
		If there is an issue the function qNtfError is executed
	*/
	qSemReleaseFast(all_sem[semaphore_id]);
	return TM_SUCCESS;
	
}






/*
	Fixed memory pools contain fixed memory blocks. Q-Kernel also supports Dynamic 
	memory blocks (variable size) and "allocate-only" heap memory but they a slower 
	(dynamic) or don't provide the required functionality ("allocate-only"). 
	
	The function qFixCreate() reserves the data structure for fixed memory blocks with a 
	specific size and returns a pointer to that structure. If a fixed memory block is 
	freed-up it will be returned to the list.
	
	The creation of fixed memory blocks can return zero indicating that there is not
	enough memory available for the memory control structures. This means that we have 
	implemented the TM_ERROR/TM_SUCCESS return values.
*/

/*
	This function creates the specified memory pool that can support one or more
	allocations of 128 bytes.  If successful, the function should
	return TM_SUCCESS. Otherwise, TM_ERROR should be returned.  
*/
int  tm_memory_pool_create(int pool_id) {
#if VARIABLE_MEMORY==1
	all_mem[pool_id] = qMemPool(10);
	if (all_mem[pool_id]) {
		qMemPoolAdd(all_mem[pool_id],10);
		return TM_SUCCESS;
		}
	return TM_ERROR;	
#else
	all_mem[pool_id] = qFixCreate(10,128);
	if (all_mem[pool_id]) return TM_SUCCESS;
	return TM_ERROR;
#endif
}

/*
	This function allocates a 128 byte block from the specified memory pool.  
	If successful, the function should return TM_SUCCESS. Otherwise, TM_ERROR 
	should be returned.  
*/
int  tm_memory_pool_allocate(int pool_id, unsigned char **memory_ptr) {
	/*
		QKernel: Central error handling so this means always success in the main flow.
		If there is an issue the function qNtfError is executed
	*/
#if VARIABLE_MEMORY==1
	*memory_ptr = qMemAllocFromPoolFast(all_mem[pool_id]);
	return TM_SUCCESS;
#else
	*memory_ptr = qFixAlloc(all_mem[pool_id]);
	return TM_SUCCESS;
#endif
}

/*	
	This function releases a previously allocated 128 byte block from the specified 
	memory pool. If successful, the function should return TM_SUCCESS. Otherwise, TM_ERROR 
	should be returned.  
*/
int  tm_memory_pool_deallocate(int pool_id, unsigned char *memory_ptr) {
	/*
		QKernel: Central error handling so this means always success in the main flow.
		If there is an issue the function qNtfError is executed.
	*/
#if VARIABLE_MEMORY==1
	qMemFreeFast(memory_ptr);
	return TM_SUCCESS;
#else
	qFixFree(memory_ptr);
	return TM_SUCCESS;
#endif
}






int main_cooperative_scheduling();
int main_preemptive_scheduling();
int main_interrupt_processing();
int main_interrupt_preemption_processing();
int main_message_processing();
int main_synchronization_processing();
int main_memory_allocation();
int main_basic_processing();

int main() {
	volatile int test = 5;
	if (test==1) main_cooperative_scheduling();
	if (test==2) main_preemptive_scheduling();
	if (test==3) main_interrupt_processing();
	if (test==4) main_interrupt_preemption_processing();
	if (test==5) main_message_processing();
	if (test==6) main_synchronization_processing();
	if (test==7) main_memory_allocation();
	return 0; 
}

/* Before starting testing please check the following:
 *  -   Use the kernel configuration "ThreadMetric". This configuration is optimized
 *      for speed. 
 *  -   The ThreadMetric project should set the optimize level to 3.
 *  -   This project requires the XC16 compiler
 *
 *  -   The results are (with XC16 version 1.20):
 *       1) cooperative_scheduling           17,141,251
 *       2) preemptive_scheduling            11,741,503
 *       3) interrupt_processing             18,461,234
 *       4) interrupt_preemption_processing   6,315,823
 *       5) message_processing               10,908,076
 *       6) synchronization_processing       32,431,922
 *       7) memory_allocation                22,221,851
 */
