/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"

#if defined(__HAS_EDS__)
    unsigned* qvSharedStack = 0;    // Working stack for all EDS stacks
    pTCB qvSharedStackOwner = 0;    // The owner of the shared stack
    unsigned qvSharedStackSize=0;   // The size of the shared stack
#endif





void yKernelInit() {
    IPC6bits.T4IP = 1;                          // Interrupt priority = 1
}



static void xThrCloseCurrent() {
    qThrClose(qThrCurrent());                  //
}

extern int _STACKGUARD_SIZE;
void yThrStkInit(void (*task)(void*), void* pData, pTCB Tcb, unsigned* Stack, unsigned StackSize) {
    unsigned lim;
    unsigned* tos = Stack;

    lim = (unsigned)Stack + StackSize - (unsigned)&_STACKGUARD_SIZE;
    Tcb->SPLIM = lim;
    
    *tos++ = (uint16_t)task;            // Address of thread on stack so we can find it
                                        // with the RTOS-Viewer
    *tos++ = (uint16_t)xThrCloseCurrent;// this will kill the process
    *tos++ = 0;

// Simulate an interrupt
    *tos++ = (uint16_t)task;            // Address of the thread 
    *tos++ = 0;                         // PC (16-23) always 0 because C30 makes jumps
    Tcb->Registers[15] = (unsigned)tos;

#ifdef __HAS_EDS__
    Tcb->DSRPSVPAG = DSRPAG;                    //
    Tcb->DSWPAG = DSWPAG;                       //
    xKrnSetEdsSupported();                      // Indicate that we are EDS aware
#else
    Tcb->DSRPSVPAG = PSVPAG;
#endif
    Tcb->TBLPAG = TBLPAG;        //
    Tcb->RCOUNT = RCOUNT;

    Tcb->Registers[0] = (uint16_t)pData;
#ifndef __NO_CHECK__
    Tcb->Registers[1] = 0x1111;
    Tcb->Registers[2] = 0x2222;
    Tcb->Registers[3] = 0x3333;
    Tcb->Registers[4] = 0x4444;
    Tcb->Registers[5] = 0x5555;
    Tcb->Registers[6] = 0x6666;
    Tcb->Registers[7] = 0x7777;
    Tcb->Registers[8] = 0x8888;
    Tcb->Registers[9] = 0x9999;
    Tcb->Registers[10] = 0xAAAA;
    Tcb->Registers[11] = 0xBBBB;
    Tcb->Registers[12] = 0xCCCC;
    Tcb->Registers[13] = 0xDDDD;
    Tcb->Registers[14] = 0xEEEE;
    Tcb->RCOUNT = 0xFFFF;
#endif
}





// +----------------------------+ 0x0800
// |                            |
// | Static memory              |
// |                            |
// +----------------------------+ 0x095e  
// |                            |
// | C30 Heap (0x0400)          |
// |                            |
// +----------------------------+ 0x0d5e  <--- qvIntStack  (Stack grows up)
// |                            |
// | Interrupt stack  (0x100)   |
// |                            |
// +----------------------------+ 0x0e5e  <--- qvIntStack  (Stack grows up)
// | Temporary extension        |
// +----------------------------+ 0x0f12  <--- qvMemHeap (Stack before qKrnStart)
// |                            |
// |                            |
// | Heap Size=0x18ee           |
// |                            |
// |                            |
// |                            |
// |                            |
// |                            |
// |                            |
// |                            |
// |                            |
// +----------------------------+ 0x2800   last
//
//
// This routine initializes the memory. It must set the following variables:
// qvMemHeap         The pointer points to the start of the heap and the function must initialize the 
//                    size in the first integer. The heap must be cleared
// qvIntStack        The pointer points to the stack. The stack must be initialized with 0xff
//
// The system uses the interrupt stack between calling qKrnInit() and qKrnStart().
// During this period the stack requirements may be higher than the interrupt stack
// can supply. For that reason the interrupt stack is created with a 96 bytes
// extension. The 96 bytes is enough space for 6 OBJ struct.
// Because the SP is currently in the space of the interrupt stack, the init of the 
// interrupt stack with 0xff is done in yStart. This gives the system time to work.
//
extern int _STACKGUARD_SIZE;
extern unsigned _SP_init;
extern unsigned _SPLIM_init;

unsigned yMemSetup(unsigned InterruptStackSize) {
    unsigned heapSize;
    
    unsigned last = (unsigned)&_SPLIM_init;
    last += (unsigned)&_STACKGUARD_SIZE;
    if (last>0x8000) {
        last=0x8000;            // Prevent memory space in upper 32k
    }
    qvIntStack = &_SP_init;
    InterruptStackSize += (sizeof(OBJ)+sizeof(int))*8;
    qvMemHeap = qvIntStack + (InterruptStackSize>>1);

    heapSize = last - (unsigned)qvMemHeap;
    qBytClr(qvMemHeap, heapSize);
    *qvMemHeap = heapSize;

    qvIntStackLimit = qvIntStack + ((InterruptStackSize - TCB_STACK_GUARD_SIZE)>>1);
    SPLIM = (unsigned)qvIntStackLimit;
    return heapSize;
}

extern void yStart1();
void yStart() {
    pMBK pMbk;
    int n;
/* The interrupt stack was created with a size to get two TCB's from its end.
 * So we first free two TCB's back into its memory pool
 */
    qvIntStackLimit += (TCB_STACK_GUARD_SIZE>>1);
    for (n=0; n<8; n++){
        qvIntStackLimit = qvIntStackLimit - ((sizeof(struct sOBJ)+sizeof(int))>>1);
        pMbk = (pMBK)qvIntStackLimit;           //
        pMbk->pMpl = qvMemObjPool;
        qMemFree(&(pMbk->Memory));
    }
    qvIntStackLimit -= (TCB_STACK_GUARD_SIZE>>1);
    yStart1();
}

pERR yErrorInit() {
    if ((RCON&0x3) == 0 && (RCON&1<<6)) {   // POR and BOR must be both 0 and software reset ...
        return &qvErr;                      // ... set for valid data
    }
    RCON = 0;                               // Clear all
    qWrdClr((unsigned*)&qvErr, sizeof(qvErr));
    return NULL;
}

void yErrorReset(unsigned Err) {
    unsigned* sp;
    qvErr.ErrNbr = Err;
    qvErr.Uptime = qRtcGetUptime();
    qvErr.pThread = qvCurTcb;
    qvErr.SPLIM = SPLIM;
    qvErr.StackType = 0;
    __asm__ __volatile__("mov WREG15, %0 \n" : "=r"(sp));
    if (sp<qvIntStackLimit) {
        qvErr.StackType = 1;
    }
    unsigned zone = (SR&0b11100000);
    zone = zone >>5;
    if (CORCON & (1<<3)) {
        zone +=8;
    }
    qvErr.Zone = zone;
    qvErr.InCritical = qvInCritical;
     // Trap address filled in by the trap function
    __asm__ __volatile__ ("reset");

}

