/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"

#if defined(__HAS_EDS__)
unsigned yEdsPage(void* __attribute__((space(eds))) Page);     // in ySupport.s
unsigned yEdsOffset(void* __attribute__((space(eds))) Offset); // in ySupport.s

static void qqqThrCloseCurrent() {
    qThrClose(qvCurTcb);    
}
extern int _STACKGUARD_SIZE;

void yThrStkInitEds(void (*task)(void*), void* pData, pTCB Tcb, unsigned int* __attribute__((space(eds))) Stack, unsigned StackSize) {
    unsigned lim;
    __eds__ unsigned int* tos = Stack;
    
    lim = (unsigned)qvSharedStack + StackSize - (unsigned)&_STACKGUARD_SIZE;
    Tcb->SPLIM = lim;

    *tos++ = (uint16_t)task;                // Addressof thread on stack so we can find it with the RTOS-Viewer
    *tos++ = (uint16_t)qqqThrCloseCurrent;  // this will kill the process 
    *tos++ = 0;

// Simulate an interrupt
    *tos++ = (uint16_t)task;                // Address of the thread
    *tos++ = 0;                             // PC (16-23) always 0 because C30 makes jumps

    
    Tcb->DSRPSVPAG = DSRPAG;                //
    Tcb->DSWPAG = DSWPAG;                   //
    Tcb->TBLPAG = TBLPAG;                   //
    Tcb->RCOUNT = RCOUNT;
    Tcb->Registers[0] = (uint16_t)pData;
    Tcb->Registers[15] = (unsigned)qvSharedStack + 10;
}

// ******************************************************************************************************
pERR qEdsKrnInit(uint32_t ClockFrequency, unsigned IdleThreadStackSize, unsigned InterruptStackSize, unsigned FiberQueueSize, unsigned SharedStackSize) {
    pERR err = qKrnInit(ClockFrequency, IdleThreadStackSize, InterruptStackSize, FiberQueueSize);
    if (err==0)
        qvSharedStack = qHeaAlloc(SharedStackSize);
        if (qvSharedStack == 0) {
            xErrNtf(qERR_KRN_MEM_SIZE);
            return err;
        }
    return err;
}
static pTCB xThrCreateEds(char* pName, void (*Thread)(void*), void* pParam, unsigned StackSize, uint8_t Prio, unsigned* __attribute__((space(eds))) Stack, uint8_t Suspended) {
    unsigned nbrThread;
    pTCB p;
    pTCB tcb;
    
    #ifndef __NO_CHECK__
        unsigned prio = Prio;
        if (xKrnIsPriorityCheckOn() && (prio==0 || prio>250)) {
            xErrNtf(qERR_THR_PRIO);
            return NULL;
        }
        if (qKrnInIsr()) {
            xErrNtf(qERR_THR_ISR);
            return NULL;
        }
        if (StackSize > qvSharedStackSize) {
            xErrNtf(qERR_THR_SHARED_STACK_SIZE);
            return NULL;
        }
        if (StackSize <= 32) {
            xErrNtf(qERR_THR_STACK);
            return NULL;
        }
        if (pName!=0) {
            if (xGetTcb(pName)) {                    // Search if there is someting with the name
                xErrNtf(qERR_THR_NAME_IN_USE);
                return NULL;
            }
        }
        if (!xKrnIsEdsSupported()) {
            xErrNtf(qERR_THR_NO_EDS);
            return NULL;
        }
    #endif
    xKrnClrPriorityCheckOn();                        // Checking on so next time we create a thread we don't check
                                                     // The Idle thread is the first thread that will be created
    qCrtEnter();
    tcb = (pTCB)qMemAllocFromPoolClr((pMPL)qvMemTcbPool);
    #ifndef __NO_CHECK__
        if (tcb==0) {
            qCrtExit();
            xErrNtf(qERR_THR_MEMORY);
            return NULL;
        }
    #endif

    nbrThread=0;
    if (xKrnIsInitDone()) {                         // Do we have an idle thread?
        p = qvIdleTcb;                              // Start with the idle thread
        while (p->pTcbList) {                       // All Tcb List is done
            nbrThread++;                            // add to the number of threads
            p = p->pTcbList;                        // next one
        }
        p->pTcbList = tcb;                          // add us to the end of the list
    }
        
    tcb->pStack = (unsigned*)yEdsOffset(Stack);     // The stack here gives the offset in the EDS space
    tcb->StackEdsPage = yEdsPage(Stack);            // And the page indicates that

    yThrStkInitEds(Thread, pParam, tcb, Stack, StackSize); // Initialize the thread's stack
    tcb->Prio = Prio;                               // Load thread priority into TCB
    tcb->StackSize = StackSize;                     // Store stack size
    tcb->pName = pName;
    tcb->Type = xTYPE_TCB;
    // Set tracing to its own bit. IOR will do nothing
    
    if (!Suspended) {
        if (xInsRdyList(tcb)) {                     // Insert the new TCB. Is it higher?
            ySwitchThread();                        // Switch Thread and EXIT_CRITICAL
        }                                           // only if readied thread > prio
        else {
            qCrtExit();
        }
    }
    else {                                          // Tcb->ListTYpe = qLIST_NONE because TCB is zeroed
        qCrtExit();
    }
    xObjRemOpenList(pName, xTYPE_OPN_TCB);   // Make waiting Open threads runnable
    return (pTCB)tcb;
}

//// ******************************************************************************************************

pTCB qThrCreateEds(char* pName, void (*Thread)(void*), void* pParam, unsigned StackSize, uint8_t Prio, unsigned* __attribute__((space(eds))) Stack) {
    return xThrCreateEds(pName, Thread, pParam, StackSize, Prio, Stack, 0);
}

// ******************************************************************************************************

pTCB qThrCreateSuspendEds(char* pName, void (*Thread)(void*), void* pParam, unsigned StackSize, uint8_t Prio, unsigned* __attribute__((space(eds))) Stack) {
    return xThrCreateEds(pName, Thread, pParam, StackSize, Prio, Stack, 1);
}

#endif //__HAS_EDS__
