/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

.ifdef __dsPIC30F
    .include "p30Fxxxx.inc"
.endif
.ifdef __dsPIC33E__
    .include "p33Exxxx.inc"
.endif
.ifdef __dsPIC33F
    .include "p33Fxxxx.inc"
.endif
.ifdef __PIC24E
    .include "p24Exxxx.inc"
.endif
.ifdef __PIC24F
    .include "p24Fxxxx.inc"
.endif
.ifdef __PIC24FK
    .include "p24Fxxxx.inc"
.endif
.ifdef __PIC24H
    .include "p24Hxxxx.inc"
.endif
.ifdef __PIC32MX
    .include "p32xxxx.inc"
.endif
.macro SET_KERNEL_INTERRUPT
    bset IFS1, #11
.endm
.macro CLR_KERNEL_INTERRUPT
    bclr IFS1, #11
.endm
.macro TST_KERNEL_INTERRUPT
    btst IFS1, #11
.endm
.macro qCrtEnter
    bclr IEC1, #11
    inc        _qvInCritical
.endm
.macro qCrtExit
    dec    _qvInCritical
    bset IEC1, #11
.endm
.macro DISABLE_KERNEL_INTERRUPT
    bclr IEC1, #11
.endm
.macro ENABLE_KERNEL_INTERRUPT
    bset IEC1, #11
.endm

.section qKernel,code
;//         +------------------------+
;//     +-- |  pFirstBlk             |
;//     |   |  NextPool              |
;//     |   |  Type/Filler           |
;//     |   |  BlockSize (10)        |
;//     |   |  Entries (4)           |
;//     |   |  BitTable (1 integer)  | <-- This is variable 1 int for 16 2 int for 32 etc 
;//     |   +------------------------+
;//     +-> |  Entry number          | <--- pBlk points here
;//         |  InUse + pnt BitTable  |
;//         |  Memory                | <--- Pointer given back 
;//         |  Memory                |
;//         |  Memory                |
;//         |  Memory                |
;//         |  Memory                |
;//         +------------------------+
;//         |  Entry number          |      Entry (bitNbr 0->15) 
;//         |  InUse + pnt BitTable  |      Pnt the the bittable word for this entry
;//         |  Memory                | <--- Pointer given back
;//         |  Memory                |
;//         |  Memory                |
;//         |  Memory                |
;//         |  Memory                |
;//         +------------------------+
;//         |  Entry number          |
;//         |  InUse + pnt BitTable  |
;//         |  Memory                | <--- Pointer given back 
;//         |  Memory                |
;//         |  Memory                |
;//         |  Memory                |
;//         |  Memory                |
;//         +------------------------+
;//         |  Entry number          |
;//         |  InUse + pnt BitTable  |
;//         |  Memory                | <--- Pointer given back
;//         |  Memory                |
;//         |  Memory                |
;//         |  Memory                |
;//         |  Memory                |
;//         +------------------------+
;//

.set _FPL_pFirstBlk, 0
.set _FPL_BlockSize, 6        ; Size of memory including EntryNumber + InUse/*BitTable
.set _FPL_Total, 8
.set _FPL_BitTable, 10        ; Bit table 1=free 0=inuse

    .global _qFixAlloc
_qFixAlloc: 
;
; This is allocating from interruptable pool
;    
; w0 = pointer to pFPL (Input)
; w7 = pointer (variable) in bit-table
; w6 = bit number free (0-15)
; w5 = Fixed (sort of) pointer in Bit-Table
; w4 = entry number
    
50: add     w0, #_FPL_BitTable-2, w7    ; pointer in bit-table Start before
    add     w0, #_FPL_BitTable, w5      ; pointer to bit Table
60: ff1r    [++w7], w6
    bra     c, 60b
    dec     w6, w6                      ; w6 1-16 must be from 0-15
    sub     w7, w5, w4                  ; w4 = offset in bit table (w7-w5) 0-2-4-6 etc
    sl      w4, #3, w4                  ; now *8 *2 = *16
    add     w6, w4 ,w4                  ; w4 = entry number 
    cp      w4, [--w5]                  ; pos - number_entries (W5 not on Bit-Table)
    bra     ge, 99f                     ; no room
    mov     [w0+_FPL_BlockSize], w1     ; Size of memory
    add     #4, w1                      ; Allocate 4 bytes more for Entry# and Inuse
    mul.uu  w4, w1, w2                  ; entry number * size in w2-w3
    add     w2, [w0], w1                ; w1 now points to the memory for us [w0]=pBlk
                                        ; (this will clear the carry flag)
    bsw.c   [w7], w6                    ; clear the bit (insert carry flag)
    btsts   [++w1], #0                  ; is memory it in use?
    bra     nz, 50b                     ; try again 
    add     w7, #1, [w1]                ; set address of memory block
    mov     w6, [--w1]                  ; set the bit position 
    add     w1, #4, w0                  ; pointer to real usable memory
    return
99:
    retlw    #0, w0

; w0 = pojnter to memory and later pointer in bit table
; w1 = offset in bit-table
; w2 = position in bit table
    .global _qFixFree
_qFixFree: 
    dec     [--w0], w1                  ; w1 pnt to bittable (will clear the zero flag)
    bclr    [w0], #0                    ; clear the in use bit in memory
    mov     [--w0], w2                  ; BitNumber in bit table in w2
    bsw.z   [w1], w2                    ; this sets the bit in the bit-table (complement of zero flag)
    return


