/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

.ifdef __dsPIC30F
    .include "p30Fxxxx.inc"
.endif
.ifdef __dsPIC33E__
    .include "p33Exxxx.inc"
.endif
.ifdef __dsPIC33F
    .include "p33Fxxxx.inc"
.endif
.ifdef __PIC24E
    .include "p24Exxxx.inc"
.endif
.ifdef __PIC24F
    .include "p24Fxxxx.inc"
.endif
.ifdef __PIC24FK
    .include "p24Fxxxx.inc"
.endif
.ifdef __PIC24H
    .include "p24Hxxxx.inc"
.endif
.ifdef __PIC32MX
    .include "p32xxxx.inc"
.endif
.macro SET_KERNEL_INTERRUPT
    bset IFS1, #11
.endm
.macro CLR_KERNEL_INTERRUPT
    bclr IFS1, #11
.endm
.macro TST_KERNEL_INTERRUPT
    btst IFS1, #11
.endm
.macro qCrtEnter
    bclr IEC1, #11
    inc        _qvInCritical
.endm
.macro qCrtExit
    dec    _qvInCritical
    bset IEC1, #11
.endm
.macro DISABLE_KERNEL_INTERRUPT
    bclr IEC1, #11
.endm
.macro ENABLE_KERNEL_INTERRUPT
    bset IEC1, #11
.endm

.section *,data,near  
_qvSaveW0:        .word 0 





_xInitDone = 0
_xOsStarted = 1
_xStatisticsOn = 2
_xSwitchNotificationOn = 3
_xUSecOn = 4
_xPriorityCheckOn = 5
_xTimerInterruptOn = 6
_xRtccInterruptOn = 7


; TCB
.set __TrackBit,          17
.set __TblPag,            18
.set __pStack,            26
.set __StackSize,         28
.set __DsrPag,            30
.set __PsvPag,            30
.set __DswPag,            32
.set __rcount,            34
.set __w0,                36
.set __w1,                38
.set __w2,                40
.set __w3,                42
.set __w4,                44
.set __w5,                46
.set __w6,                48
.set __w7,                50
.set __w8,                52
.set __w9,                54
.set __w10,               56
.set __w11,               58
.set __w12,               60
.set __w13,               62
.set __w14,               64
.set __sp,                66
.set __splim,             68
.set __StackEdsPage,      76
.set __TrackAdr,          78
.set __Cycles,            80
.set __LastCycles,        88    ; LastCycles must follwow Cycles


; ********************************************************* EDS Stack ***************************************************
; All threads that used EDS memory as their stack use the shared stack if they are active. This stack is created during 
; initialization of Q-Kernel and the variable qvSharedStack points to it. Which thread (the owner) is stored in this 
; memory is specified by qvSharedStackOwner. (qvSharedStack resides in the first 32kb)
; Q-Kernel will save the content of the EDS Stack in EDS memory if the ownership of the EDS stack changes.
; To minimize the moving of data and to guarantee fast switches to threads with standard stacks the following algorithm:
;
; Std Tcb --> Std Tcb    Nothing needs to be done
; Eds Tcb --> Std Tcb    Nothing needs to be done
; Std Tcb --> Eds Tcb    If the owner of the EDS stack is the same as the Eds Tcb nothing has to be done. But if the 
;                        owner is different the EDS stack needs to be saved and the new one need to be restored. 
; Eds Tcb --> Eds Tcb    Save the EDS stack and restore the stack for the new Tcb
;
; This way we guarantee that switching to a thread with a standard stack always is as fast as possible.
;
; Context switch:                Fully optimized no EDS stacks:                    40 cylces
;                                Std Stack Tcb --> Std Stack Tcb                   46 cycles
;                                Std Stack Tcb --> Eds Stack Tcb    (cached)       46 cycles
;                                Std Stack Tcb --> Eds Stack Tcb    (non-cached)   63 cycles + (StkSizeFrm + StkSizeTo)/2
;                                Esd Stack Tcb --> Std Stack Tcb                   46 cycles
;                                Esd Stack Tcb --> Eds Stack Tcb                   63 cycles + (StkSizeFrm + StkSizeTo)/2
;
; Preemptive Context switch:    Fully optimized no EDS stacks:                     60 cylces
;                                Std Stack Tcb --> Std Stack Tcb                   66 cycles
;                                Std Stack Tcb --> Eds Stack Tcb    (cached)       66 cycles
;                                Std Stack Tcb --> Eds Stack Tcb    (non-cached)   83 cycles + (StkSizeFrm + StkSizeTo)/2
;                                Esd Stack Tcb --> Std Stack Tcb                   66 cycles
;                                Esd Stack Tcb --> Eds Stack Tcb                   83 cycles + (StkSizeFrm + StkSizeTo)/2
;
; ***********************************************************************************************************************

.section qKernel,code


;
;  **** yStart1()
;
    .global  _yStart1
_yStart1:
    setm    SPLIM
    mov     _qvIntStack,w0
    mov     _qvIntStackLimit,w1
    add     w1,#16,w1                   ; also make the guard 0xff (TCB_STACK_GUARD_SIZE=16)
10: setm    [w0++]
    sub     w0,w1,w2
    bra     nz, 10b
    bra     _RestoreTcb                 ; Start the thread by restoring its TCB
;
;
;
; This routine does a context switch.
; Because this is a call (no interrupt we don't have to save w0 to w7, SR, RCOUNT, TBLPAG (
; and PSVPAG)  On the way back we have to restore everything because the new thread could 
; have been preempted by an interrupt. 
; This function can not be called from an interrupt. That must be tested for functions that 
; can be called from an interrupt.
;
; If there is still a critical section going or this call is called from the scheduler then we
; just return. We always do a EXIT_CRITICAL. 
;
;
    .global  _ySwitchThread
_ySwitchThread:
    cp0     _qvInCritical               ; Critical level must be level 1 (value 0)
    bra     nz, 99f                     ; If not in level 1 (our own critical section) return with exit critical
;
; Compiler can't rely on TBLPAG/DSRPAG, PSVPAG/DSWPAG and RCOUNT so they don't have to be saved
; We only have to save the callee saveble register (W8-W15)
;
    mov     _qvCurTcb, w0
    mov     w8, [w0+__w8]
    mov     w9, [w0+__w9]
    mov     w10, [w0+__w10]
    mov     w11, [w0+__w11]
    mov     w12, [w0+__w12]
    mov     w13, [w0+__w13]
    mov     w14, [w0+__w14]
    mov     w15, [w0+__sp]

.ifndef __FAST__
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    btst    _qvKernel,#_xStatisticsOn   ; statistics ON?
    bra     z, 17f
;
; Add cycles for current thread and subtract them from the next thread
    rcall   _yGetCycles                 ; returns current cycles in W0-3
    mov     _qvRdyList, w7              ; Context of tcb to become active
    add     #__LastCycles, w7           ; pointer to the last cycle counter
    mov     w0, [w7+0]
    mov     w1, [w7+2]
    mov     w2, [w7+4]
    mov     w3, [w7+6]

    mov     _qvCurTcb, w7
    add     #__Cycles, w7               ; pointer to the total number of cycles
    mov     w7, w6                      ; w6=pointer to Cycles
    add     w0, [w7++], w8              ; tmp = Cycles + Now()
    addc    w1, [w7++], w9              ; tmp = Cycles + Now()
    addc    w2, [w7++], w10             ; tmp = Cycles + Now()
    addc    w3, [w7++], w11             ; tmp = Cycles + Now()
    sub     w8, [w7++], [w6++]          ; Cycles = cycles + Now() - last
    subb    w9, [w7++], [w6++]          ; Cycles = cycles + Now() - last
    subb    w10, [w7++], [w6++]         ; Cycles = cycles + Now() - last
    subb    w11, [w7++], [w6++]         ; Cycles = cycles + Now() - last

    mov     _qvCurTcb, w0               ; Restore the current TCB
17:
    mov     _qvRdyList, w1              ; Next one required for qKrnNtfSwitch()
    btsc    _qvKernel,#_xSwitchNotificationOn  ; SwitchNotification
    call    _qKrnNtfSwitch              ; qKrnNtfSwitch(CurTcb,NextTcb)
    mov     [w0+__TrackAdr], w1         ; Tracking address
    mov.b   [w0+__TrackBit], w2         ; Tracking bit number
    ior.b   w2, [w1], [w1]              ; Do it
.endif

_RestoreTcb:
    mov     _qvRdyList, w0              ; load the Tcb on ready list into W0
    mov     w0, _qvCurTcb               ; Now rhe Tcb in RdyList as current Tcb
.ifndef __FAST__
    mov     [w0+__TrackAdr], w1         ; Tracking address
    mov.b   [w0+__TrackBit], w2         ; Tracking bit number
    ior.b   w2, [w1], [w1]              ; Do it
  .ifdef __HAS_EDS__
    mov     [w0+__StackEdsPage], w1     ; EDS page in w1
    cp0     w1                          ; Does the new thread has a EDS stack?
    bra     z, 20f
;
; Here if the new thread uses a EDS stack
    cp0     _qvSharedStackOwner         ; Is there a shhared stack owner
    bra     z, 18f
    cp      _qvSharedStackOwner         ; Do we own the shared Stack (cp _qvSharedStackOwner with w0)
    bra     z, 20f                      ; yes so we are done
;
; We have to save the shared stack in its EDS place 
    mov     _qvSharedStackOwner, w6
    mov     [w6+__StackEdsPage], w1
    mov     w1, 0x0034                  ; To page DSWPAG (we can use this because it will be restored later)
    mov     [w6+__StackSize], w1        ; number of bytes to copy
    asr     w1, w1                      ; number of words to copy
    dec     w1, w1                      ; compensate for the repeat instruction
    mov     [w6+__pStack], w4           ; To offset (bit 15 already set)
    mov     _qvSharedStack, w3          ; from address
    repeat  w1
    mov     [w3++], [w4++]
;
; Now restore the new thread
18: mov     [w0+__StackEdsPage], w1
    mov     w1, 0x0032                  ; To page DSRPAG (we can use this because it will be restored later)
    mov     [w0+__StackSize], w1        ; number of bytes to copy
    asr     w1, w1                      ; number of words to copy
    dec     w1, w1                      ; compensate for the repeat instruction
    mov     [w0+__pStack], w3           ; From offset (bit 15 already set)
    mov     _qvSharedStack, w4          ; To address
    repeat  w1
    mov     [w3++], [w4++]
    mov     w0, _qvSharedStackOwner     ; indicate we own the EDS stack
  .endif
.endif

20:
.ifndef __FAST__
  .ifdef __HAS_EDS
    mov     [w0+__DsrPag], w1           ; DSRPAG
    mov     w1, DSRPAG                  ; DSRPAG
    mov     [w0+__DswPag], w1           ; DSWPAG
    mov     w1, DSWPAG                  ; DSWPAG
  .else
    mov     [w0+__PsvPag], w1           ; PSVPAG
    mov     w1, PSVPAG                  ; PSVPAG
  .endif
    mov     [w0+__TblPag], w1           ; TBLPAG
    mov     w1, TBLPAG                  ; TBLPAG
.endif
    mov     [w0+__rcount], w1           ; RCOUNT
    mov     w1,RCOUNT                   ; RCOUNT

    mov     [w0+__w14], w14
    mov     [w0+__w13], w13
    mov     [w0+__w12], w12
    mov     [w0+__w11], w11
    mov     [w0+__w10], w10
    mov     [w0+__w9], w9
    mov     [w0+__w8], w8
    mov     [w0+__w7], w7
    mov     [w0+__w6], w6
    mov     [w0+__w5], w5
    mov     [w0+__w4], w4
    mov     [w0+__w3], w3
    mov     [w0+__w2], w2

; SPLIM and SP
.ifndef __FAST__
    setm    SPLIM                       ;SPLIM Set so there is no problem during switching SP
    mov     [w0+__splim],w1             ;SPLIM now in w1
    mov     [w0+__sp],w15               ;SP  Set the stack pointer
    mov     w1,SPLIM                    ;SPLIM (We switch to another task so the SPLIM is not predictable
.else
    mov     [w0+__sp],w15               ;SP  Set the stack pointer
.endif

; last registers    
    mov     [w0+__w1],w1
    mov     [w0+__w0],w0
; Still have to test for the Kernel interrupt (2 part of EXIT_CRITICAL)
    qCrtExit                            ;
    retfie                              ; Set the interrupt-level back 
; 
; Return because we were in a critical
;
99:    
    SET_KERNEL_INTERRUPT                ; Kernel interrupt will switch the thread.
    dec     _qvInCritical               ; Exit Critical but don't enable kernel interrupts  ...
    return                              ; ... because the previous level will do that






; This is the interrupt service routine (ISR) for the cheduler. The scheduler runs
; at interrupt level 1. It will be activated by the Kernel-Interrupt.
; The ISR start a critical section. We do this so other functions know that they
; can only use non-blocking functions. 
; We can come in here while still in a critical section. That can only be triggred if previous code
; is in a critical section (-1->0), goes in another critical section (0->1) and then exit the 
; critical section. That exit will enable the kernel interrupt and  we come in here with 
; qvInCritical = 0 means we are in a critical. In that case we have to disable the 
; kernel interrupt on exit. (label 499) 
;
.global __T4Interrupt
__T4Interrupt:
; We will clear the interrupt because the code that set the critical section will 
; set the interrupt when it leaeves the critical section 
    inc     _qvInCritical               ; Increment critical section. Must go from -1 to 0            
    bra     nz, 499f                    ; In critical section skip the scheduler.
    CLR_KERNEL_INTERRUPT                ; Clear the kernel interrupt (macro)
; 
;    Here save the w0, w1 and SP
;
    mov     w0, _qvSaveW0
    mov     _qvCurTcb, w0
    mov     w15, [w0+__sp]              ; save SP
    mov     _qvIntStack, w15            ; Stack always switched. SPLIM is not a problem because new limit...
                                        ; ...is always smaller then the interrupted one.
    mov     w1,[w0+__w1]                ; save w1
    mov     _qvSaveW0, w1               ; old w0
    mov     w1,[w0+__w0]                ; save w0 (stored in w1)

.ifndef __FAST__
    mov     _qvIntStackLimit,w1
    mov     w1,SPLIM                    ; New stack limit This new SPLIM is always smaller then the interrupted one
.endif

    mov     w2, [w0+__w2]               ; save w2
    mov     w3, [w0+__w3]
    mov     w4, [w0+__w4]
    mov     w5, [w0+__w5]
    mov     w6, [w0+__w6]
    mov     w7, [w0+__w7]

.ifndef __FAST__
    mov     [w0+__TrackAdr], w1         ; Tracking address
    mov.b   [w0+__TrackBit], w2         ; Tracking bit number
    ior.b   w2, [w1], [w1]              ; Do it
    mov     _qvFbrTrackAdr, w1          ; Fiber tracking address
    mov     _qvFbrTrackBit, w2
    ior     w2, [w1], [w1]              ; Do it
  .ifdef __HAS_EDS
    mov     DSRPAG, w2                  ; DSRPAG
    mov     w2, [w0+__DsrPag]           ; DSRPAG
    mov     DSWPAG, w2                  ; DSWPAG
    mov     w2, [w0+__DswPag]           ; DSWPAG
  .else
    mov     PSVPAG, w2                  ; PSVPAG
    mov     w2, [w0+__PsvPag]           ; PSVPAG
  .endif
    mov     TBLPAG, w2                  ; TBLPAG
    mov     w2, [w0+__TblPag]           ; TBLPAG
.endif
    mov     RCOUNT, w2                  ; RCOUNT
    mov     w2, [w0+__rcount]           ; RCOUNT

.ifndef __FAST__
    btst    _qvKernel,#_xStatisticsOn   ; statistics ON?
    bra     z, 18f
;
; Store the cycles we have used so far
;
    push    w8
    rcall   _yGetCycles                 ; returns current cycles in W0-3
    mov     _qvCurTcb, w8
    add     #__Cycles, w8               ; pointer to the total number of cycles
    add     w0, [w8++], w4              ; tmp = Cycles + Now()
    addc    w1, [w8++], w5              ; tmp = Cycles + Now()
    addc    w2, [w8++], w6              ; tmp = Cycles + Now()
    addc    w3, [w8++], w7              ; tmp = Cycles + Now()
    sub     w8, #8, w0                  ; w0 points to w0
    sub     w4, [w8++], [w0++]          ; Cycles = cycles + Now() - last
    subb    w5, [w8++], [w0++]          ; Cycles = cycles + Now() - last
    subb    w6, [w8++], [w0++]          ; Cycles = cycles + Now() - last
    subb    w7, [w8++], [w0++]          ; Cycles = cycles + Now() - last
    pop     w8
18:
.endif

;
; Here we test the priority fibers and internal functions 
;
TestSpawn:
    mov     _qvSpawn, w1
    ff1r    w1, w0                      ; w0 is offset 1 for bit-0, 2 for bit-1, 16 for bit-15
    bra     c, ExitInterrupt            ; if fibers need to be executed
;
; Here to execute fibers. This is a jump table and it executes the fiber
;
    clr     w1                          ; later used to compare if address is set
    mul.uu  w0,#0x5, w0
    bra     w0                        
;    
; We start with dummy NOP's because value 0 is not used in the ff1r instuction
    nop                ; NOT USED (VALUE 0 is tested)
    nop
    nop
    nop
    nop

;;;;;; PRIORITY FIBER 4
Prio4Fiber:
    bclr    _qvSpawn, #0
    mov     _qvSpawnFun+6, w0
    cpseq   w0, w1                      ; If function not set skip
    call    w0
    bra     TestSpawn

;;;;;; PRIORITY FIBER 3
Prio3Fiber:
    bclr    _qvSpawn, #1
    mov     _qvSpawnFun+4, w0
    cpseq   w0, w1                      ; If function not set skip
    call    w0
    bra     TestSpawn

;;;;;; PRIORITY FIBER 2
Prio2Fiber:
    bclr    _qvSpawn, #2
    mov     _qvSpawnFun+2, w0
    cpseq   w0, w1                      ; If function not set skip
    call    w0
    bra     TestSpawn

;;;;;; PRIORITY FIBER 1
Prio1Fiber:
    bclr    _qvSpawn, #3
    mov     _qvSpawnFun+0, w0
    cpseq   w0, w1                      ; If function not set skip
    call    w0
    bra     TestSpawn

;;;;;; FREE POSITION 
    bclr    _qvSpawn, #4
    nop
    nop
    nop
    bra     TestSpawn

;;;;;; Timer Expire
TimerExpireFiber:
    bclr    _qvSpawn, #5
    call   _xFbrTimerExpire ; two
    nop
    bra     TestSpawn

;;;;;; FREE POSITION  #6
    bclr    _qvSpawn, #6
    nop
    nop
    nop
    bra     TestSpawn

;;;;;; RTCC Expire
RtccExipereFiber:
    bclr    _qvSpawn, #7
    call    _xFbrRTCCExpire
    nop
    bra     TestSpawn

;;;;;; QUEUED FIBER  (Last one so size does not matter for computed branch)
; Only one place where we read the info from a the queue
QueuedFiber:
    mov     _qvFbrGet, w2               ; Get where we were
    bclr    _qvSpawn, #8                ; clear the bit that indicates queued fibers (here to prevent stall)
    mov     [w2++], w3                  ; w3 points to the next entry in the list and w2 to the function to call
                                        ; if this value is ok it must have the first bit set and if the bit is cleared
                                        ; it must not be zero. (it is posible that qFbrEnqueue() is interrupt by the 
                                        ; kernel if it was executed from a thread.)
    btst    [w2], #0                    ; Test if this is position is in use
    bra     z, TestSpawn                ; If not we are done 
    dec     [w2], w4                    ; remove the bit in the register and W4 is now address of function
    bra     z, TestSpawn                ; If address of function 0 then it was in the process of writing 
    mov     w3, _qvFbrGet               ; store the next entry in qvFbrGet
    mov     [w2+2], w0                  ; param 1
    mov     [w2+4], w1                  ; param 2
    clr     [w2]                        ; indicate free spot. Now qFbrEnqueue can write in this position
    call    w4                          ; call
    bra     QueuedFiber                 ; and again

;
;
;
; Here we will exit the interrupt by testing if we just exit the interrupt or need to do a
; full context switch
;
ExitInterrupt:
.ifndef __FAST__
    btst    _qvKernel,#_xStatisticsOn   ; statistics ON?
    bra     z, 169f

    rcall   _yGetCycles                 ; returns current cycles in W0-3
    mov     _qvRdyList, w7              ; Context of tcb to become active
    add     #__LastCycles, w7           ; pointer to the last cycle counter
    mov     w0, [w7+0]
    mov     w1, [w7+2]
    mov     w2, [w7+4]
    mov     w3, [w7+6]

169:    
.endif
    mov     _qvCurTcb,w0                ; W0 address of TCB
    cp      _qvRdyList                  ; Compare with Ready list
    bra     z, noContextSwitch          ; No thread switching so continue there
;
; Here if the scheduler wants us to switch the thread. Save all the stuff that we want to save
; on the TCB of the thread that is active
; w15, w7, w6, w5, w4, w3, w2, w1, w0, and RCOUNT are already saved SR comes from the retfie
    mov     w14,[w0+__w14]
    mov     w13,[w0+__w13]
    mov     w12,[w0+__w12]
    mov     w11,[w0+__w11]
    mov     w10,[w0+__w10]
    mov     w9,[w0+__w9]
    mov     w8,[w0+__w8]
    
.ifndef __FAST__
    mov     _qvRdyList, w1              ; Param1=CurTcb(w0) Param2=NextTcb(w1)
    btsc    _qvKernel,#_xSwitchNotificationOn            ; SwitchNotification
    call    _qKrnNtfSwitch              ; qKrnNtfSwitch(CurTcb,NextTcb)
.endif

    mov     _qvRdyList,w0               ; Make the highest job waiting, running
    mov     w0,_qvCurTcb                ; W0 is the thread to start
; 
; Here then test if we have to restore some EDS stacks
.ifndef __FAST__
    mov     [w0+__StackEdsPage], w1
    cp0     w1                          ; Does the new thread has a EDS stack?
    bra     z, 200f
;
; Here if the new thread uses a EDS stack
  .ifdef __HAS_EDS__
    cp0     _qvSharedStackOwner         ; Is there a shhared stack owner
    bra     z, 180f
    cp      _qvSharedStackOwner         ; Do we own the shared Stack (cp _qvSharedStackOwner with w0)
    bra     z, 200f                     ; yes so we are done
;
; We have to save the sharedstack in its EDS place 
    mov     _qvSharedStackOwner, w6
    mov     [w6+__StackEdsPage], w1
    mov     w1, 0x0034                  ; To page DSWPAG (we can use this because it will be restored later)
    mov     [w6+__StackSize], w1        ; number of bytes to copy
    asr     w1, w1                      ; number of words to copy
    dec     w1, w1                      ; compensate for the repeat instruction
    mov     [w6+__pStack], w4           ; To offset (bit 15 already set)
    mov     _qvSharedStack, w3          ; from address
    repeat  w1
    mov     [w3++], [w4++]
;
; Now restore the new thread
180:
    mov     [w0+__StackEdsPage], w1
    mov     w1, 0x0032                  ; To page DSRPAG (we can use this because it will be restored later)
    mov     [w0+__StackSize], w1        ; number of bytes to copy
    asr     w1, w1                      ; number of words to copy
    dec     w1, w1                      ; compensate for the repeat instruction
    mov     [w0+__pStack], w3           ; To offset (bit 15 already set)
    mov     _qvSharedStack, w4          ; from address
    repeat  w1
    mov     [w3++], [w4++]
    mov     w0, _qvSharedStackOwner     ; indicate we own the EDS stack
  .endif
.endif
200:
; Restore the registers for the new thread
    mov     [w0+__w14], w14
    mov     [w0+__w13], w13
    mov     [w0+__w12], w12
    mov     [w0+__w11], w11
    mov     [w0+__w10], w10
    mov     [w0+__w9], w9
    mov     [w0+__w8], w8

;
;     Here we got back when there is no context switch 
noContextSwitch:    
.ifndef __FAST__
    mov     _qvFbrTrackAdr, w1          ; Fiber tracking address
    mov     _qvFbrTrackBit, w2
    ior     w2, [w1], [w1]              ; Do it
    mov     [w0+__TrackAdr], w1         ; Tracking address
    mov.b   [w0+__TrackBit], w2         ; Tracking bit number
    ior.b   w2, [w1], [w1]              ; Do it
  .ifdef __HAS_EDS
    mov     [w0+__DsrPag], w1           ; DSRPAG
    mov     w1, DSRPAG                  ; DSRPAG
    mov     [w0+__DswPag], w1           ; DSWPAG
    mov     w1, DSWPAG                  ; DSWPAG
  .else
    mov     [w0+__PsvPag], w1           ; PSVPAG
    mov     w1, PSVPAG                  ; PSVPAG
  .endif
    mov     [w0+__TblPag], w1           ; TBLPAG
    mov     w1, TBLPAG                  ; TBLPAG
.endif
    mov     [w0+__rcount], w1           ; RCOUNT
    mov     w1,RCOUNT                   ; RCOUNT

    mov     [w0+__w7], w7
    mov     [w0+__w6], w6
    mov     [w0+__w5], w5
    mov     [w0+__w4], w4
    mov     [w0+__w3], w3
    mov     [w0+__w2], w2
    
.ifndef __FAST__                        ; The Thread SPLIM is always greater the the Int-Stack limit
    mov     [w0+__splim], w1            ; The new limit (SPLIM from thread is constant)
    mov     w1,SPLIM                    ; and activate the new limit
.endif

    mov     [w0+__w1], w1               ; 
    mov     [w0+__sp], w15              ; Set the stack pointer
    mov     [w0+__w0], w0               ; W0 itself
    dec     _qvInCritical
    retfie                              ; restore the SR
    
;
;
; Here if the critical section was higher than level 0 (Value -1) It is already 
; incremented so we have to decrement the value here to compensate. 
; Because there is an increment done and disable interrupt we are sure that 
; the Exit critical that will follow will activate us again. We do NOT clear the 
; IFS register.
;
499:
    dec     _qvInCritical               ; This is to compensate for the increment
    DISABLE_KERNEL_INTERRUPT            ; We disable the kernel interrupt. The level lower will enable it again.
    retfie
    
