/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

.macro SET_KERNEL_INTERRUPT
	bset IFS1, #11
.endm
.macro CLR_KERNEL_INTERRUPT
	bclr IFS1, #11
.endm
.macro TST_KERNEL_INTERRUPT
	btst IFS1, #11
.endm
.macro qCrtEnter
	bclr IEC1, #11
	inc		_qvInCritical
.endm
.macro qCrtExit
	dec	_qvInCritical
	bset IEC1, #11
.endm
.macro DISABLE_KERNEL_INTERRUPT
	bclr IEC1, #11
.endm
.macro ENABLE_KERNEL_INTERRUPT
	bset IEC1, #11
.endm

.section qKernel,code


.set __Entries, 0
.set __pObjList, 2
.set __pName, 4
.set __Type, 6
.set __Status, 7
.set __BlockSizeINT, 8
.set __BlockSize, 10
.set __MaxBlocks, 12
.set __pNtfWriter, 14 
.set __pNtfReader, 16
.set __pOut, 18				; Offset relative to Entries
.set __pIn, 20				; Offset relative to Entries
.set __pMax, 22				; Offset relative to Entries
.set __Data, 24				; Offset relative to Entries

;*******************************************************************************************
;
; INPUT
; W0 = Pip
; W1 = Data
; W2 = Nbr struct MUST > 0
;
; OUTPUT
; W0 Entries done
;
; Used
; W7 = AddressOf(Entries)
; W6 = BlockSize
; W5 = AddressOf(pOut)
; W4 = value pMax
; W0 = Entries done
;
;
; Only compile if we don't want to check params, otherwise use the C version.
;
	.global _qPipGetFast
_qPipGetFast:
	dec		[w0], [w0]                  ; Decrement entries. If it was 0 then the result would be -1
	bra		lt, 99f                     ; This is less than zero so jump and add later
;
	mov		[w0 + __BlockSizeINT], w6 	; BlockSize in integers
	mov		[w0 + __pOut], w5
	mov		[w0 + __pMax], w4
	dec		w6, w6						; if 0 then not word moves
	bra		lt, 50f                     ; goto byte moves
	repeat	w6
	mov		[w5++], [w1++]
	cpsne	w5, w4						; if we are not at the end of the buffer skip moving in front of buffer
	add		w0, #__Data, w5 			; New Address of Data to read
	dec		w2, w2
	bra		nz, 10f
	mov		w5, [w0 + __pOut]
	retlw	#1, w0
10:
	mov		w0, w7
	mov		#1, w0                      ; one structure has been done
20:
	dec		[w7], [w7]
	bra		lt, 95f
	inc		w0, w0
	Repeat	w6
	mov		[w5++], [w1++]
	cpsne	w5, w4						; if we are not at the end of the buffer skip moving in front of buffer
	add		w7, #__Data, w5 			; New Address of Data to read
	dec		w2, w2
	bra		nz, 20b
	mov		w5, [w7 + __pOut]
	return
;
;
; Here if byte moves are required
50:	
	mov		[w0 + __BlockSize], w6
	dec		w6, w6						; BlockSize is exact the size
	repeat	w6
	mov.b	[w5++], [w1++]
	cpsne	w5, w4						; if we are not at the end of the buffer skip moving in front of buffer
	add		w0, #__Data, w5 			; New Address of Data to read
	dec		w2, w2                      ; Decrement to do. Entries already decremented
	bra		nz, 60f
	mov		w5, [w0 + __pOut]
	retlw	#1, w0
60:
	mov		w0, w7
	mov		#1, w0
70:
	dec		[w7], [w7]
	bra		lt, 95f
	inc		w0, w0
	Repeat	w6
	mov.b	[w5++], [w1++]
	cpsne	w4, w5						; if we are not at the end of the buffer skip moving in front of buffer
	add		w7, #__Data, w5 			; set at beginning of buffer
	dec		w2, w2
	bra		nz, 70b
	mov		w5, [w7 + __pOut]
	return

95:
	mov		w5, [w7 + __pOut]
	inc		[w7], [w7]
	return

99:
	inc		[w0],[w0]
	retlw	#0,w0


;
; Very fast simple routine. Can only be used if the blockSize = sizeof(int)
; This is not checked
; only uses W0, W1, W2, W3
; W0 = Pip
; W1 = Data
;
	.global _qPipGetWrdFast
_qPipGetWrdFast:
	dec		[w0],[w0]
	bra		lt, 99f
;
	mov		[w0 + __pOut], w2
	mov		[w0 + __pMax], w3
	mov		[w2++], [w1++]
	cpsne	w2, w3						; if we are not at the end of the buffer skip moving in front of buffer
	add		w0, #__Data, w2 			; New Address of Data to read
	mov		w2, [w0 + __pOut]
	retlw	#1, w0

99:
	inc		[w0], [w0]
	retlw	#0, w0

;
; Very fast simple routine. Can only be used if the blockSize = sizeof(char)
; This is not checked
; only uses W0, W1, W2, W3
; W0 = Pip
; W1 = Data
;
	.global _qPipGetBytFast
_qPipGetBytFast:
	dec		[w0], [w0]
	bra		lt, 99f
;
	mov		[w0 + __pOut], w2
	mov		[w0 + __pMax], w3
	mov.b   [w2++], [w1++]
	cpsne	w2, w3						; if we are not at the end of the buffer skip moving in front of buffer
	add		w0, #__Data, w2 			; New Address of Data to read
	mov		w2, [w0 + __pOut]
	retlw	#1, w0
99:
	inc		[w0], [w0]
	retlw	#0,w0

;*******************************************************************************************
;
; INPUT
; W0 = Pip
; W1 = Data
; W2 = Nbr struct
;
; OUTPUT
; W0 Blocks done
;
; Used
; W7 = AddressOf(Entries) and pip (only used in multiple entries)
; W6 = BlockSize
; W5 = AddressOf(pIn)
; W4 = value pMax
; W3 = MaxBlocks in Pipe
; W0 = Entries done
;
;
; Only compile if we don't want to check params, otherwise use the C version.
;
	.global _qPipPutFast
_qPipPutFast:
	mov		[w0 + __MaxBlocks], w3
	cp		w3, [w0]					; MaxBlocks - Entries so is there space in pipe
	bra		le, 99f
;
	mov		[w0 + __BlockSizeINT], w6 	; BlockSize in integers
	mov		[w0 + __pIn], w5
	mov		[w0 + __pMax], w4
	dec		w6, w6						; if < 0 then it is an integer move
	bra		lt, 50f
	repeat	w6
	mov		[w1++], [w5++]
	inc		[w0], [w0]					; incrment entries (after the move)
	cpsne	w5, w4						; if we are not at the end of the buffer skip moving in front of buffer
	add		w0, #__Data, w5 			; New Address of Data to write
	dec		w2, w2
	bra		nz, 10f
	mov		w5, [w0 + __pIn]
	retlw	#1, w0
10:
	mov		w0, w7
	mov		#1, w0
20:
	cp		w3, [w7]					; MaxBlocks - Entries
	bra		le, 95f
	inc		w0, w0						; increment blocks done
	Repeat	w6
	mov		[w1++], [w5++]
	inc		[w7], [w7]					; increment entries in pipe (after the move)
	cpsne	w5, w4						; if we are not at the end of the buffer skip moving in front of buffer
	add		w7, #__Data, w5 			;New Address of Data to read
	dec		w2, w2
	bra		nz, 20b
	mov		w5, [w7 + __pIn]
	return
;
;
; Here if byte moves are required
50:
	mov		[w0 + __BlockSize], w6
	dec		w6, w6						; BlockSize is exact the size
	repeat	w6
	mov.b	[w1++], [w5++]
	inc		[w0], [w0]
	cpsne	w5, w4
	add		w0, #__Data, w5 			;New Address of Data to read
	dec		w2, w2
	bra		nz, 60f
	mov		w5, [w0 + __pIn]
	retlw	#1, w0
60:
	mov		w0, w7
	mov		#1, w0
70:
	cp		w3, [w7]					; MaxBlocks - Entries
	bra		le, 95f
	inc		w0, w0
	Repeat	w6
	mov.b	[w1++], [w5++]
    inc     [w7], [w7]
	cpsne	w4, w5
	add		w7, #__Data, w5 			;New Address of Data to read
	dec		w2, w2
	bra		nz, 70b
95:
	mov		w5, [w7 + __pIn]
	return

99:
	retlw	#0, w0

;
; Very fast simple routine. Can only be used if the blockSize = sizeof(int)
; This is not checked
; only uses W0, W1, W2, W3
; W0 = Pip
; W1 = Data
;
	.global _qPipPutWrdFast
_qPipPutWrdFast:
	mov		[w0 + __MaxBlocks], w2
	cp		w2, [w0]					; MaxBlocks - Entries so is there space in pipe
	bra		le, 99f
;
	mov		[w0 + __pIn], w2
	mov		[w0 + __pMax], w3
	mov		[w1++], [w2++]
	inc		[w0], [w0]					; incrment entries (after the move)
	cpsne	w2, w3						; if we are not at the end of the buffer skip moving in front of buffer
	add		w0, #__Data, w2 			; New Address of Data to write
	mov		w2, [w0 + __pIn]
	retlw	#1, w0
99:
	retlw	#0, w0

;
; Very fast simple routine. Can only be used if the blockSize = sizeof(char)
; This is not checked
; only uses W0, W1, W2, W3
; W0 = Pip
; W1 = Data
;
	.global _qPipPutBytFast
_qPipPutBytFast:
	mov		[w0 + __MaxBlocks], w2
	cp		w2, [w0]					; MaxBlocks - Entries so is there space in pipe
	bra		le, 99f
;
	mov		[w0 + __pIn], w2
	mov		[w0 + __pMax], w3
	mov.b	[w1++], [w2++]
	inc		[w0], [w0]					; incrment entries (after the move)
	cpsne	w2, w3						; if we are not at the end of the buffer skip moving in front of buffer
	add		w0, #__Data, w2 			; New Address of Data to write
	mov		w2, [w0 + __pIn]
	retlw	#1, w0
99:
	retlw	#0, w0

	
	
;*******************************************************************************************
;
; INPUT
; W0 = Pip
; W1 = Data
;
; OUTPUT
; W0 Entries done
;
; Used
; W6 = BlockSize
; W5 = AddressOf(pOut)
; W4 = value pMax
	.global _qPipReadFast
_qPipReadFast:
	dec		[w0],[w0]
	bra		lt, 90f
;
	qCrtEnter
	mov		[w0 + __BlockSizeINT], w6 	; BlockSize in integers
	mov		[w0 + __pOut], w5
	mov		[w0 + __pMax], w4
	dec		w6, w6						; if < 0 then byte moves
	repeat	w6							; do the word moves
	mov		[w5++], [w1++]
	cpsne	w5, w4						; if we are not at the end of the buffer skip moving in front of buffer
	add		w0, #__Data, w5 			; New Address of Data to read
	mov		w5, [w0 + __pOut]			; we are done so write pointer back
	qCrtExit
	mov		[w0 + __pNtfWriter], w2
	cp0		w2							; is there a notify reader?
	bra		nz, 11f						; yes so continu there
	retlw	#1, w0						; else return with 1 entry moved
11:	mov		#1,	w1						; 1 entry, w0 is qtPIP
	call	w2							; call the reader notification function
	retlw	#1, w0
;
; Here if there is nothing in the pipe after the first test
;
90:	inc		[w0], [w0]
	mov		[w0 + __pNtfWriter], w2
	cp0		w2
	bra		nz, 92f
	retlw	#0, w0
	
92:	clr		w1							; Entries done
	call	w2							; call the Notify reader
	retlw	#0,w0


;*******************************************************************************************
;
; INPUT
; W0 = Pip
; W1 = Data
;
; OUTPUT
; W0 Blocks done
;
; Used
; W7 = AddressOf(Entries) and pip (only used in second part)
; W6 = BlockSize
; W5 = AddressOf(pIn)
; W4 = value pMax
; W3 = MaxBlocks in Pipe
	.global _qPipWriteFast
_qPipWriteFast:
	mov		[w0 + __MaxBlocks], w3
	cp		w3, [w0]					; MaxBlocks - Entries so is there space in pipe
	bra		le, 90f
;
	qCrtEnter							;
	mov		[w0 + __BlockSizeINT], w6 	; BlockSize in integers
	mov		[w0 + __pIn], w5
	mov		[w0 + __pMax], w4
	dec		w6, w6						; if > 0 then it is an integer move
	repeat	w6
	mov		[w1++], [w5++]
	inc		[w0],[w0]					; increment entries (after the move)
	cpsne	w5, w4						; if we are not at the end of the buffer skip moving in front of buffer
	add		w0, #__Data, w5 			; New Address of Data to write
	mov		w5, [w0 + __pIn]
	qCrtExit						;
	mov		[w0 + __pNtfReader], w2
	cp0		w2							; is there a notify reader?
	bra		nz, 11f						; yes so continu there
	retlw	#1, w0						; else return with 1 entry moved
11:	mov		#1,	w1						; 1 entry, w0 is qtPIP
	call	w2							; call the reader notification function
	retlw	#1, w0
;
; Here if we don't have any space in the pipe after the first test
;
90:
	mov	[w0+ __pNtfReader], w2
	cp0		w2
	bra		nz, 92f
	retlw	#0, w0

92:	clr		w1							; Entries done
	call	w2							; call the Notify reader
	retlw	#0, w0


