/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

.ifdef __dsPIC30F
	.include "p30Fxxxx.inc"
.endif
.ifdef __dsPIC33E__
	.include "p33Exxxx.inc"
.endif
.ifdef __dsPIC33F
	.include "p33Fxxxx.inc"
.endif
.ifdef __PIC24E
	.include "p24Exxxx.inc"
.endif
.ifdef __PIC24F
	.include "p24Fxxxx.inc"
.endif
.ifdef __PIC24FK
	.include "p24Fxxxx.inc"
.endif
.ifdef __PIC24H
	.include "p24Hxxxx.inc"
.endif
.ifdef __PIC32MX
	.include "p32xxxx.inc"
.endif
.macro SET_KERNEL_INTERRUPT
	bset IFS1, #11
.endm
.macro CLR_KERNEL_INTERRUPT
	bclr IFS1, #11
.endm
.macro TST_KERNEL_INTERRUPT
	btst IFS1, #11
.endm
.macro qCrtEnter
	bclr IEC1, #11
	inc		_qvInCritical
.endm
.macro qCrtExit
	dec	_qvInCritical
	bset IEC1, #11
.endm
.macro DISABLE_KERNEL_INTERRUPT
	bclr IEC1, #11
.endm
.macro ENABLE_KERNEL_INTERRUPT
	bset IEC1, #11
.endm

__E_TYPE__ = 0
.ifdef __dsPIC33E
    __E_TYPE__ = 1
.else
    .ifdef __PIC24E
        __E_TYPE__ = 1
    .endif
.endif

.section qKernel,code
;
.set __pFunctName, 0
.set __TrapAddress, 0
.set __Uptime, 4
.set __pThread, 8
.set __ErrNbr, 10
.set __StackType, 12
.set __Zone, 13
.set __SPLIM, 14
.set __InCritical, 16

_xInitDone = 0
_xOsStarted = 1
_xStatisticsOn = 2
_xSwitchNotificationOn = 3
_xUSecOn = 4
_xPriorityCheckOn = 5
_xTimerInterruptOn = 6
_xRtccInterruptOn = 7


	.weak __OscillatorFail
__OscillatorFail:
	mov		#0x1000, w13				;  error
	bra		_restError
	
	.weak  __AddressError
__AddressError:
	mov		#0x1001,w13				    ; error
	bra		_restError

	.weak  __HardTrapError
__HardTrapError:
	mov		#0x1002,w13                 ; error
	bra		_restError

	.weak   __StackError
__StackError:
	mov		#0x1003, w13                ;  error
	bra		_restError

	.weak  __MathError
__MathError:
	mov		#0x1004, w13                ;  error
	bra		_restError

    .weak  __DMACError
__DMACError:
	mov		#0x1005,w13                 ; error
	bra		_restError

    .weak  __SoftTrapError
__SoftTrapError:
	mov		#0x1006,w13                 ; error
	bra		_restError

    .weak  __ReservedTrap7
__ReservedTrap7:
	mov		#0x1007,w13                 ; error
	bra		_restError





; W1 = Error then call xErrNtf
_restError:
	mov		[w15-4],w12
	mov		w12, _qvErr+__TrapAddress	; address of error
	mov		[w15-2],w12
	and		#0x7f, w12
	mov		w12, _qvErr+__TrapAddress+2 ;
	mov		w13, w0						; address error
	call	_xErrNtf
	retfie


;	.global _yErrorReset
;_yErrorReset:
;	mov		#0xe0,w0				; mask ISR level 4, 2 and 1
;	and		SR, wreg				; keep the IPL bits that are set
;	asr		w0, #5, w0				; shift the so we have the number
;    btst    CORCON, #3              ; Test IPL3
;    bra     z, 2f                   ; if zero skip
;    add     #8, w0                  ; Add the 8 to get the real interrupt priority level
;2:	mov.b	wreg, _qvErr+__Zone		; and store in the zone
;	mov		_qvCurTcb, w0			; current TCB
;	mov		w0, _qvErr+__pThread	; Save the current thread
;	rcall	_yRtccGetUptime
;	mov		w0,  _qvErr+__Uptime
;	mov		w1,  _qvErr+__Uptime+2
;	mov		SPLIM, w0
;	mov		w0,  _qvErr+__SPLIM
;	mov		_qvIntStackLimit, w1
;	cp		w1,w0					; SPLIM =? IntStackLimit
;	bra		nz, 10f					; no then the thread stack
;	mov.b	#1,w0
;	mov.b	wreg, _qvErr+__StackType
;    mov.b
;	reset
;10:
;	mov.b	#0,w0
;	mov.b	wreg, _qvErr+__StackType
;	reset

.global _yModeIdle
_yModeIdle:
    mov     _qvTrackingIdleAdr, w6
    mov     _qvTrackingRunAdr, w7
    mov.b   _qvTrackingRunBit, wreg
    mov     w0, w1
    mov.b   _qvTrackingIdleBit, wreg
    ior.b   w0, [w6], [w6]          ; Clear the idle bit
    ior.b   w1, [w7], [w7]          ; set the run bit
.if __E_TYPE__ == 0
    disi	#4
	nop ; 1
	nop ; 2
	nop ; 3
.else
    disi	#13                     ; Maximum 13 cycles for the
	nop ; 1
	nop ; 2
	nop ; 3
	nop ; 4
	nop ; 5
	nop ; 6
	nop ; 7
	nop ; 8
	nop ; 9
	nop ; 10
	nop ; 11
	nop ; 12
.endif
	btss    INTTREG, #15            ; If there is an outstanding interrupt skip Idle mode
	pwrsav	#1                      ; IDLE_MODE
    ior.b   w1, [w7], [w7]          ; Clear the run bit
    ior.b   w0, [w6], [w6]          ; Set the idle bit
	return

.global _yModeSleep
_yModeSleep:
    mov     _qvTrackingSleepAdr, w6
    mov     _qvTrackingRunAdr, w7
    mov.b   _qvTrackingRunBit, wreg
    mov     w0, w1
    mov.b   _qvTrackingSleepBit, wreg
    ior.b   w0, [w6], [w6]          ; Clear the sleep bit
    ior.b   w1, [w7], [w7]          ; Set the run bit
.if __E_TYPE__ == 0
    disi	#4
	nop ; 1
	nop ; 2
	nop ; 3
.else
    disi	#13
	nop ; 1
	nop ; 2
	nop ; 3
	nop ; 4
	nop ; 5
	nop ; 6
	nop ; 7
	nop ; 8
	nop ; 9
	nop ; 10
	nop ; 11
	nop ; 12
.endif
	btss    INTTREG, #15            ; If there is an outstanding interrupt skip Sleep mode
	pwrsav	#0                      ; SLEEP_MODE
    ior.b   w0, [w6], [w6]          ; Clear the idle bit
    ior.b   w1, [w7], [w7]          ; Set the run bit
	return



; W0/W1 = EDS address (Page/Offset)
; returns the Page in W0
; unsigned yEdsPage(__eds__ void* p);
.global _yEdsPage
_yEdsPage:
	mov		w1, w0
	return
	
; W0/W1 = EDS address (Page/Offset)
; returns the Offset in W0 and will set bit 15
; unsigned yEdsOffset(epVOID p);
.global _yEdsOffset
_yEdsOffset:
	return
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; uint32_t qDivU3216R(uint32_t dividend, uint16_t divisor, uint16_t* remainder)
;
; Total cycles is 45 (0.6 uSec on a PIC24E @70MIPS)
    .global _qDivU3216R
_qDivU3216R:
    mov     w0, w4        ; save LSW of dividend
    repeat  #17
    div.u   w1, w2        ; 16/16 divide
    exch    w0, w4        ; save result, remainder becomes MSW of next operation
    repeat  #17
    div.ud  w0, w2        ; 32/16 divide
    mov     w1, [w3]      ; Save remainder
    mov     w4, w1        ; move in MSW
    return
;
;
;
; uint32_t qDivU3216(uint32_t dividend, uint16_t divisor)

; Total cycles is 44 (0.6 uSec on a PIC24E @70MIPS)
    .global _qDivU3216
_qDivU3216:
    mov     w0, w4        ; save LSW of dividend
    repeat  #17
    div.u   w1, w2        ; 16/16 divide
    exch    w0, w4        ; save result, remainder becomes MSW of next operation
    repeat  #17
    div.ud  w0, w2        ; 32/16 divide
    mov     w4, w1        ; move in MSW
    return
;
;
;
; uint64_t qDivU6416R(uint64_t dividend, uint16_t divisor, uint16_t* remainder);
;
; w3-w0 = dividend
; w4 = divisor
; w5 = pointer to remainder
;
; Total cycles is 91 (1.3 uSec on a PIC24E @70MIPS)
    .global _qDivU6416R
_qDivU6416R:
    mov     w4, w7            ; divisor
    mov     w5, w6            ; address of remainder
    mov     w0, w4
    mov     w1, w5            ; dividend is now in w3-w2-w5-w4
    repeat  #17
    div.u   w3, w7            ; w3/w7 res=w0 rem=w1
    push    w0                ; result on stack
    mov     w2, w0
    repeat  #17
    div.ud  w0, w7            ; w1:w0/w7 res=w0 rem=w1
    push    w0
    mov     w5, w0
    repeat  #17
    div.ud  w0, w7
    push    w0
    mov     w4, w0
    repeat  #17
    div.ud  w0, w7
    mov     w1, [w6]
    pop     w1
    pop     w2
    pop     w3
    return
;
;
;
; uint64_t qDivU6416(uint64_t dividend, uint16_t divisor);
; This function rounds the result.
; w3-w2-w1-w0 = dividend
; w4 = divisor
;
; Total cycles is 93 (1.31 uSec on a PIC24E @70MIPS)
    .global _qDivU6416
_qDivU6416:
    mov     w4, w7            ; divisor stays in w7
    mov     w0, w4
    mov     w1, w5            ; dividend is now in w3-w2-w5-w4
    repeat  #17
    div.u   w3, w7            ; w3/w7 res=w0 rem=w1
    push    w0                ; result on stack
    mov     w2, w0
    repeat  #17
    div.ud  w0, w7            ; w1:w0/w7 res=w0 rem=w1
    push    w0
    mov     w5, w0
    repeat  #17
    div.ud  w0, w7
    push    w0
    mov     w4, w0
    repeat  #17
    div.ud  w0, w7
    pop     w1
    pop     w2
    pop     w3
    return

;
;
; uint16_t qDivScaling(uint16_t v1, uint16_t v2, uint16_t v3);
; This is a often used scaling function result=(v1*v2)/v3
;
	.global _qDivScaling
_qDivScaling:
	mul.uu	w0, w1, w0		; w0-w1 = v1*v2
	repeat	#17
	div.ud	w0, w2			; 32/16 divide W0=result W1=remainder
	return

;
; uint32_t qMulU3216(uint32_t v1, uint16_t v2)
; //TODO make this a inline function
;
    .global _qMulU3216
_qMulU3216:
    mul.uu  w1, w2, w4      ;
    mul.uu  w0, w2, w0
    add     w4, w1, w1
    return

;
; uint32_t qMulU3232(uint32_t v1, uint32_t v2)
; //TODO make this a inline function
;
    .global _qMulU3232
_qMulU3232:
	mul.su w1,w2,w4
	mul.su w3,w0,w6
	mul.uu w0,w2,w0
	add w4,w1,w1
	add w6,w1,w1
    return

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

.set _MPL_pFirstBlk, 0
.set _MPL_pNextPool, 4
.set _MPL_BlockSize, 8
.set _MPL_Total, 10

	.global _qMemAllocFromPoolFast
_qMemAllocFromPoolFast:
	qCrtEnter
	dec2 	[w0], w1				; Ugly code to save 1 cycle Mov and next cp0 is
	bra		ltu, 5f
	mov		[++w1], [w0]			; [++w1] to compensate for ugly dec2
									; write next back to pMpl->pBlk
	mov		w0, [w1++]				; write pool in memory
	mov		w1, w0					; w1 points to usable memory and now in w0
	qCrtExit
	return
;
; No block so get it from the heap
;
5:	push	w0
	mov		[w0+_MPL_BlockSize],w0
	add		#2,w0					; add integer for the pointer to mem block
	call	_qHeaAlloc
	pop		w1						; W1 now the pool address
	cp0		w0						; Did the heap returned memory
	bra		z, 7f					; no then skip the writing of the pool
	add		w1, #_MPL_Total, w2		; W2 points to total
	inc		[w2], [w2]				; Increment total
	mov		w1, [w0++]				; Write the address of the pool in first word

7:	qCrtExit
	return
	
	
;;
;; w0 = pojnter to memory INPUT
;	.global _qMemFreeFast
;_qMemFree:
;	qCrtEnter
;	mov		[--w0], w1				; w1 points to memory descriptor w0 points to the memory
;	mov		[w1], [w0]				; save the pointer
;	mov.w	w0, [w1]				; ((pMBK)mem)->pNext = p->pBlk;
;	qCrtExit
;	return
;



__E_TYPE__ = 0
.ifdef __dsPIC33E
    __E_TYPE__ = 1
.else
    .ifdef __PIC24E
        __E_TYPE__ = 1
    .endif
.endif


; int16_t = qThrDelay(uint16_t USec)
; w0=USec
.global _qThrDelay
_qThrDelay:
    mov		_qvClockFrequencyDiv15625, w2
	mul.uu	w0, w2, w0
	; The result we divide by 64 (is >>6) to get
	; uSec*(Freq/15625/64) is uSec*(Freq/1000000) the number of cycles
	sl		w1, #10, w2
	lsr		w0, #6, w0
	ior		w2, w0, w0
.if __E_TYPE__ ==1
    sub     #21, w0             ; Subtract 13 for call/ret overhead and 8 for uSec to Cycles
.else
    sub     #14, w0             ; Subtract 6 for call/ret overhead and 8 for uSec to Cycles
.endif
    repeat  w0
	nop
	return
