/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

.ifdef __dsPIC30F
	.include "p30Fxxxx.inc"
.endif
.ifdef __dsPIC33E__
	.include "p33Exxxx.inc"
.endif
.ifdef __dsPIC33F
	.include "p33Fxxxx.inc"
.endif
.ifdef __PIC24E
	.include "p24Exxxx.inc"
.endif
.ifdef __PIC24F
	.include "p24Fxxxx.inc"
.endif
.ifdef __PIC24FK
	.include "p24Fxxxx.inc"
.endif
.ifdef __PIC24H
	.include "p24Hxxxx.inc"
.endif
.ifdef __PIC32MX
	.include "p32xxxx.inc"
.endif
.macro SET_KERNEL_INTERRUPT
	bset IFS1, #11
.endm
.macro CLR_KERNEL_INTERRUPT
	bclr IFS1, #11
.endm
.macro TST_KERNEL_INTERRUPT
	btst IFS1, #11
.endm
.macro qCrtEnter
	bclr IEC1, #11
	inc		_qvInCritical
.endm
.macro qCrtExit
	dec	_qvInCritical
	bset IEC1, #11
.endm
.macro DISABLE_KERNEL_INTERRUPT
	bclr IEC1, #11
.endm
.macro ENABLE_KERNEL_INTERRUPT
	bset IEC1, #11
.endm


.set __ListType , 7
.set __Prio, 9
.set __pPrev, 10


;
;
.section qKernel,code

	.global _qThrResumeV4
_qThrResumeV4: 
	mov		SR, w7
	and		#0b11000000, w7
	bra		nz, _qThrResumeISR
;******************************************************************************************
; Here if we are in a thread or fiber
;******************************************************************************************
;
_qThrResumeTHR:
	qCrtEnter
;	xInsRdyList		; insert the thread in the ready list
;	xInsRdyList		; insert the thread in the ready list
;	xInsRdyList		; insert the thread in the ready list
	mov		#1, w1				; LIST_TYPE_READY
	mov.b	w1,[w0+__ListType]
	mov		_qvRdyList, w1
	mov		#_qvRdyList, w3
	mov.b	[w0+__Prio], w4
	mov.b	[w1+__Prio], w2
	sub.b	w4, w2, [w15]
	bra		nc, 50f
	mov		w1, [w0]
	mov		w3, [w0+__pPrev]
	mov		w0, [w3]
	mov		w0, [w1+__pPrev]
;	xInsRdyList		; insert the thread in the ready list
;	xInsRdyList		; insert the thread in the ready list
;	xInsRdyList		; insert the thread in the ready list
	bra		_ySwitchThread	; NOT in an ISR so we can switch ySwitchThread will return us
;
;	Here if the thread to resume not the first thread in the ready list
50:
	mov		w1,w3
	clr		w7					; not the first in the ready list
	mov		[w1], w1
;
	mov.b	[w1+__Prio], w2
	sub.b	w4, w2, [w15]
	bra		nc, 50b
	mov		w1, [w0]
	mov		w3, [w0+__pPrev]
	mov		w0, [w3]
	mov		w0, [w1+__pPrev]
;	xInsRdyList		; insert the thread in the ready list
;	xInsRdyList		; insert the thread in the ready list
;	xInsRdyList		; insert the thread in the ready list
	qCrtExit
	return

;
;******************************************************************************************
; Here if we are in an interrupt (interrupt level 2 and higher)
;******************************************************************************************
;
_qThrResumeISR: 
	qCrtEnter				; first time it goes from 0xff to 0x00 if 0x00 we are the first so do work
	bra		nz,100f				; if (!yCrtEnterFirst()) goto 100
;
; DO work
;
;	xInsRdyList		; insert the thread in the ready list
;	xInsRdyList		; insert the thread in the ready list
;	xInsRdyList		; insert the thread in the ready list
	mov		#1, w1				; LIST_TYPE_READY
	mov.b	w1,[w0+__ListType]
	mov		_qvRdyList, w1
	mov		#_qvRdyList, w3
	mov.b	[w0+__Prio], w4
	mov.b	[w1+__Prio], w2
	sub.b	w4, w2, [w15]
	bra		nc, 10f
;	Here if new thread the first in the ready list and must run
	mov		w1, [w0]
	mov		w3, [w0+__pPrev]
	mov		w0, [w3]
	mov		w0, [w1+__pPrev]
;	xInsRdyList		; insert the thread in the ready list
;	xInsRdyList		; insert the thread in the ready list
;	xInsRdyList		; insert the thread in the ready list
	SET_KERNEL_INTERRUPT		; Let the kernel interrupt switch the threads
	qCrtExit
	return						
;
;	Here if the thread to resume not the first thread in the ready list. We don't 
;	have to switch the context. So just an exit 
10:
	mov		w1,w3
	clr		w7					; not the first in the ready list
	mov		[w1], w1
;
	mov.b	[w1+__Prio], w2
	sub.b	w4, w2, [w15]
	bra		nc, 10b
	mov		w1, [w0]
	mov		w3, [w0+__pPrev]
	mov		w0, [w3]
	mov		w0, [w1+__pPrev]
;	xInsRdyList		; insert the thread in the ready list
;	xInsRdyList		; insert the thread in the ready list
;	xInsRdyList		; insert the thread in the ready list
	qCrtExit
	return	
;
; qFbrEnqueue1() implementation 
; W0 = pTCB
100:	
	qCrtExit
	mov		_qvFbrPut, w2
	mov		[w2++],w3
	mov		w3, _qvFbrPut
	btsts	[w2],#0
	bra		nz,	100b
	mov		#handle(_qThrResume), w3
	mov 	w0, [w2+2]
	add	 	w3, #1, [w2]
	bset 	_qvSpawn, #8
; END qFbrEnqueue1() Implementation. 
;	Kernel interrupt not set because we are in the second or higher critical section.
;	The Exit critical of the first will set the kernel interrupt.
	SET_KERNEL_INTERRUPT
	return
;

