/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  **
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  **
 **  <http://www.quasarsoft.com/license.html>                                       **
 **                                                                                 **
 *************************************************************************************/

#include "qKernel.h"

unsigned qPipPutFast(pPIP pPip, void* pData, unsigned NbrBlocks) {
    unsigned entriesDone=0;
/*
 * qPipPut will only touch the Entries and pIn. This function can be interrupted by
 * qPipGet. That function will only touch Entries and pOut so the only field that needs
 * to be updated in one instruction is Entries.
 * All entities that write in the pipe with qPipPut must have the same interrupt priority
 * and threads can only use this function in a critical section.
 */
    while (NbrBlocks-- && pPip->Entries < pPip->MaxBlocks) {
        if (pPip->BlockSizeINT!=0) {
            qWrdMov((unsigned*)pData, (unsigned*)pPip->pIn, pPip->BlockSize);
        }
        else {
            qBytMov(pData, pPip->pIn, pPip->BlockSize);
        }
        qWrdIncAtomic(&pPip->Entries);
        pData += pPip->BlockSize;
        pPip->pIn += pPip->BlockSize;
        if (pPip->pIn==pPip->pMax) {
            pPip->pIn = &pPip->Data;
        }
        entriesDone++;
    }
    return entriesDone;
}

bool qPipPutWrdFast(pPIP pPip, unsigned* pData) {
    unsigned entriesDone=0;
    if (pPip->Entries < pPip->MaxBlocks) {
        *(unsigned*)pPip->pIn = *pData;
        qWrdIncAtomic(&pPip->Entries);
        pPip->pIn += 4;
        if (pPip->pIn==pPip->pMax) {
            pPip->pIn = &pPip->Data;
        }
        entriesDone++;
    }
    return entriesDone;
}

bool qPipPutBytFast(pPIP pPip, uint8_t* pData) {
    unsigned entriesDone=0;
    if (pPip->Entries < pPip->MaxBlocks) {
        *(uint8_t*)pPip->pIn = *pData;
        qWrdIncAtomic(&pPip->Entries);
        pPip->pIn += 1;
        if (pPip->pIn==pPip->pMax) {
            pPip->pIn = &pPip->Data;
        }
        entriesDone++;
    }
    return entriesDone;
}

unsigned qPipGetFast(pPIP pPip, void* pData, unsigned NbrBlocks) {
    unsigned entriesDone=0;
/*
 * qPipGet will only touch the Entries and pOut. This function can be interrupted
 * by qPipPut. That function will only touch Entries and pIn so the only field that needs
 * to be updated in one instruction is Entries.
 * All entities that read from the pipe with qPipGet must have the same interrupt priority
 * and threads can only use this function in a critical section.
 */
    while (NbrBlocks-- && pPip->Entries > 0)  {
        if (pPip->BlockSizeINT!=0) {
            qWrdMov((unsigned*)pPip->pOut, (unsigned*)pData, pPip->BlockSize);
        }
        else {
            qBytMov(pPip->pOut, pData, (unsigned)pPip->BlockSize);
        }
        qWrdDecAtomic(&pPip->Entries);
        pData += pPip->BlockSize;
        pPip->pOut += pPip->BlockSize;
        if (pPip->pOut==pPip->pMax) {
            pPip->pOut = &pPip->Data;
        }
        entriesDone++;
    }
    return entriesDone;
}

bool qPipGetWrdFast(pPIP pPip, unsigned* pData) {
    unsigned entriesDone=0;
    if (pPip->Entries > 0) {
        *pData = (unsigned)*pPip->pOut;
        pPip->pOut += 4;
        if (pPip->pOut==pPip->pMax) {
            pPip->pOut = &pPip->Data;
        }
        entriesDone = 1;
    }
    return entriesDone;
}

bool qPipGetBytFast(pPIP pPip, uint8_t* pData) {
    unsigned entriesDone=0;
    if (pPip->Entries > 0) {
        *pData = (uint8_t)*pPip->pOut;
        qWrdDecAtomic(&pPip->Entries);
        pPip->pOut += 1;
        if (pPip->pOut==pPip->pMax) {
            pPip->pOut = &pPip->Data;
        }
        entriesDone = 1;
    }
    return entriesDone;
}


