/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  **
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  **
 **  <http://www.quasarsoft.com/license.html>                                       **
 **                                                                                 **
 *************************************************************************************/

# Bit numbers in qvKernel
__InitDone = 0
__OsStarted = 1
__StatisticsOn = 2
__SwitchNotificationOn = 3
__USecOn = 4
__PriorityCheckOn = 5
__TimerInterruptOn = 6
__RtccInterruptOn = 7

# Bit numbers in qvSpawn
__FiberPrio4 = 0
__FiberPrio3 = 1
__FiberPrio2 = 2
__FiberPrio1 = 3
__FiberReserved4 = 4
__FiberTimerExpire = 5
__FiberReserved6 = 6
__FiberRtccExpire = 7
__FiberQueued = 8

.set noreorder
.set noat


.extern qvTotalTMR
.extern qvSpawn
.extern qvCycles


__t0   = 0
__t1   = 4
__t2   = 8
__EPC  = 12
__SR   = 16
__size = 20

__CLR = 0x4
__SET = 0x8
__INV = 0xC


.text
.global yTimerInit
yTimerInit:
	lui		$t0, 0xBF88 						# Virtual high address part (0xBF88) for everything interrupt
	lui		$t1, 0xBF80         				# Virtual high address part (0xBF80) for everything timers
	lui		$t2, 0x10							### bitnumber TMR5 (0001.0000.0000.0000.0000)

	sw		$zero, %lo(T4CON)($t1)				###
	sw		$zero, %lo(T5CON)($t1)				###
	sw		$zero, %lo(TMR4)($t1)				###

	addiu	$t4, $zero, 0x8 
	sw		$t4, %lo(T4CONSET)($t1)				### and set the 32 mode bit
	
	addi	$t4, $zero, -1
	lw		$t4, %lo(PR4)($t1)					###

	addiu	$t4, $zero, 0x9						### priority 2 sub-priority 1 
	sw		$t4, %lo(IPC5SET)($t0)				###and set the 32 mode bit 
	 
	sw		$t2, %lo(IFS0CLR)($t0)				### Clear the interrupt signal 
	sw		$t2, %lo(IEC0SET)($t0)				### Interrupt enable
	jr		$ra
	nop

.section .vector_20, code						### Is interrupt 5 for tmr4-5
.global q_TIMER_5_VECTOR						###
q_TIMER_5_VECTOR:								###
	j		TimerInterrupt
	addiu	$sp, $sp, -__size

.text
TimerInterrupt:
	mfc0	$k1, $14							# Copy EPC in k1
	mfc0	$k0, $13							# Cause in k0 
	sw		$k1, __EPC($sp)						# Save EPC 
	mfc0	$k1, $12							# Copy STATUS in k1
	sw		$k1, __SR($sp)						# Save STATUS 
	srl		$k0, $k0, 0xa						# shift RIPL to bit 0
	ins		$k1, $k0, 10, 6						# Insert RIPL in IPL 
	ins		$k1, $zero, 1, 4					# Clear ERL and EXL (will enable interrupts) 
	mtc0	$k1, $12							# Store in Status (IPL now RIPL) 
												# Interrupts now enabled 
	sw		$t0, __t0($sp)
	sw		$t1, __t1($sp)
	sw		$t2, __t2($sp)

	la		$t1, PR4							### Get the address of PR4
	lw		$t2, 0($t1)							# Get PR  Save it for later
	lw		$t0, %gp_rel(qvTotalTMR)($gp)		# get qvTotalTMR
	addu	$t0, $t2, $t0						# add qvTotalTMR and PR 
	addiu	$t0, $t0, 2							# add 2 
	sw		$t0, %gp_rel(qvTotalTMR)($gp)		# and store back in the qvTotalTMR
	addi	$t0, $zero, -1						# k0 = -1 
	sw		$t0, 0($t1)							# Set PR to 0xffffffff
#
# Adding cycles to qvCycles and 
# t2 = PR register 
#
# qvCycles += PR
	lw		$t1, %gp_rel(qvCycles)($gp)			# Get cycles until now 
	lw		$t0, %gp_rel(qvCycles+4)($gp)		# Full 64 bits 
	addu	$t1, $t1, $t2						# Add PR register to this
	sw		$t1, %gp_rel(qvCycles)($gp)			# Store LSW cycles 
	sltu	$t1, $t1, $t2						# t1 = 1 if carry 
	addu	$t0, $t0, $t1
	sw		$t0, %gp_rel(qvCycles+4)($gp)		# Store MSW cycles
#
# Exit the interrupt
# Clear the interrupt, restore registers and set the __TimerInterruptOn flag
#
10:	lui		$t1, 0x10							### TMR bit
	la		$t0, IFS0
	sw		$t1, __CLR($t0)						# and clear the interrupt signal

15:	ll		$t0, %gp_rel(qvKernel)($gp)			# Set bit 1 in qvKernel
	ori		$t0, 1<<__TimerInterruptOn			# Interrupts are disabled so we can do this ...
	sc		$t0, %gp_rel(qvKernel)($gp)			# ... Without using the disabling interrupts
    beq     $t0, $zero, 15b
    nop

18:	ll		$t0, %gp_rel(qvSpawn)($gp)			#
	ori		$t0, 1<<__FiberTimerExpire 			# Set bit 5 in qvSpawn xFbrTimerExpire
	sc		$t0, %gp_rel(qvSpawn)($gp)			# ... Without using the disabling interrupts
    beq     $t0, $zero, 18b

	lw		$t0, __t0($sp)						# restore the registers
	lw		$t1, __t1($sp)
	lw		$t2, __t2($sp)

    di
    ehb
    addiu	$k0, $zero, 2						# The bit to set the qKernel interrupt
	la		$k1, IFS0							# Get interrupt register address
	sw		$k0, __SET($k1)						# and set the qKernel interrupt signal 

	lw		$k0, __EPC($sp)						# Restore EPC
	mtc0	$k0, $14
	lw		$k0, __SR($sp)						# Restore SR
	mtc0	$k0, $12
	addiu	$sp, $sp, __size
	eret




#
# INTU yTimerStop();
#
.global yTimerStop
yTimerStop:
	lui		$t0, 0xBF88 						# Virtual high address part (0xBF88) for everything interrupt
	lui		$t1, 0xBF80             			# Virtual high address part (0xBF80) for everything timers
	lui		$t2, 0x10							### bitnumber TMR5

	sw		$t2, %lo(IEC0CLR)($t0)				# Disable interrupt 

	addi	$t4, $zero, 0x8000					# Stop Timer Bit
	sw		$t4, %lo(T4CONCLR)($t1)				### Stop Timer 

	sw		$zero, %gp_rel(qvCycles)($gp)
	sw		$zero, %gp_rel(qvCycles+4)($gp)
	sw		$zero, %gp_rel(qvTotalTMR)($gp)		# clear Total Tmer
	sw		$zero, %lo(TMR4)($t1)				### clear TMR
	addi	$t4, $zero, -1
	sw		$t4, %lo(PR4)($t1)					### set PR to  0xffffffff

	sw		$t2, %lo(IFS0CLR)($t0)				# Clear interrupt flag 
	jr		$ra
	sw		$t3, %lo(IEC0SET)($t0)				# Enable interrupt 

#
# INTU yTimerReshuffleNeeded();
#
.global yTimerReshuffleNeeded
yTimerReshuffleNeeded:
	lw		$v0, %gp_rel(qvTotalTMR)($gp)		# Get qvTotalTMR 
	ext		$v0, $v0, 31, 1						# Extract bit 31
	bne		$v0, $zero, 10f
	nop
	jr		$ra									# v0 was already 0
    nop

10:	lui		$t0, 0xBF88     					# Virtual high address part (0xBF88) for everything interrupt
	lui		$t1, 0x10							### bitnumber TMR5
    sw		$t1, %lo(IEC0CLR)($t0)				# Enable interrupt
	jr		$ra
	addiu	$v0, $zero, 1						# return 1

#
# void yTimerReshuffleEnd();
#
.global yTimerReshuffleEnd
yTimerReshuffleEnd:
	lui		$t0, 0xBF88							# Virtual high address part (0xBF88) for everything interrupt 
	lui		$t2, 0x10							### bitnumber TMR5

# Load/Store no problem because timer interrupt is disabled
	lw		$t4, %gp_rel(qvTotalTMR)($gp)		# Get qvTotalTMR 
	ins		$t4, $zero, 31, 1					# clear bit 31 
	sw		$t4, %gp_rel(qvTotalTMR)($gp)		# Write back 
	
	jr		$ra
	sw		$t2, %lo(IEC0SET)($t0)				# Enable interrupt 

#
# INT32U yTimerGetTime();
#
.global yTimerGetTime
1:	nop											# Some time for the interrupt to start
	nop
	nop
	nop
yTimerGetTime:
	lui		$t0, 0xBF88							# Virtual high address part (0xBF88) for everything interrupt 
	lui		$t1, 0xBF80							# Virtual high address part (0xBF80) for everything timers
	lui		$t2, 0x10							### bitnumber TMR5

	sw		$t2, %lo(IEC0CLR)($t0)				# Disable interrupt
	lw		$v0, %lo(TMR4)($t1)					### Get the timer 
	lw		$t4, %gp_rel(qvTotalTMR)($gp)		# Get qvTotalTMR
	addu	$v0, $v0, $t4						# Add to the timer value
	lw		$t4, %lo(IFS0)($t0)					# get interrupt signal
	and		$t4, $t4, $t2						# Keep the IFS bit 
	bne		$t4, $zero, 1b						# Do all again if interrupt was set
	sw		$t2, %lo(IEC0SET)($t0)				# Enable interrupt 

	jr		$ra
	nop
#
# a0=Alarm TIME
# void yTimerSetAlarm(INT32U TimeAlarm);
#
.global yTimerSetAlarm
1:	nop											# Some time for the interrupt to start
	nop
	nop
	nop
yTimerSetAlarm:
	lui		$t0, 0xBF88							# Virtual high address part (0xBF88) for everything interrupt 
	lui		$t1, 0xBF80							# Virtual high address part (0xBF80) for everything timers
	lui		$t2, 0x10							### bitnumber TMR5 interrupt

	sw		$t2, %lo(IEC0CLR)($t0)				# Disable interrupt

	lw		$t4, %gp_rel(qvTotalTMR)($gp)		# Get qvTotalTMR 
	subu	$t4, $a0, $t4						# t4 = newtime - TotalTMR
	sw		$t4, %lo(PR4)($t1)					# Set the new time

    lw      $t5, %lo(TMR4)($t1)
    addiu   $t5, 1000                           # It is possible that we are hit by an interrupt
    sltu    $t5, $t5, $t4                       # We add 1000 cycles for that and that should be enough
    bne     $t5, $zero, 8f                      # if difference is big skip setting
    nop

3:	ll		$t4, %gp_rel(qvSpawn)($gp)			# get qvSpwan
	ori		$t4, 1<<__FiberTimerExpire          # set the fibertimer
 	sc		$t4, %gp_rel(qvSpawn)($gp)			# and write back
    beq     $t4, $zero, 3b

	addiu	$t4, $zero, 2						# the kernel interrupt bit
	sw		$t4, %lo(IFS0SET)($t0)				# set the bit

8:	lw		$t4, %lo(IFS0)($t0)					# get interrupt signal
	and		$t4, $t4, $t2						# Keep the IFS bit 
	bne		$t4, $zero, 1b						# Do all again if interrupt was set
	sw		$t2, %lo(IEC0SET)($t0)				# Enable interrupt Also delay slot


	addi	$t4, $zero, 0x8000					# Timer on/off bit 
	sw		$t4, %lo(T4CONSET)($t1)				### Start Timer in case it was off 

	jr		$ra
	nop
	

.global yGetCycles
1:
	nop											# Some time for the interrupt to start
	nop
	nop
	nop
yGetCycles:
	lui		$t0, 0xBF88							# Virtual high address part (0xBF88) for everything interrupt 
	lui		$t1, 0xBF80							# Virtual high address part (0xBF80) for everything timers
	lui		$t2, 0x10							### bitnumber TMR5

	sw		$t2, %lo(IEC0CLR)($t0)				# Disable interrupt

	lw		$t4, %lo(TMR4)($t1)					### Get the timer 
	lw		$v0, %gp_rel(qvCycles+0)($gp)
	lw		$v1, %gp_rel(qvCycles+4)($gp)
	addu	$v0, $v0, $t4
	sltu	$t4, $v0, $t4						# see if any carry
	addu	$v1, $v1, $t4						# result in v0-v1 
	
	lw		$t4, %lo(IFS0)($t0)					# get interrupt signal
	and		$t4, $t4, $t2						# Keep the IFS bit 
	bne		$t4, $zero, 1b						# Do all again if interrupt was set
	sw		$t2, %lo(IEC0SET)($t0)				# Enable interrupt 

	jr		$ra
	nop


