/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"

static const uint16_t qcMonthTotal[] = {0,0,31,59,90,120,151,181,212,243,273,304,334};
static const uint8_t qcMonthDays[] = {0,31,28,31,30,31,30,31,31,30,31,30,31};
static const uint8_t qcBcd2Bin[] =  {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
                                     0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x10, 0x11, 0x12, 0x13, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
                                     0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
                                     0x1E, 0x1F, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
                                     0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F, 0x30, 0x31, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
                                     0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
                                     0x3C, 0x3D, 0x3E, 0x3F, 0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
                                     0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
                                     0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
                                     0x5A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F, 0x60, 0x61, 0x62, 0x63};
static const uint8_t qcBin2Bcd[] =  {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09,
                                     0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19,
                                     0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29,
                                     0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
                                     0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
                                     0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
                                     0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
                                     0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
                                     0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
                                     0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99};

uint32_t qDtmAddDays(uint32_t DatTim, int Day) {
    return DatTim + (int32_t)Day*3600*24;
}


uint32_t qDtmAddHours(uint32_t DatTim, int Hour) {
    return (uint32_t)DatTim + (int32_t)Hour*3600;
}


uint32_t qDtmAddMinutes(uint32_t DatTim, int Min) {
    return DatTim + (int32_t)Min*60;
}


uint32_t qDtmAddSeconds(uint32_t DatTim, int Sec) {
    return DatTim + (int32_t)Sec;
}


uint32_t qDtmAddYears(uint32_t DatTim, int Year) {
    BCD_DTM dt;
    unsigned tmp;
    // Ignore the warning "Taking the address of 'dt' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    qDtmToBcdDtm(DatTim, &dt);
    tmp = qcBcd2Bin[dt.year];
    Year += tmp;
    dt.year = qcBin2Bcd[Year];
    // Ignore the warning "Taking the address of 'dt' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    return qDtmFromBcdDtm(&dt);
}


//Sunday=0, Monday=1, Tuesday=2, Wednessday=3, Thursday=4, Friday=5, and Saterday=6.
unsigned qDtmDayOfWeek(uint32_t DatTim) {
    uint32_t tmp;
    unsigned days;
    unsigned day;
    uint16_t dummy;

    // we have to divide by 24*3600 to get the number of days. We prefer a divide by 
    // a 16bit number so we first divide by 12*3600. The result is then 2 times to big but
    // it will fit in a 16-bit number. By shifting it to the right we get the correct value 
    // in a 16bit variable.
    // then we divide by 7 (days per week) and adjust it against 2000/1/1 

    // Ignore the warning "Taking the address of 'dummy' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    tmp = qDivU3216R(DatTim, (uint16_t)(12u*3600u), &dummy);  // Result is 2 time to big
    days = (uint16_t)tmp;                       // only keep the LSW
    days = days>>1;                             // and divide by 2
    day = (days%7) + 6;                         // 2000/1/1 is Saturday (6)
    if (day>=7) day -= 7;
    return day;
}


uint32_t qDtmFromBcdDtm(pBCD_DTM dt) {
    unsigned days;
    unsigned leap;
    unsigned tmp;
    uint32_t time;
    unsigned sec;
    unsigned min;
    unsigned hour;
    unsigned mday;
    unsigned mon;
    unsigned year;

    if (dt->sec > 0x59) return 0;
    sec = qcBcd2Bin[dt->sec];
    if (sec == 0xff) return 0;
    
    if (dt->min > 0x59) return 0;
    min = qcBcd2Bin[dt->min];
    if (min == 0xff) return 0;
    
    if (dt->hour > 0x23) return 0;
    hour = qcBcd2Bin[dt->hour];
    if (hour == 0xff) return 0;
    
    if ((dt->mday > 0x31) || (dt->mday==0)) return 0;
    mday = qcBcd2Bin[dt->mday];
    if (mday == 0xff) return 0;
    
    if ((dt->mon > 0x12) || (dt->mon==0)) return 0;
    mon = qcBcd2Bin[dt->mon];
    if (mon == 0xff) return 0;
    
    if (dt->year > 0x99) return 0;
    year = qcBcd2Bin[dt->year];
    if (year == 0xff) return 0;

    days = qcMonthDays[mon];
    leap = 0;
    if ((year & 0x3)==0) leap++;
    if (mday > (days + leap))  return 0;
    tmp = sec + (min*60);
    time = (uint32_t)tmp + (((uint32_t)(hour)) * 3600);
    if (leap && mon <= 2) leap--;
    time += ((uint32_t)(qcMonthTotal[mon]+ mday + leap - 1)) * 24ul * 3600ul;
    tmp = (year+3)>>2;                            // number of leed days
    tmp += (unsigned)year*365;                    // add days based on 365
    time += ((uint32_t)tmp * 24ul) * 3600ul;
    return time;
}


uint32_t qDtmFromYMDHMS(uint8_t Year, uint8_t Month, uint8_t Day, uint8_t Hour, uint8_t Minute, uint8_t Second) {
    BCD_DTM dt;
    dt.sec = qcBin2Bcd[Second];
    dt.min = qcBin2Bcd[Minute];
    dt.hour = qcBin2Bcd[Hour];
    dt.mday = qcBin2Bcd[Day];
    dt.mon = qcBin2Bcd[Month];
    dt.year = qcBin2Bcd[Year];
    // Ignore the warning "Taking the address of 'dt' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    return qDtmFromBcdDtm(&dt);
}    


void qDtmToBcdDtm(uint32_t DatTim, pBCD_DTM bcdDatTim) {
    uint32_t tmp;
    unsigned dat;
    unsigned days;
    uint16_t sec;
    uint16_t min;
    uint16_t hour;
    unsigned mday;
    unsigned mon;
    unsigned year;
    unsigned day;
    
    // Ignore the warning "Taking the address of 'sec' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    tmp = qDivU3216R(DatTim, 60, &sec);
    bcdDatTim->sec = qcBin2Bcd[sec];
    // Ignore the warning "Taking the address of 'min' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    tmp = qDivU3216R(tmp, 60, &min);
    bcdDatTim->min = qcBin2Bcd[min];
    // Ignore the warning "Taking the address of 'hour' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    tmp = qDivU3216R(tmp, 24, &hour);
    bcdDatTim->hour = qcBin2Bcd[hour];

    dat = tmp;                                  // dat is the number of days since 2000/1/1
    year = 0;                                   // Start with 2000
    days = 366;                                 // Leap year
    while (dat >= days) {
        dat -= 365;
        days = 365;
        if ((year++ & 0x03)== 0) dat--;
        if ((year & 0x03)== 0) days++;
    }
    mon = 1;                                    // Start with Jan
    days = 31;                                  // Jan is 31 days
    while (dat > days) {
        dat -= days;
        days = qcMonthDays[++mon];
        if (days==28 && (year && 3)==0) days++;
    }
    mday = ++dat;
    
    bcdDatTim->mday = qcBin2Bcd[mday];
    bcdDatTim->mon = qcBin2Bcd[mon];
    bcdDatTim->year = qcBin2Bcd[year];
    
    day = (((uint16_t)tmp)%7) + 6;              // tmp contains the number of days sinse 2000/1/1
    if (day >= 7) day -= 7;                     // See qDtmDayOfWeek for a description of this process
    bcdDatTim->wday = day;
}
