/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


unsigned qEvtWaitTO(pEVT pEvt, unsigned Events, WAIT_TYPE WaitType, int32_t TimeOut) {
    unsigned events;
    unsigned done;


#ifndef __NO_CHECK__
    if (xTestWait(pEvt, xTYPE_EVT, qERR_EVT_ID, TimeOut))
        return 0;
    if (!Events) {
        xErrNtf(qERR_EVT_NO_FLAGS);
        return 0;
    }
    if (WaitType < WAIT_TYPE_ALL || WaitType > WAIT_TYPE_ANY_AND_CLEAR) {
        xErrNtf(qERR_EVT_WAIT_TYPE);
        return 0;
    }
#endif

    qCrtEnter();
    done = 0;
    events = pEvt->CurEvents&Events;            // Compare the results and save
    if (WaitType&1) {                           // OR case in Bit 0
        if (events) done=1;
    }
    else {                                      // AND case
        if (events==Events) done=1;
    }
    if (done) {
        if (WaitType&2) {                       // Clear in Bit 1
            pEvt->CurEvents &= ~events;         // clear the events that were set
        }
        qCrtExit();
        return events;
    }
    qvCurTcb->WaitEvtEvents = Events;
    qvCurTcb->Status &= ~xTCB_WAIT_TYPE_OR;
    qvCurTcb->Status &= ~xTCB_WAIT_AND_CLEAR;
    if (WaitType&1) qvCurTcb->Status |= xTCB_WAIT_TYPE_OR;
    if (WaitType&2) qvCurTcb->Status |= xTCB_WAIT_AND_CLEAR;
    
    xThrTimeoutOnObject((pOBJ)pEvt,TimeOut);    // and wait on the time-out
                                                // EXIT_CRITICAL will switch the thread because 
                                                // the wait-time has set the kernel interrupt
    if (qvCurTcb->TimedOut) {                   // Did we time-out on a wait?
        return 0;                               // Time-out error
    }
    return qvCurTcb->WaitEvtEvents;             // returns the events that readied thread
}
