/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


//         +------------------------+
//     +-- |  pFirstBlk             |
//     |   |  NextPool              |
//     |   |  Type/Filler           |
//     |   |  BlockSize (10)        | Usable memory size (Size block with overhead is 14)
//     |   |  Entries (4)           |
//     |   |  BitTable (1 integer)  | <-- This is variable 1 int for 16 2 int for 32 etc 
//     |   +------------------------+
//     +-> |  Entry number          | <--- pBlk points here
//         |  InUse + pnt BitTable  |
//         |  Memory                | <--- Pointer given back 
//         |  Memory                |
//         |  Memory                |
//         |  Memory                |
//         |  Memory                |
//         +------------------------+
//         |  Entry number          |
//         |  InUse + pnt BitTable  |
//         |  Memory                | <--- Pointer given back
//         |  Memory                |
//         |  Memory                |
//         |  Memory                |
//         |  Memory                |
//         +------------------------+
//         |  Entry number          |
//         |  InUse + pnt BitTable  |
//         |  Memory                | <--- Pointer given back 
//         |  Memory                |
//         |  Memory                |
//         |  Memory                |
//         |  Memory                |
//         +------------------------+
//         |  Entry number          |
//         |  InUse + pnt BitTable  |
//         |  Memory                | <--- Pointer given back
//         |  Memory                |
//         |  Memory                |
//         |  Memory                |
//         |  Memory                |
//         +------------------------+
//


pFPL qFixCreate(unsigned Size, unsigned NbrBlocks) {
    // size + space for entries and pFPL
    unsigned blockSize = xRoundToInteger(Size);    
    unsigned totalMem = (blockSize+sizeof(unsigned)+sizeof(unsigned))*NbrBlocks;
    unsigned totalBitTable;
    unsigned totalFix;
    unsigned total;
    unsigned* pBits;
    int n;
    
#if qBIT_SIZE==16
    totalBitTable = ((NbrBlocks+15)>>4)<<1;
#else
    totalBitTable = ((NbrBlocks+31)>>5)<<2;    
#endif
    totalFix = sizeof(struct sFPL) + totalBitTable;        // pFPL has the extra word
    total = totalMem + totalFix;

    pFPL p = qHeaAlloc(total);
    if (p) {
        p->pFirstBlk = (unsigned*)((unsigned)p+totalFix);
        p->Total = NbrBlocks;
        p->pNextPool = qvFixPoolList;
        qvFixPoolList = p;
        p->BlockSize = blockSize;
        p->Type = xTYPE_FPL;
        pBits = &p->BitTable[0];
        for (n=0; n<(totalBitTable/sizeof(int)); n++) {
            *pBits++=~0;
        }
    }
    return (pFPL)p;
}


void qFixClose(pFPL pFix) {
    qMemFree(pFix);
}

