/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  


// This function is new and is delivered as until th next release. If there is a problem 
// please inform QuasarSoft Ltd.

#include "qKernel.h"
#include <stdarg.h>
//#include <stdio.h>

#define INT16_BUF_LEN         (7)     // 7  (unsigned 5 decimals) (signed 5 decimals + sign) + terminator
#define INT32_BUF_LEN         (11)    // 11 (unsigned 10 decimals) (signed 9 decimals + sign) + terminator
#define INT64_BUF_LEN         (21)    // 21 (unsigned 20 decimals) (signed 19 decimals + sign) + terminator

const char _Zero[] = {'0', 0};
const char _HexUpper[] = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
const char _HexLower[] = {'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};


//__attribute__((weak,far))
void qFmtPutChar(int c) {
    xErrNtf(qERR_PRINTF_PUTCHAR);
    return;
}
__attribute__((weak, far))
void qFmtWait() {
    xErrNtf(qERR_PRINTF_PUTWAIT);
    return;
}

static void xSaveChar(char **str, int c) {
    if (str==NULL) {
        qFmtPutChar(c);
        return;
    }
    **str = c;
    (*str)++;
}

static unsigned xFmtString(char* *Buf, char *String, unsigned StrLen, unsigned Width, bool Right) {
    unsigned count = 0;
    char pad = ' ';

    if (Width > 0) {
        if (StrLen >= Width)
            Width = 0;
        else
            Width -= StrLen;
    }
    if (!Right) {
        for (; Width > 0; --Width) {
            xSaveChar (Buf, pad);
            count++;
        }
    }
    for (; *String ; ++String) {
        xSaveChar(Buf, *String);
        count++;
    }
    for (; Width > 0; --Width) {
        xSaveChar(Buf, pad);
        count++;
    }
    return count;
}


static unsigned xFmtINT16(char* *Buf, int16_t Value, bool Sign, unsigned Width) {
    char tmp[INT16_BUF_LEN];
    unsigned len;
    char* p;
    unsigned t;
    bool neg = false;
    unsigned count = 0;
    unsigned uValue = Value;

    if (Value == 0) {
        return xFmtString (Buf, (char*)_Zero, Width, 1, true);
    }

    if (Sign==true && Value < 0) {
        neg = true;
        uValue = -Value;
    }

    p = tmp + INT16_BUF_LEN;
    *--p = 0;

    while (uValue) {
        t = uValue % 10;
        uValue /= 10;
        *--p = t + '0';
    }

    if (neg) {
        *--p = '-';
    }
    len = tmp + INT16_BUF_LEN - p -1;
    return count + xFmtString(Buf, p, len, Width, true);
}


static unsigned xFmtINT32(char* *Buf, int32_t Value, bool Sign, unsigned Width) {
    char tmp[INT32_BUF_LEN];
    unsigned len;
    char *p;
    uint16_t t;
    bool neg = false;
    unsigned count = 0;
    uint32_t uValue = Value;

    if (Value == 0) {
        return xFmtString(Buf, (char*)_Zero, 1, Width, true);
    }

    if (Sign==true && Value < 0) {
        neg = true;
        uValue = -Value;
    }

    p = tmp + INT32_BUF_LEN;
    *--p = 0;

    while (uValue) {
    // Ignore the warning "Taking the address of 't' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
        uValue = qDivU3216R(uValue, 10, &t);
        *--p = t + '0';
    }

    if (neg) {
        *--p = '-';
    }
    len = tmp + INT32_BUF_LEN - p - 1;
    return count + xFmtString(Buf, p, len, Width, true);
}



static unsigned xFmtINT64(char* *Buf, int64_t Value, bool Sign, unsigned Width) {
    char tmp[INT64_BUF_LEN];
    unsigned len;
    char *p;
    uint16_t t;
    bool neg = false;
    unsigned count = 0;
    uint64_t uValue = Value;

    if (Value == 0) {
        return xFmtString(Buf, (char*)_Zero, 1, Width, true);
    }

    if (Sign==true && Value < 0) {
        neg = true;
        uValue = -Value;
    }

    p = tmp + INT64_BUF_LEN;
    *--p = 0;

    while (uValue) {
    // Ignore the warning "Taking the address of 't' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
        uValue = qDivU6416R(uValue, 10, &t);
        *--p = t + '0';
    }

    if (neg) {
        *--p = '-';
    }
    len = tmp + INT64_BUF_LEN - p - 1;
    return count + xFmtString(Buf, p, len, Width, true);
}

static unsigned xFmtINT16Hex(char* *Buf, uint16_t Value, bool UpperCase) {
    char tmp[4];
    char *p;
    unsigned t;
    int n;

    p = tmp + 4;
    for (n=0; n<4; n++) {
        t = Value & 0xf;
        if (UpperCase) {
            *--p = _HexUpper[t];
        }
        else {
            *--p = _HexLower[t];
        }
        Value = Value>>4;
    }
    return xFmtString(Buf, p, 4, 4, false);
}

static unsigned xFmtINT32Hex(char* *Buf, uint32_t Value, bool UpperCase) {
    char tmp[8];
    char *p;
    unsigned t;
    int n;

    p = tmp + 8;
    for (n=0; n<8; n++) {
        t = Value & 0xf;
        if (UpperCase) {
            *--p = _HexUpper[t];
        }
        else {
            *--p = _HexLower[t];
        }
        Value = Value>>4;
    }
    return xFmtString(Buf, p, 8, 8, false);
}

static unsigned xFmtINT64Hex(char* *Buf, uint64_t Value, bool UpperCase) {
    char tmp[16];
    char *p;
    unsigned t;
    int n;

    p = tmp + 16;
    for (n=0; n<16; n++) {
        t = Value & 0xf;
        if (UpperCase) {
            *--p = _HexUpper[t];
        }
        else {
            *--p = _HexLower[t];
        }
        Value = Value>>4;
    }
    return xFmtString(Buf, p, 16, 16, false);
}


static unsigned xFormat(char* *Buf, char* Format, va_list args) {
    unsigned count = 0;

    while (*Format) {                       // do while there is data
        if (*Format == '%') {               // Special char
            bool right = false;
            unsigned longSize = 0;
            unsigned width = 0;
            Format++;                       // get next char
            if (*Format == 0) break;        // All done
            if (*Format == '%') {           // if %%
                xSaveChar(Buf, *Format);    // save the %
                count++;                    // One done
                Format++;                   // position next char
                continue;
            }
            if (*Format == '-') {           // %-
                Format++;                   // Get next char
                if (*Format == 0) break;    // All done
                right = true;               // right allign
            }
            // Get the numbers out for the width.
            for (; *Format >= '0' && *Format <= '9'; Format++) {
                width *= 10;
                width += *Format - '0';
            }
            if (*Format == 'l') {           // Is it long (32-bit)
                longSize = 1;
                Format++;
                if (*Format == 'l') {       // or longlong (64-bit)
                    longSize = 2;
                    Format++;
                }
            }
            if (*Format == 'd') {           // %d, %ld, %lld or with a width
                switch (longSize) {         // signed is true
                    case 1:
                        count += xFmtINT32(Buf, va_arg(args, int32_t), true, width);
                        break;
                    case 2:
                        count += xFmtINT64(Buf, va_arg(args, int64_t), true, width);
                        break;
                    default:
                        count += xFmtINT16(Buf, va_arg(args, int), true, width);
                        break;
                }
                Format++;
                continue;
            }
            if (*Format == 'u') {
                switch (longSize) {         // u so no sign
                    case 1:
                        count += xFmtINT32(Buf, va_arg(args, int32_t), false, width);
                        break;
                    case 2:
                        count += xFmtINT64(Buf, va_arg(args, int64_t), false, width);
                        break;
                    default:
                        count += xFmtINT16(Buf, va_arg(args, unsigned), false, width);
                        break;
                }
                Format++;
                continue;
            }
            if (*Format == 'x') {           // all lower case
                switch (longSize) {
                    case 1:
                        count += xFmtINT32Hex(Buf, va_arg(args, int32_t), false);
                        break;
                    case 2:
                        count += xFmtINT64Hex(Buf, va_arg(args, int64_t), false);
                        break;
                    default:
                        count += xFmtINT16Hex(Buf, va_arg(args, int), false);
                        break;
                }
                Format++;
                continue;
            }
            if (*Format == 'X') {           // all upper case
                switch (longSize) {
                    case 1:
                        count += xFmtINT32Hex(Buf, va_arg(args, int32_t), true);
                        break;
                    case 2:
                        count += xFmtINT64Hex(Buf, va_arg(args, int64_t), true);
                        break;
                    default:
                        count += xFmtINT16Hex(Buf, va_arg(args, int), true);
                        break;
                }
                Format++;
                continue;
            }
            if (*Format == 's') {           // Is it a string?
                char* s = (char*)va_arg(args, int);
                if (s==0) {                 // Is it a null string
                    count += xFmtString(Buf, "(null)", 6,  width, right);
                }
                else {
                    char* p = s;            // Otherwise get the lenght
                    unsigned len = 0;
                    while (*p++) len++;
                    count += xFmtString(Buf, s, len, width, right);
                }
                Format++;
                continue;
            }
            if (*Format == 'c') {
                char tmp[2];
                tmp[0] = (char)va_arg(args, int);
                tmp[1] = 0;
                count += xFmtString(Buf, tmp, 1, width, false);
                Format++;
                continue;
            }
        }
        else {
            xSaveChar(Buf, *Format++);
            count++;
        }
    }
    if (Buf) {
        **Buf = 0;
    }
    va_end(args);
    return count;
}

unsigned int qsprintf(char* buf, const char* format, ...) {
    va_list args;
    va_start(args, format);
    // Ignore the warning "Taking the address of 'buf' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    return xFormat(&buf, (char*)format, args);
}

void qprintf(const char* format, ...) {
    va_list args;
    va_start(args, format);
    xFormat(NULL, (char*)format, args);
    qFmtWait();
}


