/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  **
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  **
 **  <http://www.quasarsoft.com/license.html>                                       **
 **                                                                                 **
 *************************************************************************************/

#ifndef QKERNEL_COMMON_H
#define QKERNEL_COMMON_H

#ifndef qBREAKPOINT
    #error qBREAKPOINT must be defined
#endif
#ifndef qBIT_SIZE
	#error qBIT_SIZE must be defined
#endif
#ifndef qSTACK_GROWS_LOW_TO_HIGH
	#error qSTACK_GROWS_LOW_TO_HIGH must be defined
#endif
#ifndef qLITTLE_ENDIAN
    #error qLITTLE_ENDIAN must be defined
#endif

#define qMSEC(ms) ((int32_t)(ms)*(int32_t)1000)
#define qSEC(s) (0-(s))

#define _xInitDone                0
#define _xOsStarted               1
#define _xStatisticsOn            2
#define _xSwitchNotificationOn    3
#define _xUSecOn                  4
#define _xPriorityCheckOn         5
#define _xTimerInterruptOn        6
#define _xRtccInterruptOn         7
#define _xEdsSupported            8
#define _xInsertEndReadyList      9

#define xKrnIsInitDone()                    (qvKernel & (1<<_xInitDone))
#define xKrnIsOsStarted()                   (qvKernel & (1<<_xOsStarted))
#define xKrnIsStatisticsOn()                (qvKernel & (1<<_xStatisticsOn))
#define xKrnIsUSecOn()                      (qvKernel & (1<<_xUSecOn))
#define xKrnIsPriorityCheckOn()             (qvKernel & (1<<_xPriorityCheckOn))
#define xKrnIsTimerInterruptOn()            (qvKernel & (1<<_xTimerInterruptOn))
#define xKrnIsRtccInterruptOn()             (qvKernel & (1<<_xRtccInterruptOn))
#define xKrnIsEdsSupported()                (qvKernel & (1<<_xEdsSupported))
#define xKrnIsInsertEndReadyList()          (qvKernel & (1<<_xInsertEndReadyList))

#define xKrnSetInitDone()                   (qvKernel |= (1<<_xInitDone))
#define xKrnSetOsStarted()                  (qvKernel |= (1<<_xOsStarted))
#define xKrnSetStatisticsOn()               (qvKernel |= (1<<_xStatisticsOn))
#define xKrnSetPriorityCheckOn()            (qvKernel |= (1<<_xPriorityCheckOn))
#define xKrnSetUSecOn()                     (qvKernel |= (1<<_xUSecOn))
#define xKrnSetEdsSupported()               (qvKernel |= (1<<_xEdsSupported))
#define xKrnSetInsertEndReadyList()         (qvKernel |= (1<<_xInsertEndReadyList))

#define xKrnClrPriorityCheckOn()            (qvKernel &= ~(1<<_xPriorityCheckOn))
#define xKrnClrStatisticsOn()               (qvKernel &= ~(1<<_xStatisticsOn))
#define xKrnClrTimerInterruptOn()           (qvKernel &= ~(1<<_xTimerInterruptOn))
#define xKrnClrRtccInterruptOn()            (qvKernel &= ~(1<<_xRtccInterruptOn))
#define xKrnClrUSecOn()                     (qvKernel &= ~(1<<_xUSecOn))
#define xKrnClrInsertEndReadyList()         (qvKernel &= ~(1<<_xInsertEndReadyList))

#define xVER_MAJOR              6
#define xVER_MINOR              0
#define xVER_BUILD              3361
#define DTM_1_1_2010            315619200
#define xTCB_WAIT_TYPE_OR       0x04
#define xTCB_WAIT_AND_CLEAR     0x08
#define xTMR_PERIODIC           0x01
#define xTMR_DELETE             0x02
#define xTMR_STOP               0x04

#define xTHR_EVT_WAITING        (struct sOBJ *)(1)

/*
    The types for the RTOS objects. If the MSB is set then the object is opened but not created
*/

#define xTYPE_TCB               0x11    //
#define xTYPE_TMR               0x21    //
#define xTYPE_EVT               0x31    // All #30 objects with a size of 10 or 20 (32bit)
#define xTYPE_MTX               0x32
#define xTYPE_SEM               0x33
#define xTYPE_PUB               0x34
#define xTYPE_QUE               0x41    // Variable size
#define xTYPE_PIP               0x42    // Variable size
#define xTYPE_MPL               0x61    //
#define xTYPE_FPL               0x62    //
#define xTYPE_MPL_MSG           0x01    // Short type used in MSG
#define xTYPE_FPL_MSG           0x02    // Short type used in MSG

#define xTYPE_OPN_EVT           xTYPE_EVT+0x80
#define xTYPE_OPN_MTX           xTYPE_MTX+0x80
#define xTYPE_OPN_SEM           xTYPE_SEM+0x80
#define xTYPE_OPN_PUB           xTYPE_PUB+0x80
#define xTYPE_OPN_TMR           xTYPE_TMR+0x80
#define xTYPE_OPN_QUE           xTYPE_QUE+0x80
#define xTYPE_OPN_PIP           xTYPE_PIP+0x80
#define xTYPE_OPN_TCB           xTYPE_TCB+0x80

#define xLIST_NONE              0
#define xLIST_READY             1
#define xLIST_TIMER             2
#define xLIST_RTCC              3


#define xRoundToInteger(v) (((v+sizeof(unsigned)-1)>>(sizeof(unsigned)/2))<<(sizeof(unsigned)/2))
#define xRoundTo4(v) (((v+3)>>2)<<2)
#define xRoundTo8(v) (((v+7)>>3)<<3)



/*****************************************************************************************************
*                                        GLOBAL VARIABLES
*****************************************************************************************************/
extern uint64_t qvLostCycles qNEAR;             // Cycles lost due to closed threads
extern uint32_t qvTotalTMR qNEAR;               // Running timer value used in yTimer
extern uint64_t qvCycles qNEAR;                 // Number of cycles used by system yTimer

extern pTCB qvCurTcb qNEAR;                     // TCB of current thread
extern pTCB qvRdyList qNEAR;                    // Ready list
extern pTCB qvTimerList qNEAR;                  // Thread/Timer Timer list
extern pTCB qvRtccList qNEAR;                   // Thread/Timer RTCC list
extern pOBJ qvObjectList qNEAR;                 // The list of all objects exclusive threads
extern pMSG qvMsgList qNEAR;                    // List of messages in memory (non-gixed memory)
extern pTCB qvIdleTcb qNEAR;                    // Pointer to the idle list

extern unsigned* qvIntStack qNEAR;              // Pointer to interrupt stack
extern unsigned* qvIntStackLimit qNEAR;         // Pointer to the limit of the interrupt stack

extern void (*qvSpawnFun[])(void) qNEAR;        // Spawn functions
extern unsigned qvSpawn qNEAR;                  // spawn call indicators
extern unsigned* qvFbrPut qNEAR;                // Write fibers here
extern unsigned* qvFbrGet qNEAR;                // Read fibers here
extern unsigned* qvFbrTrackAdr qNEAR;           // Address of bit to set/clr for fiber tracking
extern unsigned qvFbrTrackBit qNEAR;            // Bit to set/clr for fiber tracking

extern unsigned qvKernel qNEAR;                 // All bits to indicates state of the OS
extern int qvInCritical qNEAR;                  // -1=NoCritical 0=FirstCritical 1=Next etc

extern pMPL qvMemObjPool qNEAR;                 // Pool for all objects
extern pMPL qvMemTcbPool qNEAR;                 // Pool for all TCBs
extern pMPL qvMemTmrPool qNEAR;                 // Pool for all Timers
extern pMPL qvMemPoolList qNEAR;                // Mem pool list header
extern pFPL qvFixPoolList qNEAR;                // Fix pool list header
extern unsigned* qvMemHeap qNEAR;               // Heap list header
extern void* qvSpawnRtccExpire qNEAR;           // Contains function xFbrRTCCExpire
extern uint8_t* qvTrackingSleepAdr qNEAR;       // Address of the tracking bit for sleep
extern uint8_t* qvTrackingIdleAdr qNEAR;        // Address of the tracking bit for idle
extern uint8_t* qvTrackingRunAdr qNEAR;         // Address of the tracking bit for run

extern int volatile qvInIdlePreventMode qNEAR;  // 0=Idle allowed 1->n Idle prevented
extern int volatile qvInSleepPreventMode qNEAR; // 0=Sleep allowed 1->n Sleep prevented

extern uint32_t qvClockFrequency qNEAR;         // ClockFrequency
extern uint16_t qvClockFrequencyDiv15625 qNEAR;
extern uint16_t qvClockFrequencyDiv1000000 qNEAR;// ClockFrequency/1000000
extern uint16_t qvClockFrequencyMul256Div1000000 qNEAR; // (ClockFrequency*256)/1000000
extern uint16_t qvInterruptStackSize qNEAR;

extern uint8_t qvTrackingSleepBit qNEAR;        // Sleep Tracking bit to set or clear
extern uint8_t qvTrackingIdleBit qNEAR;         // Idle Tracking bit to set or clear
extern uint8_t qvTrackingRunBit qNEAR;          // Run Tracking bit to set or clear
extern uint8_t qvClockFrequencyMhzPower2 qNEAR;
extern uint8_t qvRtcc qNEAR;                    // Type of RTCC (0=None, 1=emulated, 2=Timer, 3=HwRTCC)
extern uint8_t qvTimer qNEAR;                   // Timer number (2=TMR2/TMR3 4=TMR4/TMR5, 6=TMR6/TMR7, 8=TMR8/TMR9)
extern uint8_t qvClockFrequencyMultiplyOf1000000 qNEAR; // 1 then multiply of MHz


extern struct sERR qvErr __attribute__ ((persistent)) qNEAR;

extern const struct sVER qcVer;

extern unsigned* qvSharedStack;         // Working stack for all EDS stacks
extern pTCB qvSharedStackOwner;         // The owner of the shared stack



// Kernel functions
unsigned  xErrNtf(unsigned Error);
pTCB      xGetTcb(char* pName);
unsigned  xInsTimerList(pTCB Tcb);
unsigned  xInsRtccList(pTCB Tcb);
void      xRemRdyInsWaitList(int32_t TimeOut);
void      xObjFree(pOBJ pObj);
void*     xObjAlloc(uint8_t Type, unsigned Size);
unsigned  xObjCheck(char* pName,  unsigned ErrId, uint8_t TypeId);
pOBJ      xObjSearch(char* pName, uint8_t Type);
bool      xObjInsert(pTCB Tcb);
void      xObjRemove(pTCB tcb);
unsigned  xInsRdyList(pTCB Tcb);
unsigned  xRemWaitInsRdyList(pTCB Tcb);
unsigned  xRemRdyInsEndRdyList(void);
void      xThrTimeoutOnObject(pOBJ pObj, int32_t TimeOut);
void      xThrSuspendOnObject(pOBJ pObj);
pMPL      xMemInsPool(unsigned BlockSize);
int32_t   xUSec2Cycles(uint32_t uSec);

//The open functions
pOBJ      xObjOpen(char* pName, unsigned ErrId, uint8_t Type);
pOBJ      xObjOpenNB(char* pName, unsigned ErrId, uint8_t Type);
pOBJ      xObjOpenTO(char* pName, unsigned ErrId, uint8_t Type, int32_t TimeOut);
void      xObjRemOpenList(char* pName, uint8_t Type);
// Other functions
void      xFbrRTCCExpire();
void      xFbrTimerExpire();
void      xIdleThr(void*);
unsigned  xTestWait(void*, unsigned TypeId, unsigned ErrId, int32_t TimeOut);
void      xQueReadyWaiting(pQUE p);
void      xSemRelease(pSEM p);
unsigned* xHeapAlloc(unsigned Size);
pMPL      xMemPool(unsigned BlockSize);






//****************************************************************************************************************
//***************************************** Inline Functions **********************************************
//****************************************************************************************************************

static inline __attribute__ ((always_inline)) void qBitSet(unsigned* Ptr, unsigned Bit) {
    *Ptr |= (1<<(Bit));
}
static inline __attribute__ ((always_inline)) void qBitClr(unsigned* Ptr, unsigned Bit) {
    *Ptr &= ~(1<<(Bit));
}
static inline __attribute__ ((always_inline)) bool qBitTst(unsigned* Ptr, unsigned Bit) {
    if (*Ptr & (1<<(Bit))) {
        return true;
    }
    return false;
}
static inline __attribute__ ((always_inline)) void qBitTgl(unsigned* Ptr, unsigned Bit) {
    *Ptr ^= (1<<(Bit));
}
static inline __attribute__ ((always_inline)) void xRemRdyList(pTCB Tcb) { // Always inline
    Tcb->pNext->pPrev = Tcb->pPrev;             // We don't have to test because of the idle thread
    Tcb->pPrev->pNext = Tcb->pNext;
    Tcb->ListType=xLIST_NONE;                   // We are not in any list
}
static inline __attribute__ ((always_inline)) void xRemWaitList(pTCB Tcb) { // Handles Timer and Rtcc list Always inline
    if (Tcb->pNext) {
        Tcb->pNext->pPrev = Tcb->pPrev;
    }
    Tcb->pPrev->pNext = Tcb->pNext;
    Tcb->ListType=xLIST_NONE;                   // We are not in any list
}


typedef enum {
    TIMER_TYPE_PERIODIC = 0,                    // Timer is periodic, its keep repeating
    TIMER_TYPE_ONE_SHOT = 1,                    // Timer is one-shot, expires only once
    TIMER_TYPE_PERIODIC_MANUAL_START = 2,       // Same as periodic but start is manual
    TIMER_TYPE_ONE_SHOT_MANUAL_START = 3        // Same as manual but start is manual
} TIMER_TYPE;

typedef enum {
    WAIT_TYPE_ALL = 0,                          // Wait for all events in the event set
    WAIT_TYPE_ANY = 1,                          // Wait for any event in the event set
    WAIT_TYPE_ALL_AND_CLEAR = 2,                // Same as wait_all but clears the events also
    WAIT_TYPE_ANY_AND_CLEAR = 3                 // Same as wait any but clears the events also
} WAIT_TYPE;

//******************************************** Bit functions *************************************************

void     qBitClr(unsigned* Ptr, unsigned Bit);
void     qBitMemClrAtomic(unsigned Bit);
void     qBitMemSetAtomic(unsigned Bit);
void     qBitMemTglAtomic(unsigned Bit);
bool     qBitMemTst(unsigned Bit);
void     qBitSet(unsigned* Ptr, unsigned Bit);
void     qBitTgl(unsigned* Ptr, unsigned Bit);
bool     qBitTst(unsigned* Ptr, unsigned Bit);

//******************************************** Byte functions ************************************************
void     qBytClr(void* p, unsigned Size);
void     qBytDecAtomic(volatile uint8_t* p);
void     qBytIncAtomic(volatile uint8_t* p);
void     qBytMov(void* From, void* To, unsigned Size);
void     qBytSet(void* p, unsigned Size);

//******************************************** CRC32C functions **********************************************
uint32_t qCrcGet(uint8_t* buf, unsigned size);           //  Polynomial = 0x1EDC6F41
uint32_t qCrcCalculate(uint32_t crc, uint8_t* buf, unsigned size);

//**************************************** Critical Section functions ****************************************
void     qCrtEnter();
void     qCrtExit();

//*********************************** Fast Divide Multiply functions *****************************************
uint32_t qDivU3216(uint32_t dividend, uint16_t divisor);
uint32_t qDivU3216R(uint32_t dividend, uint16_t divisor, uint16_t* Remainder);
uint64_t qDivU6416(uint64_t dividend, uint16_t divisor);
uint64_t qDivU6416R(uint64_t dividend, uint16_t divisor, uint16_t* Remainder);
uint16_t qDivScaling(uint16_t v1, uint16_t v2, uint16_t v3); // return (v1*v2)/v3
uint32_t qMulU3216(uint32_t v1, uint16_t v2);

//******************************************** DateTime functions ********************************************
uint32_t qDtmAddDays(uint32_t DatTim, int Day);
uint32_t qDtmAddHours(uint32_t DatTim, int Hour);
uint32_t qDtmAddMinutes(uint32_t DatTim, int Min);
uint32_t qDtmAddSeconds(uint32_t DatTim, int Sec);
uint32_t qDtmAddYears(uint32_t DatTim, int Year);
unsigned qDtmDayOfWeek(uint32_t DatTim);
uint32_t qDtmFromBcdDtm(pBCD_DTM DateTime);
uint32_t qDtmFromYMDHMS(uint8_t Year, uint8_t Month, uint8_t Day, uint8_t Hour, uint8_t Minute, uint8_t Second);
void     qDtmToBcdDtm(uint32_t DatTim, pBCD_DTM DateTime);


//******************************************* EventSet functions *********************************************
unsigned qEvtClear(pEVT pEvt, unsigned EventFlags);
void     qEvtClose(pEVT pEvt);
pEVT     qEvtCreate(char* pName);
pEVT     qEvtOpen(char* pName);
pEVT     qEvtOpenNB(char* pName);
pEVT     qEvtOpenTO(char* pName, int32_t Time);
void     qEvtSignal(pEVT pEvt, unsigned EventFlags);
unsigned qEvtWait(pEVT pEvt, unsigned EventFlags, WAIT_TYPE WaitType);
unsigned qEvtWaitNB(pEVT pEvt, unsigned EventFlags, WAIT_TYPE WaitType);
unsigned qEvtWaitTO(pEVT pEvt, unsigned EventFlags, WAIT_TYPE WaitType, int32_t TimeOut);


//************************************** Central Error handling functions ************************************
unsigned qErrNotify(unsigned Error) __attribute__((far));


//********************************************** Fiber functions *********************************************
void     qFbrCreate(uint8_t Priority, void (*pFbr)());
void     qFbrEnqueue0(void (*pFbr)());
void     qFbrEnqueue1(void (*pFbr)(void*), void* pPar1);
void     qFbrEnqueue2(void (*pFbr)(void*,void*), void* pPar1, void* pPar2);
void     qFbrSpawn1();
void     qFbrSpawn2();
void     qFbrSpawn3();
void     qFbrSpawn4();
void     qFbrSpawnRtcc();
uint64_t qFbrStatCycles();
void     qFbrTracking(unsigned* pAdr, unsigned BitNbr);


//***************************************** Fixed Memory functions *******************************************
void*    qFixAlloc(pFPL pFix);
void*    qFixAllocClr(pFPL pFix);
pFPL     qFixCreate(unsigned Size, unsigned NbrBlocks);
void     qFixFree(void* );


//********************************************** Heap functions **********************************************
void*    qHeaAlloc(unsigned size);
unsigned qHeaSize();


//********************************************* Kernel functions *********************************************
void     qKrnError(unsigned Error);
unsigned qKrnInCritical();
unsigned qKrnInFiber();
unsigned qKrnInIsr();
pERR     qKrnInit(uint32_t ClockFrequency, unsigned IdleThreadStackSize, unsigned InterruptStackSize, unsigned FiberQueueSize);
pERR     qKrnInitEds(uint32_t ClockFrequency, unsigned IdleThreadStackSize, unsigned InterruptStackSize, unsigned FiberQueueSize, unsigned SharedStackSize);
void     qKrnNotifyIdle() __attribute__((far));
void     qKrnNotifySwitch(pTCB pCurThr, pTCB pNextThr) __attribute__((far));
unsigned qKrnStack();
void     qKrnStart();
void     qKrnStatOff();
void     qKrnStatOn();
void     qKrnSwitchNotificationOff();
void     qKrnSwitchNotificationOn();
void     qKrnTrackingIdle(unsigned* pAdr, unsigned Bit);
void     qKrnTrackingRun(unsigned* pAdr, unsigned Bit);
void     qKrnTrackingSleep(unsigned* pAdr, unsigned Bit);
void     qKrnUSecOff();
void     qKrnUSecOn();
pVER     qKrnVersion();


//******************************************** Memory functions **********************************************
void*    qMemAlloc(unsigned Size);
void*    qMemAllocClr(unsigned Size);
void*    qMemAllocFromPool(pMPL pMpl);
void*    qMemAllocFromPoolClr(pMPL pMpl);
void*    qMemAllocFromPoolFast(pMPL pMpl);
void     qMemFree(void* p);
void     qMemFreeFast(void* p);
pMPL     qMemPool(unsigned Size);
unsigned qMemPoolAdd(pMPL pMpl, unsigned NbrBlocks);
pMPL     qMemPoolNext(pMPL pMpl);
unsigned qMemPoolSize(pMPL pMpl);
void*    qMemRealloc(void* p, unsigned size);


//******************************************** Message functions *********************************************
pMSG     qMsgAlloc(uint16_t Size, uint16_t MsgType);
pMSG     qMsgAllocFast(pMPL pMpl, uint16_t MaxSize, uint16_t MsgType);
pMSG     qMsgCopy(pMSG Msg);
uint16_t qMsgDataSize(pMSG Msg);
pMSG     qMsgFixAlloc(pFPL pFix, uint16_t MaxSize, uint16_t MsgType);
pFPL     qMsgFixCreate(unsigned MsgSize, unsigned NbrMessages);
void     qMsgFree(pMSG pMsg);
void     qMsgPublish(pPUB pPub, pMSG pMsg);
pMSG     qMsgRead(pPIP pPip);
pMSG     qMsgReceive(pQUE pQue);
pMSG     qMsgReceiveNB(pQUE pQue);
pMSG     qMsgReceiveTO(pQUE pQue, int32_t TimeOut);
void     qMsgSend(pQUE pQue, pMSG pMsg);
bool     qMsgSendNB(pQUE pQue, pMSG pMsg);
bool     qMsgSendTO(pQUE pQue, pMSG pMsg, int32_t TimeOut);
uint16_t qMsgSize(pMSG Msg);
unsigned qMsgWrite(pPIP pPip, pMSG pMsg);


//********************************************* Mutex functions **********************************************
void     qMtxClose(pMTX pMtx);
pMTX     qMtxCreate(char* pName, bool Lock);
void     qMtxLock(pMTX pMtx);
bool     qMtxLockNB(pMTX pMtx);
bool     qMtxLockTO(pMTX pMtx, int32_t TimeOut);
pMTX     qMtxOpen(char* pName);
pMTX     qMtxOpenNB(char* pName);
pMTX     qMtxOpenTO(char* pName, int32_t Time);
pTCB     qMtxOwner(pMTX pMtx);
void     qMtxUnlock(pMTX pMtx);


//******************************************** Pipe prototypes ***********************************************
unsigned qPipBlockSize(pPIP pPip);
void     qPipClose(pPIP pPip);
pPIP     qPipCreate(char* pName, uint16_t BlockSize, uint16_t MaxBlocks, void (*pNtfReader)(pPIP, unsigned), void (*pNtfWriter)(pPIP, unsigned));
unsigned qPipEntries(pPIP pPip);
unsigned qPipFreeBlocks(pPIP pPip);
unsigned qPipGet(pPIP pPip, void* Buffer, unsigned NbrBlocks);
unsigned qPipGetFast(pPIP pPip, void* Buffer, unsigned NbrBlocks);
bool     qPipGetBytFast(pPIP pPip, uint8_t* Buffer);
bool     qPipGetWrdFast(pPIP pPip, unsigned* Buffer);
unsigned qPipMaxBlocks(pPIP pPip);
pPIP     qPipOpen(char* pName);
pPIP     qPipOpenNB(char* pName);
pPIP     qPipOpenTO(char* pName, int32_t TimeOut);
unsigned qPipPut(pPIP pPip, void* Buffer, unsigned NbrBlocks);
unsigned qPipPutFast(pPIP pPip, void* Buffer, unsigned NbrBlocks);
bool     qPipPutBytFast(pPIP pPip, uint8_t* Buffer);
bool     qPipPutWrdFast(pPIP pPip, unsigned* Buffer);
unsigned qPipRead(pPIP pPip, void* Buffer, unsigned NbrBlocks);
bool     qPipReadFast(pPIP pPip, void* Buffer);
unsigned qPipWrite(pPIP pPip, void* Buffer, unsigned NbrBlocks);
bool     qPipWriteFast(pPIP pPip, void* Buffer);


//***************************************** Publish/subscribe functions **************************************
void     qPubClose(pPUB pPub);
pPUB     qPubCreate(char* pName);
pPUB     qPubOpen(char* pName);
pPUB     qPubOpenNB(char* pName);
pPUB     qPubOpenTO(char* pName, int32_t TimeOut);
void     qPubSubscribeFun(pPUB pPub, void (*pSub)(void*,pMSG));
void     qPubSubscribePip(pPUB pPub, pPIP pPip);
void     qPubSubscribeQue(pPUB pPub, pQUE pPip);


//************************************************ Power functions *******************************************
void     qPwrPermitIdle();
void     qPwrPermitSleep();
void     qPwrPreventIdle();
void     qPwrPreventSleep();


//********************************************* Queue  functions *********************************************
void     qQueClose(pQUE pQue);
pQUE     qQueCreate(char* pName, uint16_t Size);
pQUE     qQueOpen(char* pName);
pQUE     qQueOpenNB(char* pName);
pQUE     qQueOpenTO(char* pName, int32_t TimeOut);

//********************************************* Random functions *********************************************
uint32_t qRanGet();
void     qRanNotifySeed(uint32_t* x, uint32_t* y) __attribute__((far));

//********************************************* RTCC functions ***********************************************
void     qRtcAlarm(uint32_t DatTim, void(*pFbr)(void*), void* p);
uint32_t qRtcGetDatTim();
uint32_t qRtcGetUptime();
void     qRtcSetDatTim(uint32_t DatTim);


//******************************************* Semaphore functions ********************************************
void     qSemAcquire(pSEM pSem);
bool     qSemAcquireNB(pSEM pSem);
bool     qSemAcquireTO(pSEM pSem, int32_t TimeOut);
void     qSemClose(pSEM pSem);
pSEM     qSemCreate(char* pName, unsigned Permits);
pSEM     qSemOpen(char* pName);
pSEM     qSemOpenNB(char* pName);
pSEM     qSemOpenTO(char* pName, int32_t TimeOut);
unsigned qSemPermits(pSEM pSem);
void     qSemRelease(pSEM pSem);


//********************************************** Thread functions ********************************************
void     qThrClose(pTCB pTcb);
pTCB     qThrCreate(char* pName, void(*pFun)(void*), void* Par, unsigned StackSize, uint8_t Prio);
pTCB     qThrCreateSuspended(char* pName, void(*pFun)(void*), void* Par, unsigned StackSize, uint8_t Prio);
pTCB     qThrCurrent();
void     qThrDelay(unsigned USec);
unsigned qThrEvtClear(unsigned EventFlags);
void     qThrEvtSignal(pTCB pTcb, unsigned EventFlags);
unsigned qThrEvtWait(unsigned EventFlags, WAIT_TYPE WaitType);
unsigned qThrEvtWaitNB(unsigned EventFlags, WAIT_TYPE WaitType);
unsigned qThrEvtWaitTO(unsigned EventFlags, WAIT_TYPE WaitType, int32_t TimeOut);
pTCB     qThrOpen(char* pName);
pTCB     qThrOpenNB(char* pName);
pTCB     qThrOpenTO(char* pName, int32_t TimeOut);
void     qThrResume(pTCB pTcb, int Reason);
void     qThrResumeV4(pTCB pTcb);
void     qThrSetPriority(pTCB pTcb, uint8_t Priority);
int      qThrSleep(int32_t Time);
unsigned qThrStack(pTCB pTcb);
uint64_t qThrStatCycles(pTCB pTcb);
int      qThrSuspend();
unsigned qThrTagGet();
unsigned qThrTagSet(unsigned Value);
void     qThrTracking(pTCB pTcb, unsigned *pAdr, unsigned BitNbr);
void     qThrYield();


//********************************************* Time functions ***********************************************
uint64_t qTimCycles();
uint64_t qTimMSec();
uint64_t qTimUSec();

//********************************************* Timer functions **********************************************
void     qTmrClose(pTMR pTmr);
pTMR     qTmrCreate(char* pName, void (*pFunction)(void*), void* pParam, TIMER_TYPE TimerType, int32_t WaitTime);
pTMR     qTmrOpen(char* pName);
pTMR     qTmrOpenNB(char* pName);
pTMR     qTmrOpenTO(char* pName, int32_t Time);
void     qTmrStart(pTMR pTmr);
void     qTmrStop(pTMR pTmr);

//****************************************** Word (int) functions ********************************************
void     qWrdMov(unsigned* From, unsigned* To, unsigned Size);
void     qWrdClr(unsigned* p, unsigned Size);
void     qWrdDecAtomic(volatile int* p);
bool     qWrdDecIfNotZero(volatile int* p);
void     qWrdIncAtomic(volatile int* p);
void     qWrdSet(unsigned* p, unsigned Size);


//**************************************** Buildin replacement functions *************************************
unsigned qsprintf(char* buf, const char *format, ...); //TODO
void     qprintf(const char *format, ...); //TODO
void     qFmtPutChar(int c) __attribute__((far)); //TODO
void     qFmtWait() __attribute__((far)); //TODO

//**************************************** ERRORS ************************************************************
#define qERR_TRAP_OSCILLATOR_FAIL   0X1000
#define qERR_TRAP_ADDRESS_ERROR     0X1001
#define qERR_TRAP_HARDTRAP_ERROR    0X1002
#define qERR_TRAP_STACK_ERROR       0X1003
#define qERR_TRAP_MATH_ERROR        0X1004
#define qERR_TRAP_DMAC_ERROR        0X1005
#define qERR_TRAP_SOFTTRAP_ERROR    0X1006
#define qERR_TRAP_RESERVED_TRAP7    0X1007

#define qERR_EVT_ID                 0x1100
#define qERR_EVT_NO_START           0x1101
#define qERR_EVT_ISR                0x1102
#define qERR_EVT_FBR                0x1103
#define qERR_EVT_TIMEOUT            0x1105
#define qERR_EVT_MEMORY             0x1106
#define qERR_EVT_NAME_IN_USE        0x1107
#define qERR_EVT_NO_NAME            0x1108
#define qERR_EVT_CRITICAL           0x1109
#define qERR_EVT_NO_TIMER           0x110A
#define qERR_EVT_NO_RTCC            0x110B
#define qERR_EVT_IN_USE             0x1110
#define qERR_EVT_NO_FLAGS           0x1111
#define qERR_EVT_STACK_LIMIT_1      0x1112
#define qERR_EVT_STACK_LIMIT_2      0x1113
#define qERR_EVT_WAIT_TYPE          0x1114

#define qERR_FBR_ISR                0x1210
#define qERR_FBR_PRIO               0x1211
#define qERR_FBR_MEMORY             0x1212
#define qERR_FBR_BIT_NBR            0x1213
#define qERR_FBR_THREAD             0x1214
#define qERR_FBR_QUEUE_FULL         0x1215
#define qERR_FBR_STAT_OFF           0x1216

#define qERR_KRN_ISR                0x1310
#define qERR_KRN_MEM_SIZE           0x1311
#define qERR_KRN_STARTED            0x1312
#define qERR_KRN_NO_INIT            0x1313
#define qERR_KRN_TIMER              0x1314
#define qERR_KRN_STAT_TIMER         0x1315
#define qERR_KRN_SAME_TIMER         0x1316
#define qERR_KRN_INTERRUPT          0x1317
#define qERR_KRN_NO_STAT            0x1318
#define qERR_KRN_STAT               0x1319
#define qERR_KRN_LICENSE            0x131A
#define qERR_KRN_MEMORY             0x131B
#define qERR_KRN_MHZ                0x131C
#define qERR_KRN_RUN_BIT_NBR        0x131D
#define qERR_KRN_IDLE_BIT_NBR       0x131E
#define qERR_KRN_SLEEP_BIT_NBR      0x131F

#define qERR_NTF_SEED               0x1420

#define qERR_MSG_ID                 0x1510
#define qERR_MSG_ISR                0x1511
#define qERR_MSG_SIZE               0x1512
#define qERR_MSG_PIP_SIZE           0x1513

#define qERR_MTX_ID                 0x1600
#define qERR_MTX_NO_START           0x1601
#define qERR_MTX_ISR                0x1602
#define qERR_MTX_FBR                0x1603
#define qERR_MTX_TIMEOUT            0x1605
#define qERR_MTX_MEMORY             0x1606
#define qERR_MTX_NAME_IN_USE        0x1607
#define qERR_MTX_NO_NAME            0x1608
#define qERR_MTX_CRITICAL           0x1609
#define qERR_MTX_NO_TIMER           0x160A
#define qERR_MTX_NO_RTCC            0x160B
#define qERR_MTX_IN_USE             0x1610
#define qERR_MTX_LOCKED             0x1611
#define qERR_MTX_OWNER              0x1612

#define qERR_PIP_ID                 0x1700
#define qERR_PIP_NO_START           0x1701
#define qERR_PIP_ISR                0x1702
#define qERR_PIP_FBR                0x1703
#define qERR_PIP_TIMEOUT            0x1705
#define qERR_PIP_MEMORY             0x1706
#define qERR_PIP_NAME_IN_USE        0x1707
#define qERR_PIP_NO_NAME            0x1708
#define qERR_PIP_NO_TIMER           0x170A
#define qERR_PIP_NO_RTCC            0x170B
#define qERR_PIP_BLOCK_SIZE         0x1710
#define qERR_PIP_MAX_BLOCKS         0x1711
#define qERR_PIP_NBR_BLOCKS         0x1712

#define qERR_QUE_ID                 0x1800
#define qERR_QUE_NO_START           0x1801
#define qERR_QUE_ISR                0x1802
#define qERR_QUE_FBR                0x1803
#define qERR_QUE_TIMEOUT            0x1805
#define qERR_QUE_MEMORY             0x1806
#define qERR_QUE_NAME_IN_USE        0x1807
#define qERR_QUE_NO_NAME            0x1808
#define qERR_QUE_CRITICAL           0x1809
#define qERR_QUE_NO_TIMER           0x180A
#define qERR_QUE_NO_RTCC            0x180B
#define qERR_QUE_IN_USE             0x1810
#define qERR_QUE_NO_FLAGS           0x1811
#define qERR_QUE_SIZE               0x1812

#define qERR_SEM_ID                 0x1900
#define qERR_SEM_NO_START           0x1901
#define qERR_SEM_ISR                0x1902
#define qERR_SEM_FBR                0x1903
#define qERR_SEM_TIMEOUT            0x1905
#define qERR_SEM_MEMORY             0x1906
#define qERR_SEM_NAME_IN_USE        0x1907
#define qERR_SEM_NO_NAME            0x1908
#define qERR_SEM_CRITICAL           0x1909
#define qERR_SEM_NO_TIMER           0x190A
#define qERR_SEM_NO_RTCC            0x190B
#define qERR_SEM_IN_USE             0x1910
#define qERR_SEM_OVERFLOW           0x1911

#define qERR_THR_ID                 0x1A00
#define qERR_THR_NO_START           0x1A01
#define qERR_THR_ISR                0x1A02
#define qERR_THR_FBR                0x1A03
#define qERR_THR_TIMEOUT            0x1A05
#define qERR_THR_MEMORY             0x1A06
#define qERR_THR_NAME_IN_USE        0x1A07
#define qERR_THR_NO_NAME            0x1A08
#define qERR_THR_CRITICAL           0x1A09
#define qERR_THR_NO_TIMER           0x1A0A
#define qERR_THR_NO_RTCC            0x1A0B
#define qERR_THR_IN_USE             0x1A10
#define qERR_THR_PRIO               0x1A11
#define qERR_THR_STACK              0x1A12
#define qERR_THR_CURRENT            0x1A13
#define qERR_THR_NO_FLAGS           0x1A14
#define qERR_THR_WAIT_TYPE          0x1A15
#define qERR_THR_STAT_OFF           0x1A16
#define qERR_THR_READY_STATE        0x1A17
#define qERR_THR_WAIT_STATE         0x1A18
#define qERR_THR_WAIT_OBJECT        0x1A19
#define qERR_THR_BIT_NBR            0x1A1A
#define qERR_THR_EDS                0x1A1B
#define qERR_THR_SHARED_STACK_SIZE  0x1A1C
#define qERR_THR_NO_EDS             0x1A1D
#define qERR_THR_REASON             0x1A1E

#define qERR_TMR_ID                 0x1B00
#define qERR_TMR_NO_START           0x1B01
#define qERR_TMR_ISR                0x1B02
#define qERR_TMR_FBR                0x1B03
#define qERR_TMR_TIMEOUT            0x1B05
#define qERR_TMR_MEMORY             0x1B06
#define qERR_TMR_NAME_IN_USE        0x1B07
#define qERR_TMR_NO_NAME            0x1B08
#define qERR_TMR_CRITICAL           0x1B09
#define qERR_TMR_NO_TIMER           0x1B0A
#define qERR_TMR_NO_RTCC            0x1B0B
#define qERR_TMR_IN_USE             0x1B10
#define qERR_TMR_SIGNAL_OPT         0x1B11
#define qERR_TMR_TYPE               0x1B12
#define qERR_TMR_FUNCTION           0x1B13
#define qERR_TMR_TIME               0x1B14
#define qERR_TMR_NO_USEC            0x1B15

#define qERR_RTC_ID                 0x1C10
#define qERR_RTC_NO_START           0x1C11
#define qERR_RTC_ISR                0x1C12
#define qERR_RTC_FBR                0x1C13
#define qERR_RTC_1_1_2010           0x1C15
#define qERR_RTC_MEMORY             0x1C16
#define qERR_RTC_NAME_IN_USE        0x1C17
#define qERR_RTC_NO_NAME            0x1C18
#define qERR_RTC_CRITICAL           0x1C19
#define qERR_RTC_EMULATION          0x1C1A
#define qERR_RTC_NO_RTCC            0x1C1B
#define qERR_RTC_FUNCTION           0x1C1C


#define qERR_PUB_ID                 0x1D00
#define qERR_PUB_NO_START           0x1D01
#define qERR_PUB_ISR                0x1D02
#define qERR_PUB_FBR                0x1D03
#define qERR_PUB_TIMEOUT            0x1D05
#define qERR_PUB_MEMORY             0x1D06
#define qERR_PUB_NAME_IN_USE        0x1D07
#define qERR_PUB_NO_NAME            0x1D08
#define qERR_PUB_CRITICAL           0x1D09
#define qERR_PUB_NO_TIMER           0x1D0A
#define qERR_PUB_NO_RTCC            0x1D0B

#define qERR_MEM_ID                 0x1E00
#define qERR_MEM_ISR                0x1E10
#define qERR_MEM_DAMAGE             0x1E11
#define qERR_MEM_NO_MEM_LEFT        0x1E12

#define qERR_EDS_ID                 0x1F00
#define qERR_EDS_ISR                0x1F10
#define qERR_EDS_DAMAGE             0x1F11

#define qERR_PRINTF_PUTCHAR         0x2000
#define qERR_PRINTF_PUTWAIT         0x2001
#define qERR_PWR_IDLE_OUT_OF_SYNC   0x2002
#define qERR_PWR_SLEEP_OUT_OF_SYNC  0x2003

#endif //QKERNEL_COMMON_H


