/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


void* qMemAllocFromPool(pMPL pMpl) {
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_MEM_ISR);
        return NULL;
    }    
    if (pMpl->Type!=xTYPE_MPL) {
        xErrNtf(qERR_MEM_ID);
        return NULL;
    }
#endif
    qCrtEnter();
    pMBK mbk;
    if (pMpl->pFreeBlk==0) {
        mbk = (pMBK)xHeapAlloc(pMpl->BlockSize+sizeof(void*));
        if (mbk==NULL) {
            qCrtExit();
            return NULL;
        }
        pMpl->Total++;
    }
    else {
        mbk = pMpl->pFreeBlk;
        pMpl->pFreeBlk = (pMBK)mbk->pMpl;
    }
    qCrtExit();
    mbk->pMpl = pMpl;
    return &mbk->Memory;    
}

void qMemFree(void* pMem) {
    unsigned* pInt = (unsigned*)pMem;
    pInt--; 
    pMBK mbk = (pMBK)(pInt);
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_MEM_ISR);
        return;
    }    
    if (mbk->pMpl->Type!=xTYPE_MPL) {
        xErrNtf(qERR_MEM_ID);
        return;
    }
#endif
    pMPL mpl = mbk->pMpl;
    qCrtEnter();
    mbk->pMpl = (pMPL)mpl->pFreeBlk;
    mpl->pFreeBlk = mbk;    
    qCrtExit();
}

