/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"

uint16_t qMsgSize(pMSG pMsg) {
    uint8_t poolType = pMsg->MemPoolType & 0xf; // Keep only the least significant nibble
    unsigned size;                              // Used as the size of the real message
#ifndef __NO_CHECK__
    if (pMsg==NULL) {
        xErrNtf(qERR_MSG_ID);
        return NULL;
    }
#endif
#ifndef __NO_CHECK__
    if (poolType!=xTYPE_MPL_MSG && poolType!=xTYPE_FPL_MSG) {
        xErrNtf(qERR_MSG_ID);
        return 0;
    }
    if (qKrnInIsr() && poolType != xTYPE_FPL_MSG) {
        xErrNtf(qERR_MSG_ISR);
        return 0;
    }
#endif
    uint8_t sizeDiff = pMsg->MemPoolType>>4;
    if (poolType==xTYPE_MPL_MSG) {
        pMPL pool = (pMPL)*((int*)pMsg-1);      // Location before message is..
                                                // ..pointer to the memory pool
        size = pool->BlockSize;
    }
    else {
        size = pMsg->pFixedPool->BlockSize;
    }
    return size - sizeDiff;                     // Pool size minus the size difference
}

uint16_t qMsgDataSize(pMSG pMsg) {
    return qMsgSize(pMsg) - sizeof(void*) - 4;  // One pointer, 2 bytes 1 16bit
}

