/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


void qMsgFree(pMSG pMsg) {
#ifndef __NO_CHECK__
    if (pMsg==NULL) {
        xErrNtf(qERR_MSG_ID);
        return;
    }
#endif
    uint8_t poolType = pMsg->MemPoolType & 0xf;// Keep only the least significant nibble

#ifndef __NO_CHECK__
    if (poolType!=xTYPE_MPL_MSG && poolType!=xTYPE_FPL_MSG) {
        xErrNtf(qERR_MSG_ID);
        return;
    }
#endif
    if (poolType==xTYPE_MPL_MSG) {
        if (!qCrtEnterFirst()) {
            if (qKrnInIsr()) {
                qCrtExit();
                qFbrEnqueue1((void(*)(void*))qMsgFree, pMsg);
                return;
            }
        }
        qBytDecAtomic(&pMsg->UseCnt);
        if (pMsg->UseCnt == 0) {
            pMsg->MemPoolType = 0;
            pMSG p = qvMsgList;         // Start of all messages
            pMSG t = 0;                 // Following pointer
            while (p!=pMsg) {           // If not found
                t = p;                  // save the previous pointer
                p = p->pNext;           // Get the next one
            }
            if (t) {                    // if not in the beginning
                t->pNext = p->pNext;    // just copy the pointer back
            }
            else {
                qvMsgList =  p->pNext;  // else copy into the Message list
            }
            qMemFree(pMsg);
        }
        qCrtExit();
    }
    else {
        qBytDecAtomic(&pMsg->UseCnt);
        if (pMsg->UseCnt == 0) {
            pMsg->MemPoolType = 0;
            qFixFree(pMsg);
        }
    }
}

