/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


bool qMsgSendTO(pQUE pQue, pMSG pMsg, int32_t TimeOut) {

#ifndef __NO_CHECK__
    if (xTestWait(pQue, xTYPE_QUE, qERR_QUE_ID, TimeOut))
        return false;
    if (pMsg==NULL) {
        xErrNtf(qERR_MSG_ID);
        return false;
    }
    uint8_t poolType = pMsg->MemPoolType & 0xf; // Keep only the least significant nibble
    if (poolType!=xTYPE_MPL_MSG && poolType!=xTYPE_FPL_MSG) {
        xErrNtf(qERR_MSG_ID);
        return false;
    }    
#endif
    qCrtEnter();
    if (pQue->Entries < pQue->Size) {           // Is there space in the queue
        pQue->Entries++;                        //
        *pQue->pIn++ = pMsg;                    // Save the message
        qBytIncAtomic(&pMsg->UseCnt);           // Send so increment
        if (pQue->pIn==pQue->pMax)
            pQue->pIn = &pQue->pMessages;
    }
    else {
        // Here we wait for space
        xThrTimeoutOnObject((pOBJ)pQue,TimeOut);// Suspend task until event or timeout occurs
                                                // EXIT_CRITICAL will switch the thread because 
                                                // the wait-time has set the kernel interrupt
        if (qvCurTcb->TimedOut) {               // Did we timeout on a wait?
            return false;                       // Timeout error
        }
        qCrtEnter();
        if (pQue->Entries < pQue->Size) {       // Is there space in the queue
            pQue->Entries++;                    //
            *pQue->pIn++ = pMsg;                // Save the message
            qBytIncAtomic(&pMsg->UseCnt);       // Send so increment
            if (pQue->pIn==pQue->pMax)
                pQue->pIn = &pQue->pMessages;
        }
        else {
            qCrtExit();
            return false;
        }
    }
    
    // ready waiting senders
    xQueReadyWaiting(pQue);
    return true;
}



