/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


pPIP qPipCreate(char* pName, uint16_t BlockSize, uint16_t MaxBlocks,
                            void (*pNtfReader)(pPIP, unsigned), 
                            void (*pNtfWriter)(pPIP, unsigned)) {
    pPIP p;
    unsigned size;

#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_PIP_ISR);
        return NULL;
    }
    if (qKrnInFiber()) {
        xErrNtf(qERR_PIP_FBR);
        return NULL;
    }
    if (BlockSize==0) {
        xErrNtf(qERR_PIP_BLOCK_SIZE);
        return NULL;
    }
    if (MaxBlocks<2) {
        xErrNtf(qERR_PIP_MAX_BLOCKS);
        return NULL;
    }
    if (xErrNtf(xObjCheck(pName, qERR_PIP_ID, xTYPE_PIP))) {
        return NULL;
    }
#endif

    qCrtEnter();
    /*
     * The MaxBlock size used is one larger than used to determine the size.
     * This is done so that an interrupt does not overwrite before the
     * last value is read.
     * The maximum size for the buffer is also compensated for one block
     */
    size = sizeof(struct sPIP) - sizeof(unsigned);
    size = size + (unsigned)BlockSize*((unsigned)MaxBlocks+1);
    p = xObjAlloc(xTYPE_PIP,size);
    if (!p) {
        xErrNtf(qERR_PIP_MEMORY);
        return NULL;
    }
    p->BlockSize = BlockSize;
    if ((BlockSize & (sizeof(int)-1))) {    // Is it an integer size?
        p->BlockSizeINT = 0;
    }    
    else {
        p->BlockSizeINT = (BlockSize>>((sizeof(int)>>1)));
    }    
    p->pName = pName;
    p->MaxBlocks = MaxBlocks;    
    p->pNtfWriter = pNtfWriter;
    p->pNtfReader = pNtfReader;
    p->Type = xTYPE_PIP;
    p->pMax = &p->Data + (BlockSize * (MaxBlocks+1));
    p->pIn = &p->Data;
    p->pOut = &p->Data;
    qCrtExit();
    xObjRemOpenList(pName, xTYPE_OPN_PIP);   // Make waiting Open threads runnable
    return p;
}

