/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


bool qSemAcquireTO(pSEM pSem, int32_t TimeOut) {
    pSEM p = (pSEM)(((unsigned)pSem)-(4*sizeof(int)));    
#ifndef __NO_CHECK__
    if (xTestWait(p, xTYPE_SEM, qERR_SEM_ID, TimeOut))
        return false;
#endif
    if (qWrdDecIfNotZero(&p->Permits)) return true;

    qCrtEnter();
    if (p->Permits > 0) {     // We do it again because it can be changed outside...
        qWrdDecAtomic(&p->Permits);             // ...the critical section
        qCrtExit();
        return true;                            // return success
    }
    xThrTimeoutOnObject((pOBJ)p,TimeOut);       // Suspend task until event or timeout occurs
                                                // EXIT_CRITICAL will switch the thread because 
                                                // the wait-time has set the kernel interrupt
    if (qvCurTcb->TimedOut)                     // Did we time out?
        return false;                           // Yes return failure
    qWrdDecAtomic(&p->Permits);                 // Now decrement
    return true;                                // return success
}

