/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"

void qSemRelease(pSEM pSem) {
    pSEM p = (pSEM)(((unsigned)pSem)-(4*sizeof(int)));    
    
#ifndef __NO_CHECK__
    if (p==NULL || p->Type!= xTYPE_SEM) {
        xErrNtf(qERR_SEM_ID);
        return;
    }
#endif
    if (p->Permits > 0) {       // If there are permits we don't have to switch so just increment
        qWrdIncAtomic(&p->Permits);
        return;
    }
    xSemRelease(p);
}

// this function is called from the function above or the .s version if it is in an
// interrupt or there is a waitlist.
//
void xSemRelease(pSEM p) {
    pTCB tcb;

    if (!qCrtEnterFirst()) {
        if (qKrnInIsr()) {
            qCrtExit();
            qFbrEnqueue1((void (*)(void*))xSemRelease, p);
            return;
        }
    }
    qWrdIncAtomic(&p->Permits);
    tcb = p->pWaitList;                         // List of waiting threads (prio order)
    if (tcb) {
        xObjRemove(tcb);
        if (xRemWaitInsRdyList(tcb)) {              // Remove WaitList Insert ReadyList
            if (qKrnInIsr()) {
                yKrnSetInterrupt();                 // Use the kernel interrupt to do that because
            }                                       // we cannot use switchThread in an interrupt.
            else {
                ySwitchThread();                    // Switch thread and EXIT_CRITICAL
                return;                             // only if readied thread > prio
            }
        }
    }
    qCrtExit();
}

