/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"



void qThrClose(pTCB pTcb) {
    pTCB tmp;
    pOBJ p;
    pTCB pp;
    pTCB tt;

#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_THR_ISR);
        return;
    }
    if (pTcb==0 || pTcb->Type!=xTYPE_TCB) {
        xErrNtf(qERR_THR_ID);
        return;
    }
#endif
    qCrtEnter();
    p = qvObjectList;
    // For each object find the mutexes and if we own free the mutex
    // this process is the same as unlock but we don't handle priority enhiritance
    // and we don't reschedule because we do that at the end.
    while (p) {                                 // If we own the mutex
        switch (p->Type) {
            case xTYPE_MTX:
                if (((pMTX)p)->Owner==pTcb) {
                    ((pMTX)p)->Owner = 0;       // we don't own the mutex anymore
                    tmp = ((pMTX)p)->pWaitList; // List of thread that we wait for in prio order
                    if (tmp) {                  // Are there threads waiting? (thread is readied)
                        xObjRemove(tmp);        // remove from the object list
                        xRemWaitInsRdyList(tmp);// Remove WaitList Insert ReadyList
                        }
                    }
                break;
        }
        p=p->pObjList;
    }

    pTcb->Type=0;                               // Indicate the TCB is not valid anymore
    if (pTcb->ListType==xLIST_READY) {
        xRemRdyList(pTcb);                      // Remove from ready list
    }
    else {
        if (pTcb->pObject) {                    // Is the thread wating on a object?
            xObjRemove(pTcb);                   // remove from the object list
        }
        if (pTcb->ListType!=xLIST_NONE) {
            xRemWaitList(pTcb);                 // Now remove the thread from the list
        }    
    }        
    // Remove us from the AllTcb list. We know that the first is the Idle thread so we never have
    // to change the root pointer
    pp = qvIdleTcb;                             // Start of all TCB's
    tt = 0;
    while (pp!=pTcb) {                          // Search the Tcb from the beginning
        tt = pp;
        pp = pp->pTcbList;
    }
    tt->pTcbList = pp->pTcbList;                // And give the previous one our next pointer
    
    qMemFree(pTcb->pStack);                     // Free the stack memory
    qvLostCycles += pTcb->Cycles + yGetCycles() - pTcb->LastCycles;
    qMemFree(pTcb);                             // Now we can free the TCB itself
    ySwitchThread();                            // Switch Thread and EXIT_CRITICAL
 }    

