/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"



static inline void* xAllocStack(unsigned Size) {
    unsigned* p = qMemAlloc(Size);
    if (p) {
        qWrdSet(p,Size);
    }    
    return p;
}



static pTCB xThrCreate(char* pName, void (*Thread)(void*), void* pParam, int StackSize, uint8_t Prio, uint8_t Suspended) {
    unsigned* stack;
    uint8_t nbrThread;
    pTCB tcb;
    
#ifndef __NO_CHECK__
    unsigned prio = Prio;
    if (xKrnIsPriorityCheckOn() && (prio==0 || prio>250)) {
        xErrNtf(qERR_THR_PRIO);
        return NULL;
    }
    if (qKrnInIsr()) {
        xErrNtf(qERR_THR_ISR);
        return NULL;
    }
    if (StackSize <= 16) {
        xErrNtf(qERR_THR_STACK);
        return NULL;
    }
    if (pName!=0) {
        if (xGetTcb(pName)) {                   // Search if there is something with the name
            xErrNtf(qERR_THR_NAME_IN_USE);
            return NULL;
        }
    }
#endif
    xKrnClrPriorityCheckOn();                   // Checking on so next time the switch has to be set
                                                // Off again if we don't want checking
    qCrtEnter();
    tcb = (pTCB)qMemAllocFromPoolClr((pMPL)qvMemTcbPool);
#ifndef __NO_CHECK__
    if (tcb==0) {
        qCrtExit();
        xErrNtf(qERR_THR_MEMORY);
        return NULL;
    }
#endif

    nbrThread=1;
    if (xKrnIsInitDone()) {                     // Do we have an idle thread?
        pTCB p = qvIdleTcb;                          // Start with the idle thread
        while (p->pTcbList) {                   // All Tcb List is done
            nbrThread++;                        // add to the number of threads
            p = p->pTcbList;                    // next one
        }
        p->pTcbList = tcb;                      // add us to the end of the list
    }
    stack =(unsigned*)xAllocStack(xRoundToInteger(StackSize));
#ifndef __NO_CHECK__
    if (stack == 0) {
        qCrtExit();
        xErrNtf(qERR_THR_MEMORY);
        return NULL;
    }
#endif
    yThrStkInit(Thread, pParam, tcb, stack, StackSize); // Initialize the thread's stack
    tcb->Prio = Prio;                           // Load thread priority into TCB
    tcb->StackSize = StackSize;                 // Store stack size
    tcb->pStack = stack;                        // Store pointer to bottom of stack
    tcb->pName = pName;
    tcb->Type = xTYPE_TCB;
    
    // Set tracking to its own bit. IOR will do nothing
    tcb->TrackAdr = &tcb->TrackBit;
    tcb->TrackBit = 0;
    
    if (!xKrnIsInitDone()) {                    // Is this the Idle Thread?
        tcb->ListType = xLIST_READY;
        qvRdyList=tcb;
        qCrtExit();
        return tcb;
    }
    if (!Suspended) {
        if (xInsRdyList(tcb)) {                 // Insert the new TCB. Is it higher?
            ySwitchThread();                    // Switch Thread and EXIT_CRITICAL
        }                                       // only if readied thread > prio
        else {
            qCrtExit();
        }
    }
    else {                                      // Tcb->ListTYpe = qLIST_NONE because TCB is zeroed
        qCrtExit();
    }
    xObjRemOpenList(pName, xTYPE_OPN_TCB);   // Make waiting Open threads runnable
    return tcb;
}

pTCB qThrCreate(char* pName, void (*Thread)(void*), void* pParam, unsigned StackSize, uint8_t Prio) {
    return xThrCreate(pName, Thread , pParam, StackSize, Prio, 0);
}

pTCB qThrCreateSuspended(char* pName, void (*Thread)(void*), void* pParam, unsigned StackSize, uint8_t Prio) {
    return xThrCreate(pName, Thread ,pParam, StackSize, Prio, 1);
}
