/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"

__attribute__((weak, far))
void qKrnNtfSwitch(pTCB pCurThr, pTCB pNextThr) {
}

unsigned xErrNtf(unsigned Error) {
    return qErrNotify(Error);
}


void qKrnError(unsigned Error) {
    yErrorReset(Error);
}    


void qKrnSwitchNotificationOn() {
    qvKernel |= (1<<_xSwitchNotificationOn);
}

void qKrnSwitchNotificationOff() {
    qvKernel &= ~(1<<_xSwitchNotificationOn);
}


static bool SameName(char* p, char* q) {
    if (p==0 || q==0) return false;             // compare with Null always NOT the same
    char c1 = *p++;
    char c2 = *q++;
    while (c1==c2 && c1!=0 && c2!=0) {          // while true get next character
        c1 = *p++;
        c2 = *q++;
    }
    if (c1==0 && c2==0) return true;            // Return true when they are the same
    return false;                               // Return false when they are not the same
}



pTCB xGetTcb(char* pName) {
    pTCB p;
        
    qCrtEnter();
    p = qvIdleTcb;
    while (p && (SameName(pName,p->pName)==false)) { // The Tcb with the correct name
        p = p->pTcbList;
    }
    qCrtExit();
    return p;
}



unsigned xInsTimerList(pTCB Tcb) {
    pTCB p;
    pTCB t;
    unsigned r = 1;
        
    p = qvTimerList;
    t = (pTCB)&qvTimerList;
    while (p && (Tcb->WakeUpTime > p->WakeUpTime)) { // Search for the minimum wait time
        r = 0;
        t = p;
        p = p->pNext;
    }
    Tcb->pNext = p;
    Tcb->pPrev = t;
    t->pNext = Tcb;
    if (p) {
        p->pPrev = Tcb;
    }
    Tcb->ListType = xLIST_TIMER;                // We are in the Timer list
    Tcb->TimedOut = 0;                          // Not yet timed out
    return r;
}


unsigned xInsRtccList(pTCB Tcb) {
    pTCB p;
    pTCB t;
    unsigned r = 1;

    p = qvRtccList;
    t = (pTCB)&qvRtccList;
    while (p && (Tcb->WakeUpTime > p->WakeUpTime)) { // Search for the minimum wait time
        r = 0;
        t = p;
        p = p->pNext;
    }
    Tcb->pNext = p;
    Tcb->pPrev = t;
    t->pNext = Tcb;
    if (p) {
        p->pPrev = Tcb;
    }
    Tcb->ListType = xLIST_RTCC;                 // We are in the Rtcc list
    Tcb->TimedOut = 0;                          // Not yet timed out
    return r;
}

    
// Function is based on the CurTCB
// TimeOut defines the rtcc or timer list
//
void xRemRdyInsWaitList(int32_t TimeOut) { 
    //RemRdyList    
    qvCurTcb->pNext->pPrev = qvCurTcb->pPrev;
    qvCurTcb->pPrev->pNext = qvCurTcb->pNext;

    if (TimeOut<0) {
        qvCurTcb->WakeUpTime = yRtccGetDatTimRounded() + (-TimeOut);
        if (xInsRtccList(qvCurTcb))
            yRtccSetAlarm(qvCurTcb->WakeUpTime);
    }
    else {    
        qvCurTcb->WakeUpTime = yTimerGetTime() + xUSec2Cycles(TimeOut);
        if (xInsTimerList(qvCurTcb))
            yTimerSetAlarm(qvCurTcb->WakeUpTime);
    }    
}

/* The main object has been created in qXxxCreate() and here we make the
 * waiting threads qXxxOpen() and qXxxOpenTO() runable and remove the
 * temporary object that was used to indicate open.
 */
void xObjRemOpenList(char* pName, uint8_t Type) {
    pOBJ p;                                     // Working pointer
    pTCB tcb;

    p = xObjSearch(pName, Type);
    if (p==0) return;

    qCrtEnter();
    tcb = p->pWaitList;                         // Get the start TCB
    while (tcb) {                               // Every thread waiting on this event
        tcb->pObject = 0;                       // remove the object from the TCB
        xRemWaitInsRdyList(tcb);                // Remove from wait Insert in ready
        tcb = tcb->pNextWaitObj;                // next thread that waits on the object
    }
    xObjFree(p);                                // Free the original open object

    if (qvCurTcb!=qvRdyList) {                  // if this thread is not the highest prio ready.
        ySwitchThread();                        // Switch Thread and EXIT_CRITICAL
        return;
    }
    qCrtExit();
}


// Object is removed from the Object List. MUST BE IN CRITICAL SECTION
void xObjFree(pOBJ pObj) {
    pOBJ p;
    pOBJ t;
    
    p = qvObjectList;                           // The first object
    t = 0;                                      // The trailing pointer
    while (p!=pObj) {                           // Search the objects from the beginning
        t = p;                                  // t will trail behind
        p = p->pObjList;
    }
    if (t) {                                    // Is t something
        t->pObjList = p->pObjList;              // Yes then we point to the Next
    }
    else {                                      // If not
        qvObjectList = p->pObjList;             // we have to change the header
    }

    pObj->Type = 0;                             // In case it is used it will fail
    qMemFree(pObj);
}


// Object is allocated. MUST BE IN CRITICAL SECTION
void* xObjAlloc(uint8_t Type, unsigned Size) {
    pOBJ p;

    if (Size==0)                                // Fixed objects
        p = qMemAllocFromPoolClr((pMPL)qvMemObjPool);
    else
        p = qMemAllocClr(Size);                 // Variable objects
    if (p) {
        p->pObjList = qvObjectList;             // Put the new object at the beginning at ... 
        qvObjectList = p;                       // ...Object list. 
    }
    return p;
}



unsigned xObjCheck(char* pName,  unsigned ErrId, uint8_t Type) {
    pOBJ p;
    p = xObjSearch(pName, Type);
    if (p) return ErrId+7;
    return 0;
}



pOBJ xObjSearch(char* pName, uint8_t Type) {
    pOBJ p;

    if (pName==0) return NULL;                  // return 0 if no names specified
    qCrtEnter();
    p = qvObjectList;                           // The list
    while (p) {                                 // Search the object from the beginning
        if (SameName(pName,p->pName)==true && p->Type==Type) {
            goto exit;
        }
        p = p->pObjList;
    }
exit:
    qCrtExit();
    return p;
}

// Return true if the inserted TCB it is the highest in the object
// Tcb->pObject must be set already

bool xObjInsert(pTCB Tcb) {
    pTCB p;
    pTCB t;
    bool result;

    p = Tcb->pObject->pWaitList;                // pointer in timer/Rtcc list
    t = 0;

    if (xKrnIsInsertEndReadyList()) {
        while (p && (Tcb->Prio <= p->Prio)) {   // Insert at the end of threads with same priority
            t = p;
            p = p->pNextWaitObj;
        }
    }
    else {
       while (p && (Tcb->Prio < p->Prio)) {     // Insert at the end of threads with same priority
            t = p;
            p = p->pNextWaitObj;
        }
    }
    Tcb->pNextWaitObj = p;
    if (t) {
        t->pNextWaitObj = Tcb;
        result = false;
    }
    else {
        Tcb->pObject->pWaitList = Tcb;
        result = true;
    }
    return result;
}



void xObjRemove(pTCB tcb) {
    pTCB pp;
    pTCB tt;
    pp = tcb->pObject->pWaitList;               // The first TCB waiting on this object
    tt = 0;                                     // The trailing TCB
    while (pp!=tcb) {                           // Search the Tcb from the beginning
        tt = pp;                                // t will trail behind
        pp = pp->pNextWaitObj;
    }
    if (tt) {                                   // Is t something
        tt->pNextWaitObj = pp->pNextWaitObj;    // Yes then we point to the Next
    }
    else {                                      // If not
        tcb->pObject->pWaitList = pp->pNextWaitObj; // we have to change the object itself
    }
    tcb->pObject = 0;                           // Unlink object from this thread
}

    
// return 1 if the inserted TCB is the first in the ready list
// return 0 if the tcb is not the first of the ready list 

unsigned xInsRdyList(pTCB Tcb) {                // Returns if duplicate or not
    pTCB p;
    pTCB t;
    unsigned r = 1;
                
    Tcb->ListType = xLIST_READY;
    p = qvRdyList;
    t = (pTCB)&qvRdyList;
    
    if (xKrnIsInsertEndReadyList()) {
        while (Tcb->Prio <= p->Prio) {          // Insert object at the end of it's priority list
            t = p;
            p = p->pNext;
            r = 0;
        }
    }
    else {
        while (Tcb->Prio < p->Prio) {           // Insert object at the end of it's priority list
            t = p;
            p = p->pNext;
            r = 0;
        }
    }
    Tcb->pNext = p;
    Tcb->pPrev = t;
    t->pNext = Tcb;
    if (p) p->pPrev = Tcb;
    return r;
}


// return 1 if the inserted TCB is the first of the ready list
// return 0 if the tcb is not the first of the ready list 

unsigned xRemWaitInsRdyList(pTCB Tcb) {
    pTCB p;
    pTCB t;
    unsigned r = 1;
        
// RemWaitList part (Timer or Rtcc. If in no list this will be skipped)
    if (Tcb->ListType>=xLIST_TIMER) {           // xLIST_RTCC > xLIST_TIMER
        if (Tcb->pNext) 
            Tcb->pNext->pPrev = Tcb->pPrev;
        Tcb->pPrev->pNext = Tcb->pNext;
    }

// InsRdyList part
    Tcb->ListType = xLIST_READY;
    p = qvRdyList;
    t = (pTCB)&qvRdyList;
    
    if (xKrnIsInsertEndReadyList()) {
        while (Tcb->Prio <= p->Prio) {          // Insert object at the end of the list with same priority
            t = p;
            p = p->pNext;
            r = 0;
        }
    }
    else {
        while (Tcb->Prio < p->Prio) {           // Insert object at the begining of the list with same priority
            t = p;
            p = p->pNext;
            r = 0;
        }
    }
    Tcb->pNext = p;
    Tcb->pPrev = t;
    t->pNext = Tcb;
    if (p) p->pPrev = Tcb;
    return r;
}

// return 1 if the inserted TCB is the first of the ready list
// return 0 if the tcb is not the first of the ready list
unsigned xRemRdyInsEndRdyList(void) {
    pTCB p;
    pTCB t;
    pTCB Tcb;
    unsigned r = 1;

    Tcb = qvCurTcb;
    qvCurTcb->pNext->pPrev = qvCurTcb->pPrev;   // Remove current ready task from the ready list
    qvCurTcb->pPrev->pNext = qvCurTcb->pNext;   // ...

// InsRdyList part
    Tcb->ListType = xLIST_READY;
    p = qvRdyList;
    t = (pTCB)&qvRdyList;

    while (Tcb->Prio <= p->Prio) {               // Insert object at the end of it's priority list
        t = p;
        p = p->pNext;
        r = 0;
    }
    Tcb->pNext = p;
    Tcb->pPrev = t;
    t->pNext = Tcb;
    if (p) p->pPrev = Tcb;
    return r;
}


// Must be in a critical section and we don't check if it is a multiply of 8.
// Size must be a multiply of sizeof(int)
//
pMPL xMemInsPool(unsigned BlockSize) {
    pMPL p;
    pMPL t;
    pMPL n;
    
    n = (pMPL)xHeapAlloc(sizeof(struct sMPL));
    if (!n) return NULL;
    n->BlockSize = BlockSize;
    n->pFreeBlk = 0;
    n->Type = xTYPE_MPL;
    p = qvMemPoolList;
    t = 0;
    while (p && (BlockSize > p->BlockSize)) {
        t = p;
        p = p->pNextPool;
    }
    n->pNextPool = p;
    if (t)
        t->pNextPool = n;
    else
        qvMemPoolList = n;
    return n;
}

// Must be in a critical section and we don't check if it is a multiply of sizeof(int).
// Size must be a multiply of sizeof(int)
//
unsigned* xHeapAlloc(unsigned Size) {
    unsigned free;
    unsigned* p = qvMemHeap;                    // Pointer to the heap for return
#ifndef __NO_CHECK__
    if (*(p+1)) {                               // If the next word is not 0 Error 
        xErrNtf(qERR_MEM_DAMAGE);               // Indicate the error
        return NULL;
    }
#endif
        
    free = *qvMemHeap;
    if (Size>free) {
        xErrNtf(qERR_MEM_NO_MEM_LEFT);          // Indicate the error
        return NULL;
    }
    free = free-Size;
    qvMemHeap = qvMemHeap + (Size>>(sizeof(int)>>1));
    *qvMemHeap = free;
    *p = 0;                                     // clear the old size out
    return p;
}

// Must be in a critical section
// Size must be a multiply of sizeof(int)
// We compare with the exact size and the rounded to 8 size and see whta fits best
//
pMPL xMemPool(unsigned BlockSize) {
    pMPL p;
    p = qvMemPoolList;
    while (p && p->BlockSize<xRoundTo8(BlockSize)) {
        p=p->pNextPool;
    }
    if (p && p->BlockSize>=BlockSize && p->BlockSize<=xRoundTo8(BlockSize)) {
        return p;
    }
    return NULL;
}

// Making a thread waiting on a object with timeout (Called from wait) 
// The function removes the TCB from the ready list and add it to the Timer/Rtcc List for timing
// It also add's it to the List for that object.
// Last but not least switch the context
//
void xThrTimeoutOnObject(pOBJ pObj, int32_t TimeOut) {
    qvCurTcb->pObject = pObj;                   // We wait on a object
    xRemRdyInsWaitList(TimeOut);                // Remove us from the ReadyList in the waitlist
                                                // This will also set the Timeout value to 0 ..
                                                // .. so we can use it for qThrResume()
    xObjInsert(qvCurTcb);
    ySwitchThread();                            // Switch Thread and EXIT_CRITICAL
}

// Making a thread waiting on a object (Called from wait) 
// The function removes the TCB from the ready list 
// It also add's it to the List for that object.
// Last but not least switch the context
//
void xThrSuspendOnObject(pOBJ pObj) {
    qvCurTcb->pObject = pObj;                   // We wait on a object
    qvCurTcb->TimedOut = 0;                     // This is for the case that we resume. In that case ...
                                                // ... TimedOut will be set and the waiting function ...
                                                // ... knows this and can return the appropriate values
    xRemRdyList(qvCurTcb);                      // Remove us from the ReadyList
    xObjInsert(qvCurTcb);
    ySwitchThread();                            // Switch Thread and EXIT_CRITICAL
}

int32_t xUSec2Cycles(uint32_t uSec) {
    int32_t ret;
    if (qvClockFrequencyMultiplyOf1000000) {
        ret = qMulU3216(uSec, qvClockFrequencyDiv1000000);
    }
    else if (uSec>0xffff) {
        uSec = uSec>>6;         // devide by 64
        ret = qMulU3216(uSec, qvClockFrequencyDiv15625);
    }
    else {
        uSec = qMulU3216(uSec, qvClockFrequencyDiv15625);
        ret = (int32_t)uSec>>6;
    }
    return ret;
}

