/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"

//
// The scheduler for RTCC is only called if the time has arrived. 
//

void xFbrRTCCExpire() {
    pTCB p;
    uint32_t datTime = yRtccGetDatTim();
/*
    Walk true the rtcc list and see if a thread or timer is timed-out. We do this until we reach wakeup times
    that are bigger then the date time. If that's the case we are done.
*/
    while (((p=qvRtccList)!=0) && (p->WakeUpTime<=datTime)) {
        if (p->Type==xTYPE_TCB) {
            p->TimedOut = 1;                    // Indicate a time out
            if (p->pObject) {                   // Do we time out from a object (not a Sleep)
                xObjRemove(p);                  // Yes then remove us from the object
            }
            xRemWaitInsRdyList(p);              // Remove from wait Insert in ready
        }    
        else {
            pTMR tmr = (pTMR)p;                 // change the type to timer
            (tmr->pFunction)(tmr->pParam);      // call the timer
            xRemWaitList((pTCB)tmr);
            if (tmr->Status&xTMR_DELETE) {
                xObjFree((pOBJ)tmr);            // Remove from the timer pool
            }                                
            else {
                if (tmr->Status&xTMR_PERIODIC && (tmr->Status&xTMR_STOP)==0) {
                    tmr->WakeUpTime += -tmr->WaitTime;
                    xInsRtccList((pTCB)tmr);    // Insert (Remove was done above)
                }
            }
        }    
    }
    if (qvRtccList) {
        yRtccSetAlarm(qvRtccList->WakeUpTime);
    }    
}

