/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"



pOBJ xObjOpenTO(char* pName, unsigned ErrId, uint8_t Type, int32_t TimeOut) {
    pOBJ p;
    uint8_t OpenType;

#ifndef __NO_CHECK__
    // We can't use the wait function because we don't have an object yet.
    if (qKrnInIsr()) {
        xErrNtf(ErrId+2);                       // XXX_ISR
        return NULL;
    }
    if (!xKrnIsOsStarted()) {
        xErrNtf(ErrId+1);                       // XXX_NO_START
        return NULL;
    }
    if (qKrnInFiber()) {
        xErrNtf(ErrId+3);                       // XXX_FBR
        return NULL;
    }
    if (TimeOut>=0) {
        if (TimeOut<qvClockFrequencyDiv1000000) {
            xErrNtf(ErrId+5);                   // XXX_TIMEOUT
            return NULL;
        }
    }
	else {
		if (qvRtcc==0) {
			xErrNtf(ErrId+11);				// XXX_NO_RTCC
			return NULL;
		}
	}
    if (qKrnInCritical()) {
        xErrNtf(ErrId+9);                       // XXX_CRITICAL
        return NULL;
    }    
    if (pName==0) {
        xErrNtf(ErrId+8);                       // XXX_N0_NAME
        return NULL;
    }
#endif

    if (Type == xTYPE_TCB)
        p = (pOBJ)xGetTcb(pName);
    else
        p = xObjSearch(pName, Type);            // get the object
    if (p) return p;

    qCrtEnter();
    OpenType = Type + 0x80;                     // change from normal type to _OPN_ type
    p = xObjSearch(pName, OpenType);            // something else waiting 
    if (!p) {                                   // Nothing currently waiting
        p = xObjAlloc(OpenType,0);
        if (!p) {
            xErrNtf(ErrId+6);
            return NULL;
        }
        p->pName = pName;
        p->Type = OpenType;
    }
    xThrTimeoutOnObject(p,TimeOut);             // Suspend task until event or timeout occurs
                                                // EXIT_CRITICAL will switch the thread because 
                                                // the wait-time has set the kernel interrupt
    if (qvCurTcb->TimedOut) {                   // DID we timeout??
        return NULL;                            // Timeout error
    }
    qCrtEnter();
    if (Type == xTYPE_TCB)
        p = (pOBJ)xGetTcb(pName);
    else
        p = xObjSearch(pName, Type);            // get the object
    qCrtExit();
    return p;
}

pEVT qEvtOpenTO(char* pName, int32_t TimeOut) {
    return (pEVT)xObjOpenTO(pName, qERR_EVT_ID, xTYPE_EVT, TimeOut);
}

pMTX qMtxOpenTO(char* pName, int32_t TimeOut) {
    return (pMTX)xObjOpenTO(pName, qERR_MTX_ID, xTYPE_MTX, TimeOut);
}

pPIP qPipOpenTO(char* pName, int32_t TimeOut) {
    return (pPIP)xObjOpenTO(pName, qERR_PIP_ID, xTYPE_PIP, TimeOut);
}

pQUE qQueOpenTO(char* pName, int32_t TimeOut) {
    return (pQUE)xObjOpenTO(pName, qERR_QUE_ID, xTYPE_QUE, TimeOut);
}

pSEM qSemOpenTO(char* pName, int32_t TimeOut) {
    unsigned p =(unsigned)xObjOpenTO(pName, qERR_SEM_ID, xTYPE_SEM, TimeOut);
    p = p + 4*sizeof(int);
    return (pSEM)p;
}

pTCB qThrOpenTO(char* pName, int32_t TimeOut) {
    return (pTCB)xObjOpenTO(pName, qERR_THR_ID, xTYPE_TCB, TimeOut);
}

pTMR qTmrOpenTO(char* pName, int32_t TimeOut) {
    return (pTMR)xObjOpenTO(pName, qERR_TMR_ID, xTYPE_TMR, TimeOut);
}

pPUB qPubOpenTO(char* pName, int32_t TimeOut) {
    return (pPUB)xObjOpenTO(pName, qERR_PUB_ID, xTYPE_PUB, TimeOut);
}

