/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  **
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  **
 **  <http://www.quasarsoft.com/license.html>                                       **
 **                                                                                 **
 *************************************************************************************/
#include "qKernel.h"
volatile uint32_t qvUpTime = 0;            // UpTime of the device
volatile uint32_t qvAlarm = 0;             // The alarm Date-Time in DTM format (nbr sec since 1/1/2000 00:00:00)
uint32_t qvDatTim = 0;            // Current DateTime in emulation mode
volatile uint16_t  qvHalfSec = 0;           // 0=in first half  1=second half

//
// This function is called every half-second and is called in a fiber
// This means that we are in a critical section
//
__attribute__((weak, far))
void yRtccSetDatTim(uint32_t DatTim) {
    qCrtEnter();
    qvDatTim = DatTim;
    qCrtExit();
}

//
// This function is called every half-second
//
static void y500mSec(void* p) {
    qCrtEnter();
    if (qvHalfSec==0) {
        qvHalfSec=1;
        qCrtExit();
        return;
    }
    else {
        qvHalfSec=0;
    }
    qvDatTim++;
    qvUpTime++;
    if (qvAlarm!=0) {
        if (qvDatTim>=qvAlarm) {
            qvAlarm = 0;
            xFbrRTCCExpire();
        }
    }
    qCrtExit();
}

__attribute__((weak, far))
uint8_t yRtccInit() {
    //Create a 500mSec timer.
    pTMR tmr = qTmrCreate("RTCC Emulator", y500mSec, 0, TIMER_TYPE_PERIODIC_MANUAL_START, 500000);
    // The start of the timer is queued so it will be started when the OS will start
    qFbrEnqueue1((void (*)(void*))qTmrStart, tmr);
    return 1;
}

//
//    This function returns the current time to the user
//
__attribute__((weak, far))
uint32_t yRtccGetDatTim() {
    uint32_t datTim;
    qCrtEnter();
    datTim = qvDatTim;
    qCrtExit();
    return datTim;
}

//
//    This function returns the current time to the user + 1 if in the second part of the half Second
//
__attribute__((weak, far))
uint32_t yRtccGetDatTimRounded() {
    uint32_t datTim;
    qCrtEnter();
    datTim = qvDatTim + qvHalfSec;
    qCrtExit();
    return datTim;
}

//
//    This function set the alarm time
//
__attribute__((weak, far))
void yRtccSetAlarm(uint32_t alarm){
    qCrtEnter();
    qvAlarm = alarm;
    qCrtExit();
}

//
//    This function returns the number of seconds the system is up.
//
__attribute__((weak, far))
uint32_t yRtccGetUptime() {
    uint32_t upTime;
    qCrtEnter();
    upTime = qvUpTime;
    qCrtExit();
    return upTime;
}

