/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"

void qKrnStatOn() {
    pTCB p; 
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_KRN_ISR);
        return;
    }
#endif
    qCrtEnter();
    if (!xKrnIsStatisticsOn()) {
        p = qvIdleTcb;
        while (p) {
            p->Cycles = 0;
            p = p->pTcbList;
        }
        xKrnSetStatisticsOn();
    }
    qCrtExit();
}



void qKrnStatOff() {
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_KRN_ISR);
        return;
    }
#endif
    qCrtEnter();
    xKrnClrStatisticsOn();
    qCrtExit();
}


uint64_t qThrStatCycles(pTCB pTcb) {
    uint64_t val;
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_THR_ISR);
        return 0;
    }
    if (pTcb!=0 && pTcb->Type!=xTYPE_TCB) {
        xErrNtf(qERR_THR_ID);
        return 0;
    }
    if (!xKrnIsStatisticsOn()) {
        xErrNtf(qERR_THR_STAT_OFF);
        return 0;
    }
#endif
    qCrtEnter();
    if (pTcb==0) {
        pTcb = qvIdleTcb;
    }
    if (pTcb==qvCurTcb) {
        val = pTcb->Cycles + yGetCycles() - pTcb->LastCycles;
    }
    else {
        val = pTcb->Cycles;
    }
    qCrtExit();
    return val;
}

uint64_t qThrStatLostCycles() {
    uint64_t val;
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_THR_ISR);
        return 0;
    }
    if (!xKrnIsStatisticsOn()) {
        xErrNtf(qERR_THR_STAT_OFF);
        return 0;
    }
#endif
    qCrtEnter();
    val = qvLostCycles;
    qCrtExit();
    return val;
}

uint64_t qFbrStatCycles() {
    uint64_t tot=0;
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_FBR_ISR);
        return 0;
    }
    if (!xKrnIsStatisticsOn()) {
        xErrNtf(qERR_FBR_STAT_OFF);
        return 0;
    }
#endif
    pTCB pTcb = qvIdleTcb;                      // Start of all TCB's
    while (pTcb) {                              // Search the Tcb from the beginning
        tot += qThrStatCycles(pTcb);
        pTcb = pTcb->pTcbList;
    }
    return qTimCycles() - tot - qvLostCycles;
}
